import sys
import argparse

from eeprom import EEPROM

# Try to import ouroboros for our MAC address server. If this file cannot be imported, then we cannot program the mac address, but we can setup the EEPROM.
# This is used in recovery images, so we can safely give these to end users without handing out the keys to the mac address programming server.
try:
    from ouroboros import Ouroboros
except ImportError:
    Ouroboros = None


## This file contains the main script to configure our EEPROM with a BOM number, revision number.
#  It also handles setting a new GUID if there is none in the EEPROM.
#  It only handles MAC address setting when our ouroboros module is available.
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Program the BOM number and MAC address.')
    parser.add_argument('BOM', metavar='BOM', type=int, help='The BOM number that needs to be programmed.')
    parser.add_argument('Revision', metavar='Revision', type=int, nargs='?', help='The revision number that needs to be programmed.')

    args = parser.parse_args()
    
    eeprom = EEPROM()
    if not eeprom.isValidGuid():
        print("Generated new GUID as no valid GUID was programmed yet")
        eeprom.generateNewGuid()

    if eeprom.getBom() != args.BOM:
        print("Changing BOM number into: %d from %d" % (args.BOM, eeprom.getBom()))
        eeprom.setBom(args.BOM)

    if args.Revision is not None:
        if eeprom.getRevision() != args.Revision:
            print("Changing Revision number into: %d from %d" % (args.Revision, eeprom.getRevision()))
            eeprom.setRevision(args.Revision)

    if not eeprom.isValidMac():
        if Ouroboros:
            print("Requesting a MAC address, as current is invalid. %s" % (eeprom.getMac()))
            data = Ouroboros().request("request/client/%d-A/%s/" % (eeprom.getBom(), eeprom.getGuid()))
            if data is None:
                print("Failed to get MAC address")
                sys.exit(1)
            eeprom.setMac(data['mac'].replace(":", ""))
            print("Stored MAC address: %s" % (data['mac']))
            Ouroboros().request("commit/", {"appliance": "client", "bom": "%d-A" % (eeprom.getBom()), "uuid": eeprom.getGuid(), "mac": data['mac'].replace(":", "")})
        else:
            print("Not a valid mac address, but we don't have ouroboros.")
    sys.exit(0)
