import struct
import uuid
import binascii

from crc8 import Crc8


## Class to handle reading and writing of our board EEPROM.
class EEPROM:
    # Location of the EEPROM in the filesystem (eeprom is exposed as a file-like object by linux)
    __SYSTEM_EEPROM_FILE = "/sys/bus/i2c/devices/1-0050/eeprom"
    # Offsets for various parts in the EEPROM see docs/Board_EEPROM_layout.txt for details.
    __SYSTEM_EEPROM_MAC_OFFSET = 0x008
    __SYSTEM_EEPROM_BOM_OFFSET = 0x100
    __SYSTEM_EEPROM_REV_OFFSET = 0x104
    __SYSTEM_EEPROM_GUID_OFFSET = 0x108

    ## Create a new EEPROM object, this reads the data from the EEPROM and stores it in private members.
    def __init__(self):
        with open(self.__SYSTEM_EEPROM_FILE, "rb") as f:
            f.seek(self.__SYSTEM_EEPROM_MAC_OFFSET)
            self.__mac = binascii.hexlify(f.read(6)).decode('ascii')
            if Crc8(binascii.unhexlify(self.__mac)).digest() != f.read(1):
                print(binascii.hexlify(Crc8(binascii.unhexlify(self.__mac)).digest()), binascii.hexlify(f.read(1)))
                self.__mac = "ffffffffffff"
            f.seek(self.__SYSTEM_EEPROM_BOM_OFFSET)
            self.__bom_number = struct.unpack("!I", f.read(4))[0]
            f.seek(self.__SYSTEM_EEPROM_REV_OFFSET)
            self.__rev_number = struct.unpack("!I", f.read(4))[0]
            f.seek(self.__SYSTEM_EEPROM_GUID_OFFSET)
            self.__guid = str(uuid.UUID(bytes=f.read(16)))
            
    
    ## Check if the GUID we have is valid. On an empty EEPROM there is no valid GUID.
    #  @return bool: True if the GUID is valid, False if not
    def isValidGuid(self):
        if self.__guid == "00000000-0000-0000-0000-000000000000":
            return False
        if self.__guid == "ffffffff-ffff-ffff-ffff-ffffffffffff":
            return False
        return True
    
    ## Generate a new GUID and store it in the EEPROM
    def generateNewGuid(self):
        with open(self.__SYSTEM_EEPROM_FILE, "r+b") as f:
            f.seek(self.__SYSTEM_EEPROM_GUID_OFFSET)
            f.write(uuid.uuid4().bytes)
            f.seek(self.__SYSTEM_EEPROM_GUID_OFFSET)
            self.__guid = str(uuid.UUID(bytes=f.read(16)))

    ## Get our current GUID
    # @return uuid.UUID object containing the GUID.
    def getGuid(self):
        return self.__guid

    ## Check if our MAC address is valid.
    #  The address is valid if it starts with a known range from Ultimaker.
    #  @return bool: True if the MAC address is valid, False if not.
    def isValidMac(self):
        if not self.__mac.startswith('ccbdd3'):
            return False
        return True
    
    ## Get our current mac address.
    #  @return MAC address as hex string (example: "ccbdd3000000")
    def getMac(self):
        return self.__mac

    ## Set a new MAC address in the EEPROM.
    #  @param new_mac: new MAC address as hex string (example: "ccbdd3000000")
    def setMac(self, new_mac):
        with open(self.__SYSTEM_EEPROM_FILE, "r+b") as f:
            f.seek(self.__SYSTEM_EEPROM_MAC_OFFSET)
            f.write(binascii.unhexlify(new_mac))
            f.write(Crc8(binascii.unhexlify(new_mac)).digest())
            f.seek(self.__SYSTEM_EEPROM_MAC_OFFSET)
            self.__mac = binascii.hexlify(f.read(6)).decode('ascii')

    ## Get the current BOM number as number
    #  @return integer: bom number
    def getBom(self):
        return self.__bom_number

    ## Update the current bom number in the EEPROM.
    #  @param bom: integer containing the new BOM number.
    def setBom(self, bom):
        with open(self.__SYSTEM_EEPROM_FILE, "r+b") as f:
            f.seek(self.__SYSTEM_EEPROM_BOM_OFFSET)
            f.write(struct.pack("!I", bom))
            f.seek(self.__SYSTEM_EEPROM_BOM_OFFSET)
            self.__bom_number = struct.unpack("!I", f.read(4))[0]

    ## Get the current revision number as number
    #  @return integer: revision number
    def getRevision(self):
        return self.__rev_number

    ## Update the current revision number in the EEPROM.
    #  @param rev: integer containing the new revision number.
    def setRevision(self, rev):
        with open(self.__SYSTEM_EEPROM_FILE, "r+b") as f:
            f.seek(self.__SYSTEM_EEPROM_REV_OFFSET)
            f.write(struct.pack("!I", rev))
            f.seek(self.__SYSTEM_EEPROM_REV_OFFSET)
            self.__rev_number = struct.unpack("!I", f.read(4))[0]
