import re
import logging

from typing import Dict, List, Optional

from app.libs.ultimaker3.build_plate import BuildPlate
from um3_commons.math.vector3 import Vector3
from um3_commons.math.axisAlignedBox import AxisAlignedBox
from app.models.extruder_train import ExtruderTrain
from app.models.machine_variant import MachineVariant
from app.libs.ultimaker3.material import Material

logger = logging.getLogger(__name__)


class GCodeParser:
    """
    Read and parse gcode contents into this object.
    """
    HEADER_PARSING_REGEX = "^;([A-Z0-9_\.]+):(.+)$"
    MATCH_EXTRUDER_REGEX = r"^EXTRUDER_TRAIN.(\d+)\.(.+)$"

    KEY_HEADER_VERSION = "HEADER_VERSION"
    KEY_FLAVOR = "FLAVOR"
    KEY_PRINT_SIZE_MIN_X = "PRINT.SIZE.MIN.X"
    KEY_PRINT_SIZE_MIN_Y = "PRINT.SIZE.MIN.Y"
    KEY_PRINT_SIZE_MIN_Z = "PRINT.SIZE.MIN.Z"
    KEY_PRINT_SIZE_MAX_X = "PRINT.SIZE.MAX.X"
    KEY_PRINT_SIZE_MAX_Y = "PRINT.SIZE.MAX.Y"
    KEY_PRINT_SIZE_MAX_Z = "PRINT.SIZE.MAX.Z"
    KEY_TARGET_MACHINE_NAME = "TARGET_MACHINE.NAME"

    KEY_EXTRUDER_MATERIAL_GUID = "MATERIAL.GUID"
    KEY_EXTRUDER_NOZZLE_DIAMETER = "NOZZLE.DIAMETER"
    KEY_EXTRUDER_NOZZLE_NAME = "NOZZLE.NAME"

    KEY_BUILDPLATE_TYPE = "BUILD_PLATE.TYPE"

    EXPECTED_KEYS = {
        KEY_HEADER_VERSION,
        KEY_FLAVOR,
        KEY_PRINT_SIZE_MIN_X,
        KEY_PRINT_SIZE_MIN_Y,
        KEY_PRINT_SIZE_MIN_Z,
        KEY_PRINT_SIZE_MAX_X,
        KEY_PRINT_SIZE_MAX_Y,
        KEY_PRINT_SIZE_MAX_Z,
        KEY_TARGET_MACHINE_NAME,
    }

    EXPECTED_EXTRUDER_KEYS = {
        KEY_EXTRUDER_MATERIAL_GUID,
    }

    DIMENSION_KEYS = {
        KEY_PRINT_SIZE_MIN_X,
        KEY_PRINT_SIZE_MIN_Y,
        KEY_PRINT_SIZE_MIN_Z,
        KEY_PRINT_SIZE_MAX_X,
        KEY_PRINT_SIZE_MAX_Y,
        KEY_PRINT_SIZE_MAX_Z,
    }

    FLOAT_KEYS = DIMENSION_KEYS

    def __init__(self, file_path: str) -> None:
        self.__headers = {}             # type: Dict[str, str]
        self._file_path = file_path
        self.__extruder_trains = {}     # type: Dict[int, ExtruderTrain]
        self.is_valid = self.__parse_headers()
        self.__populate_extruder_trains()
        self.__is_supported = self.__evaluate_is_supported()
        self.__buildplate_type = self.__headers.get(self.KEY_BUILDPLATE_TYPE, "glass").lower()  # type: str

    @property
    def headers(self) -> Dict[str, str]:
        return self.__headers

    @classmethod
    def get_supported_file_types(cls) -> List[str]:
        return ["gcode", "g"]

    def open_stream(self):  # return a filestream
        return open(self._file_path, "r")

    def __parse_headers(self) -> bool:
        self.__headers = {}

        try:
            with self.open_stream() as file:
                for line in file:
                    match = re.match(self.HEADER_PARSING_REGEX, line)

                    if match is not None:
                        header_name = match.group(1)

                        if self.__is_non_header_variable(header_name):
                            return True

                        header_value = match.group(2)
                        self.__headers[header_name] = header_value
        except FileNotFoundError:
            return False
        except OSError:
            logger.exception("OSError on file: %s" % self._file_path)
            return False
        return True

    @staticmethod
    def __is_non_header_variable(name: str) -> bool:
        return name in ["LAYER"]

    def __populate_extruder_trains(self) -> None:
        for extruder_index, data in self.__parse_extruder_train_data().items():
            self.__extruder_trains[extruder_index] = ExtruderTrain(
                extruder_index=extruder_index,
                material_guid=data.get(self.KEY_EXTRUDER_MATERIAL_GUID, ""),
                nozzle_diameter=data.get(self.KEY_EXTRUDER_NOZZLE_DIAMETER, ""),
                print_core_id=data.get(self.KEY_EXTRUDER_NOZZLE_NAME, None)
            )

    def __evaluate_is_supported(self) -> bool:
        """
        Return if current g-code is supported.

        If supported, all functions in the class work.
        """
        if not self.__extruder_trains:
            logger.warning("G-code file [%s]: does not contain any extruder train data", self._file_path)
            return False
        for key in self.EXPECTED_KEYS:
            if key not in self.__headers:
                logger.warning("G-code file [%s]: it should contain key [%s], but it does not",
                               self._file_path, key)
                return False
        for extruder_index, extruder_data in self.__parse_extruder_train_data().items():
            for key in self.EXPECTED_EXTRUDER_KEYS:
                if key not in extruder_data:
                    logger.warning("G-code file [%s]: extruder %s does not contain key for %s",
                                   self._file_path, extruder_index, key)
                    return False
            if self.KEY_EXTRUDER_NOZZLE_NAME not in extruder_data and \
               self.KEY_EXTRUDER_NOZZLE_DIAMETER not in extruder_data:
                logger.warning("G-code file [%s]: one of these should be in the header per extruder: %s, %s",
                               self._file_path, self.KEY_EXTRUDER_NOZZLE_DIAMETER, self.KEY_EXTRUDER_NOZZLE_NAME)
                return False

        if self.__headers[self.KEY_FLAVOR] != "Griffin":
            logger.warning("G-code file [%s]: only supporting g-code flavor Griffin, found [%s]",
                           self._file_path, self.__headers[self.KEY_FLAVOR])
            return False
        if self.__headers[self.KEY_HEADER_VERSION] != "0.1":
            logger.warning("G-code file [%s]: only supporting header version 0.1, found [%s]",
                           self._file_path, self.__headers[self.KEY_HEADER_VERSION])
            return False
        for key in self.FLOAT_KEYS:
            try:
                float(self.__headers[key])
            except ValueError:
                logger.warning("G-code file [%s]: could not parse key %s value %s",
                               self._file_path, key, self.__headers[key])
                return False
        return True

    def __parse_extruder_train_data(self) -> Dict[int, Dict[str, str]]:
        extruder_data = {}  # type: Dict[int, Dict[str, str]]

        for key, value in self.headers.items():
            match = re.match(self.MATCH_EXTRUDER_REGEX, key)
            if match is not None:
                extruder_index = int(match.group(1))
                extruder_property_name = match.group(2)

                if extruder_index not in extruder_data:
                    extruder_data[extruder_index] = {}

                extruder_data[extruder_index][extruder_property_name] = value

        return extruder_data

    @property
    def machine_variant(self) -> MachineVariant:
        if 'TARGET_MACHINE.NAME' in self.__headers:
            return MachineVariant(self.__headers['TARGET_MACHINE.NAME'])
        return None

    @property
    def extruder_trains(self) -> Dict[int, ExtruderTrain]:
        return self.__extruder_trains

    @property
    def materials(self) -> Dict[int, Material]:
        materials = {}  # type: Dict[int, Material]

        for extruder_index, extruder_train in self.extruder_trains.items():
            materials[extruder_index] = extruder_train.material

        return materials

    @property
    def print_cores(self) -> Dict[int, str]:
        print_core_ids = {}

        for extruder_index, extruder_train in self.extruder_trains.items():
            print_core_ids[extruder_index] = extruder_train.print_core_id

        return print_core_ids

    @property
    def print_time(self) -> int:
        return int(self.__headers.get('PRINT.TIME', 0))

    @property
    def is_supported(self) -> bool:
        return self.__is_supported

    @property
    def boundary_box(self) -> AxisAlignedBox:
        v_min = Vector3(x=float(self.__headers.get(self.KEY_PRINT_SIZE_MIN_X, 0)),
                        y=float(self.__headers.get(self.KEY_PRINT_SIZE_MIN_Y, 0)),
                        z=float(self.__headers.get(self.KEY_PRINT_SIZE_MIN_Z, 0)))
        v_max = Vector3(x=float(self.__headers.get(self.KEY_PRINT_SIZE_MAX_X, 0)),
                        y=float(self.__headers.get(self.KEY_PRINT_SIZE_MAX_Y, 0)),
                        z=float(self.__headers.get(self.KEY_PRINT_SIZE_MAX_Z, 0)))
        return AxisAlignedBox(v_min, v_max)

    @property
    def build_plate(self):
        return BuildPlate({"type": self.__buildplate_type})
