from um3_commons import thread
from app.libs.signal import Signal
import logging
import re
import copy
import time
from typing import List
from typing import Optional
from app.models.machine_variant import MachineVariant


logger = logging.getLogger(__name__)

try:
    # We allow the dbusif import to fail, in case of a dev environment without dbus installed.
    from um3_commons import dbusif
except ImportError:
    dbusif = None
except:
    # General exception on dbusif import. This can happen on OSX, which has a python-dbus implementation that is
    # not compatible with our dbusif implementation.
    logger.exception("Exception during import of dbusif. Most likely something is broken in your dbus setup.")
    dbusif = None

SERVICE_TYPE = u'_ultimaker._tcp.local.'


class DiscoveredPrinter:
    def __init__(self, unique_name: str, friendly_name: str, ip_address: str, machine_variant: Optional[str]) -> None:
        self.unique_name = unique_name
        self.friendly_name = friendly_name
        self.ip_address = ip_address
        self.machine_variant = machine_variant
        self.update_time = time.monotonic()

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "DiscoveredPrinter(unique_name='%s', friendly_name='%s', ip_address=%s, machine_variant='%s')" % \
               (self.unique_name, self.friendly_name, self.ip_address, self.machine_variant)


class PrinterDiscovery:
    """
    Singleton object that handles the discovery of printers on the network through zeroconf.
    This class talks to the avahi daemon for service discovery. It will auto-retry resolving,
    which can fail in rare cases when UDP packets are dropped.

    It can be used event and polling based.
    """
    __instance = None

    @staticmethod
    def getInstance():
        return PrinterDiscovery.__instance

    @staticmethod
    def startPrinterDiscovery():
        if PrinterDiscovery.__instance is None:
            PrinterDiscovery()

    @staticmethod
    def stopPrinterDiscovery():
        # We can stop the printer discovery, we use this for unit tests to properly clean up after ourselves.
        if PrinterDiscovery.__instance is not None:
            PrinterDiscovery.__instance.__stop()
            PrinterDiscovery.__instance = None

    # The values for the avahi server status
    __SERVER_INVALID, __SERVER_REGISTERING, __SERVER_RUNNING, __SERVER_COLLISION, __SERVER_FAILURE = range(0, 5)
    __AVAHI_IF_UNSPEC = -1
    __AVAHI_PROTO_UNSPEC, __AVAHI_PROTO_INET, __AVAHI_PROTO_INET6 = -1, 0, 1

    __SERVICE_TYPE = "_ultimaker._tcp"

    # Delay between the polling to resolve known or timeout services. This delay is here to prevent busy looping.
    __RESOLVE_POLL_DELAY_IN_SECONDS = 1.0
    # Time between refreshes of the resolved printer data on printers that are known in the network.
    # We occasionally update this data to get new printer names.
    # Setting this time too low will cause a lot of broadcast traffic, which will make network administrators very angry.
    __REFRESH_SERVICE_INFO_TIME = 30.0 * 60.0

    def __init__(self):
        assert PrinterDiscovery.__instance is None, \
            "PrinterDiscovery is a singleton, so we should only create a single instance"
        PrinterDiscovery.__instance = self

        self.__discovered_services_lock = thread.RLock()
        self.__discovered_services = {}
        
        self.__active = True
        self.__resolve_thread = None

        self.__avahi_server = None
        self.__service_browser = None

        # on_new_printer signal, is called when a new printer is discovered and resolved
        self.on_new_printer = Signal[DiscoveredPrinter]()
        # on_removed_printer signal, is called when a discovered printer is no longer discoverable
        self.on_removed_printer = Signal[DiscoveredPrinter]()

        if dbusif is None:
            logger.warning("No dbus python module available: No printer discovery available.")
            return

        # Connect to the avahi main dbus service, and request a browser for us to use.
        try:
            self.__avahi_server = dbusif.RemoteObject("org.freedesktop.Avahi", "/", "org.freedesktop.Avahi.Server")
        except dbusif.DBusException:
            logger.warning("Could not connect to avahi service on dbus. No printer discovery available.")
            return

        # The avahi server reports itself on dbus before it can actually handle requests
        # We need to monitor the state until it states that it is running before we can
        # restart monitoring for printers on the network.
        self.__avahi_server.connectSignal("StateChanged", self.__avahi_server_state_changed)
        self.__avahi_server_state_changed(self.__avahi_server.GetState(), "")

    def __avahi_server_state_changed(self, state, error):
        if state == PrinterDiscovery.__SERVER_RUNNING and self.__service_browser is None:
            browser_path = self.__avahi_server.ServiceBrowserNew(self.__AVAHI_IF_UNSPEC,
                                                                 self.__AVAHI_PROTO_INET,
                                                                 self.__SERVICE_TYPE, "", 0)

            # Connect to the requested browser, and start listening for new services.
            self.__service_browser = dbusif.RemoteObject("org.freedesktop.Avahi",
                                                         browser_path,
                                                         "org.freedesktop.Avahi.ServiceBrowser")
            self.__service_browser.connectSignal("ItemNew", self.__new_service)
            self.__service_browser.connectSignal("ItemRemove", self.__remove_service)

            self.__resolve_thread = thread.Thread("Printer discovery resolve thread", self.__resolve_handler).start()
            logger.info("Started printer discovery.")

    def __new_service(self, interface: int, protocol: int, hostname: str, type: str, domain: str, flags: int):
        logger.debug("New printer service: %s", hostname)
        # interface = IF_UNSPEC, protocol = PROTO_INET or PROTO_INET6, name = hostname, type = __SERVICE_TYPE, domain = "local", flags = ?
        with self.__discovered_services_lock:
            # Add this as a discovered service, but without any data.
            # The service resolve thread will fill in the data, as we need to query this from avahi, but this query can timeout/fail.
            self.__discovered_services[hostname] = None

    def __remove_service(self, interface: int, protocol: int, hostname: str, type: str, domain: str, flags: int):
        logger.debug("Remove printer service: %s", hostname)
        with self.__discovered_services_lock:
            if hostname in self.__discovered_services:
                printer = self.__discovered_services[hostname]
                del self.__discovered_services[hostname]
                if printer is not None:
                    logger.debug("Printer is no longer discoverable %s", printer)
                    self.on_removed_printer.emit(printer)

    def __resolve_handler(self):
        """
        The resolve handler is a thread function that will keep polling unresolved but discovered printers.
        As resolving can fail, we do need to repeat discovery. We cannot do the resolving in the dbus signal for
        the new service discovery, as we will never get the remove service signal if we just loop there.
        As we would block the dbus main loop.
        """
        while self.__active:
            time.sleep(self.__RESOLVE_POLL_DELAY_IN_SECONDS)
            with self.__discovered_services_lock:
                hostnames = []
                for hostname, printer in self.__discovered_services.items():
                    if printer is None or time.monotonic() - printer.update_time > self.__REFRESH_SERVICE_INFO_TIME:
                        hostnames.append(hostname)
            for hostname in hostnames:
                new_discovered_printer_info = self.__resolve(hostname)

                if new_discovered_printer_info is not None:
                    # Update our discovered services list with the new data.
                    with self.__discovered_services_lock:
                        if hostname not in self.__discovered_services:
                            # Name was removed from services while we where resolving the service
                            return
                        old_info_object = self.__discovered_services[hostname]
                        self.__discovered_services[hostname] = new_discovered_printer_info
                        if old_info_object is None:
                            # Newly discovered printer instead of an update on already previously discovered printer.
                            logger.debug("Discovered new printer %s:%s:%s",
                                        hostname, new_discovered_printer_info.friendly_name,
                                        new_discovered_printer_info.ip_address)
                            self.on_new_printer.emit(self.__discovered_services[hostname])

    def __resolve(self, hostname: str) -> DiscoveredPrinter:
        """
        Resolve the discovered printer through avahi, this can fail rarely.
        If resolving is succesful, we will update the discovered services dictionary
        and send out the signal for a new printer.
        """
        logger.debug("Going to resolve printer service: %s", hostname)
        try:
            interface, protocol, hostname, type, domain, host, aprotocol, address, port, txt_list, flags = self.__avahi_server.ResolveService(self.__AVAHI_IF_UNSPEC, self.__AVAHI_PROTO_INET, hostname, self.__SERVICE_TYPE, "local", -1, 0, byte_arrays=True)
        except:
            # Failed to resolve the service. This can happen, and we want to try again
            # unless it gets removed by the RemoveService callback.
            logger.debug("Failed to resolve printer service details: %s", hostname)
            return None

        # Check if the address is an IPv4 address. We occasionally get ipv6 addresses even if we request IPv4.
        # Generally the IPv4 address will follow on the next request.
        if re.match("^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$", address) is None:
            logger.warning("Got strange IP address for: %s:%s", hostname, address)
            return None

        # Unpack the txt data, which is a list of bytes in the form of [key=value].
        # We store this in txt_dict as key:string = value:string, for usage after this.
        txt_dict = {}
        for txt_data in txt_list:
            txt_data = txt_data.decode("utf-8", errors='replace')
            if "=" not in txt_data:
                continue
            key, value = txt_data.split("=", 1)
            txt_dict[key] = value

        # If available, get our friendly name. We currently fall back to the hostname if no friendly name is available.
        friendly_name = txt_dict.get("name", hostname)

        # Map the numeric machine property to Ultimaker 3 machine variants
        machine = None
        machine_prop = txt_dict.get("machine")
        if machine_prop == "9066.0":
            machine = MachineVariant.ULTIMAKER_3
        elif machine_prop == "9511.0":
            machine = MachineVariant.ULTIMAKER_3_EXTENDED

        return DiscoveredPrinter(hostname, friendly_name, address, machine)

    def get_discovered_printers(self) -> List[DiscoveredPrinter]:
        with self.__discovered_services_lock:
            return [copy.deepcopy(printer) for printer in self.__discovered_services.values() if printer is not None]

    def __stop(self):
        logger.info("Stopping the printer discovery")
        self.__active = False
        if self.__resolve_thread is not None:
            self.__resolve_thread.join()
            self.__resolve_thread = None
        if self.__avahi_server is not None:
            self.__avahi_server.cleanSignals()
            self.__avahi_server = None
        if self.__service_browser is not None:
            # We want to stop the discovery so avahi removes the service browser we created.
            # Else we leak resources in avahi
            self.__service_browser.Free()   # Tell Avahi we no longer need this service browser.
            self.__service_browser.cleanSignals()
            self.__service_browser = None
