from threading import Lock
from typing import Any, Dict

from app.libs.signal import Signal


# The _AtomicModificationContext is a context manager that makes all storage backed properties update in one event when
# leaving the context.
# The Modelfiels can be used and set in one action by using this class, this class is instantiated in each
# model as force_atomic_update.
# If your model has multiple ModelFields called field1 and field2 you can do the following:
# with object.force_atomic_update?:
#     object.field1 = "printing"
#     object.field2 = 5
#
# Then as soon as you exit the context of the force_atomic_update in normal circumstances the values will both
# be set by one modify event. However if you exit the context trough a Exception then No event will be triggered.
class _AtomicModificationContext:
    def __init__(self) -> None:
        # Intentionally not an Rlock, since it should just be one context at a time even if you could
        # have multiple additions within the same thread
        self.__lock = Lock()  # type: Lock
        self.__properties = {}  # type: Dict[str, Any]
        self.on_release = Signal[Dict[str, Any]]()  # type: Signal[Dict[str, Any]]

    @property
    def is_taken(self) -> bool:
        return self.__lock.locked()

    def __enter__(self) -> None:
        self.__lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type is None:
            self.on_release.emit(self.__properties)
        self.__properties = {}
        self.__lock.release()

    def add_property(self, key: str, value: Any) -> None:
        assert self.__lock.locked(), "Cannot add properties when not in this context!"
        self.__properties[key] = value
