from typing import Generic, Type, TypeVar, Optional

import logging

MYPY = False
if MYPY:
    from app.models.abstraction.abstract_model import AbstractModel

T = TypeVar('T')
logger = logging.getLogger(__name__)


# If you define a ModelField You can access it's data as if it was just data, even though it is defined in the class
# and not the instance it keeps per instance data, this is the magic that is descriptors...
# Any ModelField defined inside a AbstractModel will cause that data to be stored in A backing store (if there is one).
# Changes to ModelField's data will cause onChange signals to fire in the AbstractModel class.
# If you wish to change multiple fields as one action, you can use the AbstractModel's atomic_modification context.
# when an object is instantiated you HAVE to take the initial_properties context and set the initial values of all
# ModelField's within that context.
# You can use the field along with typing information, but you have to provide the data type, this is double.
class ModelField(Generic[T]):
    def __init__(self, *, data_type: type, read_only: bool=False, field_name: str=None) -> None:
        self.__data_type = data_type
        self.__key = field_name
        self.__read_only = read_only
        self.__label = None  # type: Optional[str]

    # This name is stored in the database
    @property
    def key(self) -> str:
        return self.__key

    @property
    def type(self) -> Type:
        return self.__data_type

    # This is the name the field has in the class, so it could be "_status" instead of the field_name, which might
    # then be "status" error messages on usage make more sense using the first while the DB stores the later.
    @property
    def label(self) -> str:
        return self.__label

    @label.setter
    def label(self, value: str) -> None:
        if self.__label is None:
            self.__label = value
            # if the DB key is not defined use the external name.
            # This prevents the annoying reiterating of the same name.
            if self.__key is None:
                self.__key = value
        else:
            assert self.__label == value

    def __get__(self, instance: "AbstractModel", owner) -> T:
        data = instance._get_via_ModelField(self.__key)
        if not isinstance(data, self.__data_type):
            try:
                data = self.__data_type(data)
            except Exception:
                logger.warning("Could not convert property %s to type %s acting upon data %s",
                               self.__label, self.__data_type, str(data))
        return data

    def __set__(self, instance: "AbstractModel", value: T) -> None:
        if not isinstance(value, self.__data_type):
            try:
                value = self.__data_type(value)
            except TypeError:
                logger.error("Ignored set: Could not convert value for property %s to type %s acting upon data %s",
                             self.__label, self.__data_type, str(value))
                return

        if instance.initial_properties.is_taken:
            instance.initial_properties.add_property(self.__key, value)
            return

        if self.__read_only:
            logger.error("field(%s) is marked as readonly!", self.label)
            return

        if instance.atomic_modification.is_taken:
            instance.atomic_modification.add_property(self.__key, value)
        else:
            instance._set_via_ModelField(self.__key, value)
