import datetime
import logging
import os
import time
import uuid as uuid_lib
from typing import Dict, Iterator, List, Optional

from app.constants.print_job_status import PrintJobStatus
from app.libs import createGCodeParser
from app.libs.ultimaker3.build_plate import BuildPlate
from app.models.abstraction.abstract_model import AbstractModel
from app.models.abstraction.model_field import ModelField
from app.models.configuration import Configuration
from app.models.extruder_train import ExtruderTrain
from app.services.scheduler.result_containers.match_results import ConfigurationChangeMatchResult, MatchResult
from um3_commons.math.axisAlignedBox import AxisAlignedBox

logger = logging.getLogger(__name__)

MYPY = False
if MYPY:
    from app.models.printer import Printer


class PrintJobCreationError(Exception):
    pass


class PrintJob(AbstractModel):
    name = ModelField[str](data_type=str, read_only=True)
    filepath = ModelField[str](data_type=str, read_only=True)
    created_at = ModelField[datetime.datetime](data_type=datetime.datetime, read_only=True)
    owner = ModelField[str](data_type=str, read_only=True)

    status = ModelField[str](data_type=str)
    constraints = ModelField[dict](data_type=dict)
    force = ModelField[bool](data_type=bool)
    external_uuid = ModelField[str](data_type=str)
    printer_uuid = ModelField[str](data_type=str)

    def __init__(
            self,
            name: str,
            filepath: str,
            uuid: Optional[str]=None,
            constraints: Optional[Dict[str, str]]=None,
            status: Optional[str]=None,
            created_at: Optional[datetime.datetime]=None,
            force: bool=False,
            owner: Optional[str]=None,
            external_uuid: Optional[str] = None,
            printer_uuid: Optional[str] = None) -> None:
        super().__init__()
        with self.initial_properties:
            self.uuid = uuid or str(uuid_lib.uuid4())
            self.name = name
            self.filepath = filepath
            self.constraints = constraints or {}
            self.created_at = created_at or datetime.datetime.utcnow()
            self.force = force
            self.owner = owner
            self.status = status or PrintJobStatus.QUEUED
            self.external_uuid = external_uuid
            self.printer_uuid = printer_uuid

        self.time_remaining = 0                         # type: float
        self.firmware_version = None                    # type: Optional[str]
        self.language = None                            # type: Optional[str]
        self.uptime = None                              # type: Optional[int]

        # This dictionary will be used to count errors with other printers based on their uuid
        self.__error_with_printer = {}  # type: Dict[str, int]

        # The assigned_to field indicates a printer that has an active reservation for this print job.
        self.__assigned_to = None   # type: "Printer"

        self.network_error_count = 0
        self.last_successful_check = 0.0  # type: float

        self.time_elapsed = 0

        gcode_parser = createGCodeParser(self.filepath)
        if gcode_parser is None or not gcode_parser.is_valid or not gcode_parser.is_supported:
            raise PrintJobCreationError("Unsupported gcode file on job creation UUID: %s, file [%s]",
                                        self.uuid, self.filepath)
        self.boundary_box = gcode_parser.boundary_box  # type: AxisAlignedBox
        self.time_total = gcode_parser.print_time
        self.machine_variant = gcode_parser.machine_variant
        self.extruder_trains = gcode_parser.extruder_trains  # type: Dict[int, ExtruderTrain]
        self.build_plate = gcode_parser.build_plate  # type: BuildPlate

        self.configuration = Configuration()  # type: Configuration
        self.__configuration_changes_required = []  # type: List[ConfigurationChangeMatchResult.ChangeSet]
        self.__impediments_to_printing = []  # type: List[MatchResult]

    @property
    def job_name(self):
        return os.path.basename(self.filepath)

    def status_check_successful(self) -> None:
        self.network_error_count = 0
        self.last_successful_check = time.monotonic()

    def status_check_failed(self) -> None:
        self.network_error_count += 1

    @property
    def started(self) -> bool:
        return self.status not in [PrintJobStatus.QUEUED]

    @property
    def printing(self) -> bool:
        return self.status in PrintJobStatus.PRINTING_STATUSES

    @property
    def last_seen(self) -> float:
        return time.monotonic() - self.last_successful_check

    @property
    def assigned_to(self) -> Optional[str]:
        if self.__assigned_to is None:
            return None
        return self.__assigned_to.uuid

    @property
    def assigned_to_printer(self) -> Optional["Printer"]:
        return self.__assigned_to

    # This is called from the printer service, only use printerService.[set/remove]_reservation,
    # to ensure the link is bi-directional
    def set_assigned_to(self, printer: Optional["Printer"]):
        self.__assigned_to = printer

    def log_start_error_with_printer(self, printer: "Printer") -> None:
        self.__error_with_printer[printer.uuid] = self.__error_with_printer.get(printer.uuid, 0) + 1

    def get_start_error_count_with_printer(self, printer: "Printer") -> int:
        return self.__error_with_printer.get(printer.uuid, 0)

    @property
    def configuration_changes_required(self) -> List[ConfigurationChangeMatchResult.ChangeSet]:
        return self.__configuration_changes_required

    # This value is exclusively set by the scheduler's executor!
    @configuration_changes_required.setter
    def configuration_changes_required(self, changes: List[ConfigurationChangeMatchResult.ChangeSet]) -> None:
        self.__configuration_changes_required = changes

    @property
    def impediments_to_printing(self) -> List[MatchResult]:
        return self.__impediments_to_printing

    # This value is exclusively set by the scheduler's executor!
    @impediments_to_printing.setter
    def impediments_to_printing(self, impediments: List[MatchResult]):
        self.__impediments_to_printing = impediments

    def __iter__(self) -> Iterator:
        return iter({
            "uuid": self.uuid,
            "name": self.name,
            "filepath": self.filepath,
            "created_at": self.created_at,
            "started": self.started,
            "status": self.status,
            "configuration": self.configuration,
            "machine_variant": self.machine_variant,
            "owner": self.owner,
        })

    def __str__(self) -> str:
        pairs = []
        for key in ["name", "uuid", "status"]:
            pairs.append("%s=%s" % (key, getattr(self, key)))

        return "PrintJob(%s)" % (",".join(pairs))

    def __repr__(self) -> str:
        return "<{}>".format(self)
