import datetime
import time
import logging

from typing import Optional, Tuple, Iterator, Any, List, cast

from app.constants.led_colors import LedColors
from app.constants.printer_status import PrinterStatus
from app.libs.printer_api import PrinterApi, GcodeType
from app.libs.printer_api_exceptions import PrinterApiException
from app.libs.printer_credentials_store import PrinterCredentialsStore
from app.libs.printer_discovery import PrinterDiscovery
from app.libs.signal import Signal
from app.libs.ultimaker3.build_plate import BuildPlate
from app.libs.ultimaker3.print_job import PrintJob as Ultimaker3PrintJob
from app.libs.ultimaker3.status import Status
from app.libs.ultimaker3.maintenance import Maintenance
from app.libs.ultimaker3.system import System
from app.models.abstraction.abstract_model import AbstractModel
from app.models.abstraction.model_field import ModelField
from app.models.configuration import Configuration
from app.models.led_color import LedColor
from app.models.machine_variant import MachineVariant
from app.models.print_job import PrintJob
import uuid as uuid_lib

IP_VERIFY_PERIOD_SECONDS = 30

logger = logging.getLogger(__name__)
try:
    from um3_commons import dbusif
    import dbus
except Exception:
    dbusif = None
    logger.warning("DBUS not available!")


class Printer(AbstractModel):
    unique_name = ModelField[str](data_type=str, read_only=True)
    # The next field is an excellent example of saving the filed as a different name then the field's name in the model.
    __machine_variant = ModelField[str](data_type=str, field_name="machine_variant", read_only=True)

    digest_auth_id = ModelField[str](data_type=str)
    digest_auth_key = ModelField[str](data_type=str)
    digest_auth_authorized = ModelField[bool](data_type=bool)
    enabled = ModelField[bool](data_type=bool)
    ip_address = ModelField[str](data_type=str)
    friendly_name = ModelField[str](data_type=str)
    _status = ModelField[str](data_type=str, field_name="status")

    def __init__(
            self,
            unique_name: str,
            ip_address: str,
            uuid: Optional[str]=None,
            friendly_name: Optional[str]=None,
            status: Optional[str]=None,
            machine_variant: Optional[str]=None,
            digest_auth_id: Optional[str]=None,
            digest_auth_key: Optional[str]=None,
            digest_auth_authorized: Optional[bool]=False,
            enabled: Optional[bool]=True) -> None:
        super().__init__()
        with self.initial_properties:
            self.uuid = uuid or str(uuid_lib.uuid4())
            self.ip_address = ip_address
            self.enabled = enabled
            self.unique_name = unique_name
            self.friendly_name = friendly_name or unique_name
            self.digest_auth_id = digest_auth_id
            self.digest_auth_key = digest_auth_key
            self.digest_auth_authorized = digest_auth_authorized
            self._status = PrinterStatus.UNKNOWN
            self.__machine_variant = machine_variant

        self.__reserved_by = None                       # type: Optional[PrintJob]
        self.configuration = Configuration()
        self.time_remaining = 0                         # type: float
        self.firmware_version = None                    # type: Optional[str]
        self.language = None                            # type: Optional[str]
        self.uptime = None                              # type: Optional[int]

        self._last_verify_time = 0  # type: float
        self.status = status                            # type: Optional[str]

        self.build_plate = BuildPlate({})                # type: BuildPlate

        self.ip_address_changed = Signal[Tuple[Printer, str]]()

        # Last/currently known led color. Updated when the leds are properly updated on the machine.
        # Prevents sending multiple LED color updates when they are not required.
        self.__current_led_color = None  # type: Optional[LedColor]

    @property
    def machine_variant(self) -> MachineVariant:
        return MachineVariant(self.__machine_variant)

    @property
    def reserved_by(self) -> Optional[str]:
        if self.__reserved_by is None:
            return None
        return self.__reserved_by.uuid

    @property
    def reserved_by_job(self) -> Optional[PrintJob]:
        return self.__reserved_by

    # This is called from the printer service, only use printerService.[set/remove]_reservation,
    # to ensure the link is bi-directional
    def set_reserved_by(self, print_job: Optional[PrintJob]):
        self.__reserved_by = print_job

    @property
    def status(self) -> Optional[str]:
        return self._status

    @status.setter
    def status(self, new_status: Optional[str]) -> None:
        if new_status in PrinterStatus.KNOWN:
            logger.debug("Printer [%s] status [%s] -> [%s]", self, self._status, new_status)
            self._status = new_status
        elif new_status is None:
            logger.warning("Printer [%s] status [%s] -> UNKNOWN", self, self._status)
            self._status = PrinterStatus.UNKNOWN
        else:
            logger.warning("Printer [%s] status [%s] -> [%s] (not in PrinterStatus)", self, self._status, new_status)
            self._status = new_status

    def is_authenticated(self) -> bool:
        return self.__get_api().authenticated()

    def create_job(self, job_name: str, file_path: str) -> Optional[str]:
        if not self.__verify_remote_printer():
            logger.error("Could not create job for printer because __verify_remote_printer failed.")
            return None
        try:
            return self.__get_api().create_job(job_name, file_path)
        except PrinterApiException:
            logger.exception("Could not create job for printer because of printer API exception.")
            return None

    def get_current_print_job(self) -> Optional[Ultimaker3PrintJob]:
        if not self.__verify_remote_printer():
            return None
        try:
            return self.__get_api().get_current_print_job()
        except PrinterApiException:
            return None

    def get_print_job_from_history(self, external_job_uuid: str) -> Optional[Ultimaker3PrintJob]:
        if not self.__verify_remote_printer():
            return None
        try:
            return self.__get_api().get_print_job_from_history(external_job_uuid)
        except PrinterApiException:
            return None

    def download_gcode(self, destination_filename: str) -> Optional[GcodeType]:
        if not self.__verify_remote_printer():
            return None

        logger.debug("Downloading gcode from %s", self)
        try:
            return self.__get_api().download_print_job(destination_filename)
        except PrinterApiException:
            logger.exception("Could not download gcode")
            return None

    def set_led_color(self, color: LedColor) -> bool:
        try:
            if self.__current_led_color != color:
                self.__get_api().set_led_color(color)
                self.__current_led_color = color
            return True
        except PrinterApiException:
            return False

    def set_display_message(self, message: str, button_text: str) -> bool:
        """
        Set display message
        """
        logger.debug("Setting display message: %s on %s", message, self)
        try:
            self.__get_api().set_display_message(message, button_text)
            return True
        except PrinterApiException:
            logger.exception("Could not set display message [%s]", message)
            return False

    def clear_display_message(self):
        self.set_display_message("", "")  # Empty stings indicate that the message needs to be removed

    def request_print_job_state(self, target_state: str) -> bool:
        try:
            self.__get_api().request_print_job_state(target_state)
            return True
        except PrinterApiException:
            logger.exception("Could not set print job state [%s]", target_state)
            return False

    def set_default_led_color(self) -> bool:
        color = LedColors.DEFAULT
        if not self.enabled and self.status != PrinterStatus.PRINTING:
            color = LedColors.DISABLED
        return self.set_led_color(color)

    def get_status(self) -> Optional[Status]:
        if not self.__verify_remote_printer():
            return None
        try:
            return self.__get_api().get_status()
        except PrinterApiException:
            return None

    def get_camera_feed_url(self) -> Optional[str]:
        if not self.__verify_remote_printer():
            return None
        return self.__get_api().get_camera_feed_url()

    def get_maintenance_history(self, offset: Optional[int]=None, count: Optional[int]=None)\
            -> Optional[List[Maintenance]]:
        if not self.__verify_remote_printer():
            return None
        try:
            return self.__get_api().get_maintenance_history(offset, count)
        except PrinterApiException:
            return None

    def add_maintenance_history_item(self, action: str, mechanic: Optional[str]=None) -> bool:
        try:
            return self.__get_api().add_maintenance_history_item(action, mechanic)
        except PrinterApiException:
            return False

    def __get_api(self) -> PrinterApi:
        credentials_store = PrinterCredentialsStore(printer=self)
        return PrinterApi(self.ip_address, credentials_store)

    def __verify_remote_printer(self) -> bool:
        # In this function we verify of the remote printer is actually the printer we are expecting at the IP address
        # This to prevent sending information to the wrong printer.
        # This function also searches for a new location in the zeroconf discovery if the printer cannot be
        # found at the previous IP address.
        # We also check if the remote printer has a valid system time, and synchronize that.

        # We have a grace period in which we do not check anything to prevent this call from polling
        # the printer all the time.
        difference_since_last_check = time.monotonic() - self._last_verify_time
        if difference_since_last_check < IP_VERIFY_PERIOD_SECONDS:
            return True
        self._last_verify_time = time.monotonic()

        logger.debug("Verifying the IP %s of printer %s", self.ip_address, self)

        printer_api = self.__get_api()
        try:
            printer_system_info = printer_api.get_system()
        except PrinterApiException:
            logger.warning("Unable to contact printer %s at IP %s", self, self.ip_address)
            return self.__resolve_ip_address()

        self.__verify_printer_system_time(printer_system_info)
        self.__update_printer_with_system_info(printer_system_info)
        self.__update_printer_name(printer_system_info)

        if not self.__verify_remote_printer_system_info(printer_system_info):
            return self.__resolve_ip_address()

        if not self.__get_api().verify_authentication():
            logger.warning("Authorization failed for printer(%s) at verification, disabling credentials", self.friendly_name)
            self.__get_api().credentials.reset()
            return False

        return True

    def __verify_remote_printer_system_info(self, printer_system_info: System) -> bool:
        if printer_system_info.hostname != self.unique_name:
            logger.warning("Expected to find printer %s at IP %s, but saw %s instead.",
                           self, self.ip_address, printer_system_info.hostname)
            return False
        return True

    def __verify_printer_system_time(self, printer_system_info: System) -> None:
        if printer_system_info.time is None:
            return
        # Update the time on the printer if we have a valid system time ourselves and the printer has no system time
        # As any time is better then no time.
        now = datetime.datetime.utcnow()
        if now.year > 1990 and printer_system_info.time.year < 1990:
            try:
                self.__get_api().set_system_time(now.timestamp())
            except PrinterApiException:
                logger.warning("Failed to set time on printer %s at IP %s", self, self.ip_address)
        # If the remote printer has a valid system time, and we have not. Update our own system time.
        if now.year < 1990 and printer_system_info.time.year > 1990:
            try:
                # Cast to Any is done here, because typeshed is wrong on clock_settime see:
                #   https://github.com/python/typeshed/pull/1629
                time.clock_settime(time.CLOCK_REALTIME, cast(Any, printer_system_info.time.timestamp()))
            except:
                logger.warning("Failed to set local time when we do not have a local time.")

    def __update_printer_with_system_info(self, printer_system_info: System) -> None:
        self.uptime = printer_system_info.uptime
        self.firmware_version = printer_system_info.firmware_version
        self.language = printer_system_info.language

    def __update_printer_name(self, printer_system_info: System) -> None:
        if self.friendly_name != printer_system_info.name:
            # while the function above(__update_printer_with_system_info) saves data to the class it does not
            # store anything in the eventstore, this does.
            self.friendly_name = printer_system_info.name

    def __resolve_ip_address(self) -> bool:
        new_ip_addr = self.__find_printer_ip_via_zeroconf()
        if new_ip_addr is None:
            # Zeroconf has proven itself to random and unreliable. Sometimes the host can't even find itself.
            # This can especially happen after changes in the local network adapters. In this case it is
            # useful to directly examine the local machine's active IP addresses to find itself.
            new_ip_addr = self.__find_printer_ip_via_network_adapters()
        if new_ip_addr is not None:
            logger.warning("Found printer %s at IP %s", self, new_ip_addr)
            self.ip_address = new_ip_addr
            self.ip_address_changed.emit((self, new_ip_addr))
            return True
        logger.warning("Unable to contact printer %s at IP %s or find it via zeroconf.", self, self.ip_address)
        return False

    def __find_printer_ip_via_zeroconf(self) -> Optional[str]:
        for printer in PrinterDiscovery.getInstance().get_discovered_printers():
            ZEROCONF_HOSTNAME_SUFFIX = "._ultimaker._tcp.local."
            if printer.unique_name.endswith(ZEROCONF_HOSTNAME_SUFFIX):
                discovered_unique_name = printer.unique_name[:-len(ZEROCONF_HOSTNAME_SUFFIX)]
            else:
                discovered_unique_name = printer.unique_name

            if discovered_unique_name == self.unique_name:
                return printer.ip_address
        return None

    def __find_printer_ip_via_network_adapters(self) -> Optional[str]:
        # Check the network interfaces on this printer. This is a special
        # case for when the host printer is looking for itself.
        if dbusif is not None:
            local_unique_name = dbusif.RemoteObject("system").getHostName()
            if local_unique_name == self.unique_name:
                local_ip_address = dbusif.RemoteObject("network").getIPv4Address()
                if local_ip_address is not None and local_ip_address != "":
                    return local_ip_address
        return None

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        # This method is mostly for the benefit of Flask when it converts
        # one of these objects to JSON before sending it back across the
        # network. It only needs to have the properties which are set.
        dict_properties = {"uuid": self.uuid,
                           "unique_name": self.unique_name,
                           "ip_address": self.ip_address,
                           "enabled": self.enabled,
                           "reserved_by": self.reserved_by}

        if self.digest_auth_id is not None:
            dict_properties["digest_auth_id"] = self.digest_auth_id
        if self.digest_auth_authorized is not None:
            dict_properties["digest_auth_authorized"] = self.digest_auth_authorized
        if self._status is not None:
            dict_properties["status"] = self._status

        return iter(dict_properties.items())

    def __str__(self) -> str:
        return "Printer(unique_name=%s, friendly_name=%s, uuid=%s, " \
               "enabled=%s, reserved_by=%s, status=%s, authenticated=%s)" % (
               self.unique_name, self.friendly_name, self.uuid, self.enabled, self.reserved_by,
               self._status, self.is_authenticated())

    def __repr__(self) -> str:
        return "<{}>".format(self)
