import logging
import os.path
import shutil
import uuid

from typing import Tuple, Optional

from app.libs.printer_api import GcodeType
from app.libs.signal import Signal
from app.models.printer import Printer
from um3_commons.thread import Thread

logger = logging.getLogger(__name__)


class AlienPrintJobDownloader:
    def __init__(self, printer: Printer, gcode_spool_directory: str, job_uuid: str, filename: str) -> None:
        self.__printer = printer
        self.__job_uuid = job_uuid
        self.__gcode_spool_directory = gcode_spool_directory
        self.__print_job_filename = filename
        self.__gcode_temp_filename = None   # type: Optional[str]
        self.__thread = Thread(
            name="alien print job downloader for printer {}".format(printer.unique_name),
            target=self.__run
        )
        self.finished = Signal[str]()
        self.failed = Signal[Tuple]()
        self.__thread.start()

    def __run(self) -> None:
        logger.info("Starting gcode download from printer {}".format(self.__printer.unique_name))

        gcode_dir = os.path.join(self.__gcode_spool_directory, self.__job_uuid)
        temp_filename = str(uuid.uuid4())
        self.__gcode_temp_filename = os.path.join(gcode_dir, temp_filename)

        created_dir = False
        try:
            if not os.path.isdir(gcode_dir):
                os.makedirs(gcode_dir, exist_ok=True)
                created_dir = True
            logger.info("Downloading gcode from printer {} to file {}"
                        .format(self.__printer.unique_name, self.__gcode_temp_filename))

            gcode_type = self.__printer.download_gcode(self.__gcode_temp_filename)
            if gcode_type is not None:
                logger.info("Successfully downloaded gcode from printer {} to file {}"
                            .format(self.__printer.unique_name, self.__gcode_temp_filename))

                # Move the temp download file to the correct destination filename with extension.
                filename_suffix = ".gcode" if gcode_type == GcodeType.PLAIN_GCODE else ".gcode.gz"
                gcode_filename = os.path.join(gcode_dir, self.__print_job_filename + filename_suffix)
                os.rename(self.__gcode_temp_filename, gcode_filename)

                self.finished.emit(gcode_filename)
                return

        except OSError:
            logger.exception("An error occurred")

        logger.warning("Unable to download the alien gcode from printer {}".format(self.__printer.unique_name))

        # Clean up any left over mess.
        if created_dir:
            def report_errors(_, path, __) -> None:
                logger.warning("An error occurred while cleaning up %s", path)
            shutil.rmtree(gcode_dir, ignore_errors=True, onerror=report_errors)

        self.failed.emit(())
