import logging
import time
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Tuple

from app.libs import watchdog
from app.libs.signal import Signal
from app.libs.ultimaker3.print_job import PrintJob as PrintJob
from app.libs.ultimaker3.status import Status
from app.models.printer import Printer
from um3_commons import thread

logger = logging.getLogger('PrintMonitorService')

POLL_SLEEP_SECONDS = 5

PrinterChangeSubscriber = Callable[[str, Status], None]


# The callable here has the signature: function(print_job_identifier: PrintJobIdentifier,
#                                               print_job_uuid: str, response_code: int,
#                                               json_response: Any)
PrintJobChangeSubscriber = Callable[[str, str, int, Any], None]


PrintJobIdentifier = NamedTuple("PrintJobIdentifier", (
    ("printer", Printer),
    ("print_job_uuid", str),
    ("external_uuid", str)
))


class PrinterMonitorService:
    def __init__(self, use_thread=True) -> None:
        # This code can be called from from multiple threads. Access to the member variables
        # should be done while holding this reentrant lock. Most of the variables are persistent
        # immutable structures, thus they can be happily used without worrying about changes
        # although it may be necessary to use the lock to grab a consistent snapshot of all
        # the variables.

        logger.info("Starting Printer Monitor Service...")

        self.__lock = thread.RLock()
        self.__printer_monitor_list = []  # type: List[Printer]
        self.__print_job_monitor_list = []  # type: List[PrintJobIdentifier]

        self.printer_status_updated = Signal[Tuple[Printer, Status]]()
        self.print_job_status_updated = Signal[Tuple[Printer, str, PrintJob]]()
        self.print_job_status_unavailable = Signal[Tuple[Printer, str]]()
        self.print_job_alien_detected = Signal[Tuple[Printer, PrintJob]]()

        if use_thread:
            self.__monitor_thread = thread.Thread(
                name="PrinterMonitorService",
                target=self._monitor_loop
            )
        else:
            self.__monitor_thread = None

    def start(self):
        logger.info("Starting printer monitor service.")
        if self.__monitor_thread is not None:
            self.__monitor_thread.start()

    def start_monitoring_printer(self, printer: Printer) -> None:
        with self.__lock:
            if printer not in self.__printer_monitor_list:
                self.__printer_monitor_list.append(printer)

    def stop_monitoring_printer(self, printer: Printer) -> None:
        with self.__lock:
            if printer in self.__printer_monitor_list:
                self.__printer_monitor_list.remove(printer)
            else:
                logger.warning("Tried to remove a printer that was not in the printer monitor list.")
            for printjob_identifier in self.__print_job_monitor_list:
                if printjob_identifier.printer == printer:
                    self.stop_monitoring_print_job(printjob_identifier.print_job_uuid)

    def start_monitoring_print_job(
            self,
            printer: Printer,
            print_job_uuid: str,
            external_uuid: str) -> None:

        logger.info("Start monitoring print job %s with external uuid %s" %
                     (print_job_uuid, external_uuid))
        with self.__lock:
            print_job_id = PrintJobIdentifier(printer=printer,
                                              print_job_uuid=print_job_uuid,
                                              external_uuid=external_uuid)
            self.__print_job_monitor_list.append(print_job_id)

    def stop_monitoring_print_job(self, print_job_uuid: str) -> None:
        with self.__lock:
            for job in self.__print_job_monitor_list:
                if job.print_job_uuid == print_job_uuid:
                    logger.info("Stop monitoring print job %s", print_job_uuid)
                    self.__print_job_monitor_list.remove(job)

    def _monitor_loop(self) -> None:
        watchdog.get_instance().start_monitoring()
        while True:
            self.fetch_status()
            watchdog.get_instance().tick()
            time.sleep(POLL_SLEEP_SECONDS)

    def fetch_status(self) -> None:
        watchdog.get_instance().tick()

        with self.__lock:
            # The whole poll action can take quite a long time but it is important that the lists it operates
            # on aren't modified while we are using them. So we don't lock everything but we do use copies.
            printer_monitor_list = self.__printer_monitor_list.copy()
            print_job_monitor_list = self.__print_job_monitor_list.copy()

        poll_executor = _PollExecutor(printer_monitor_list, print_job_monitor_list)
        poll_executor.execute()

        for printer_status_result in poll_executor.printer_status_results:
            self.printer_status_updated.emit(printer_status_result)

        for print_job_status_updated_result in poll_executor.print_job_status_updated_results:
            self.print_job_status_updated.emit(print_job_status_updated_result)

        for print_job_status_unavailable_result in poll_executor.print_job_status_unavailable_results:
            self.print_job_status_unavailable.emit(print_job_status_unavailable_result)

        for print_job_alien_detected_result in poll_executor.print_job_alien_detected_results:
            self.print_job_alien_detected.emit(print_job_alien_detected_result)


class _PollExecutor:
    def __init__(self, printer_monitor_list: Iterable[Printer], print_job_monitor_list: Iterable[PrintJobIdentifier])\
            -> None:
        self.__printer_monitor_list = printer_monitor_list
        self.__print_job_monitor_list = print_job_monitor_list

        self.__printer_current_job_cache = {}               # type: Dict[str, Optional[PrintJob]]
        self.__seen_printer_jobs = []                       # type: List[Tuple[Printer, Optional[PrintJob]]]

        self.__printer_status_results = []                  # type: List[Tuple[Printer, Status]]
        self.__print_job_status_updated_results = []        # type: List[Tuple[Printer, str, PrintJob]]
        self.__print_job_status_unavailable_results = []    # type: List[Tuple[Printer, str]]
        self.__print_job_alien_detected_results = []        # type: List[Tuple[Printer, PrintJob]]

    @property
    def printer_status_results(self) -> List[Tuple[Printer, Status]]:
        return self.__printer_status_results

    @property
    def print_job_status_updated_results(self) -> List[Tuple[Printer, str, PrintJob]]:
        return self.__print_job_status_updated_results

    @property
    def print_job_status_unavailable_results(self) -> List[Tuple[Printer, str]]:
        return self.__print_job_status_unavailable_results

    @property
    def print_job_alien_detected_results(self) -> List[Tuple[Printer, PrintJob]]:
        return self.__print_job_alien_detected_results

    def execute(self) -> None:
        self.__poll_printers()
        self.__poll_print_jobs()
        self.__detect_alien_print_jobs()

    def __poll_printers(self) -> None:
        # We send out the notifications of the jobs before we do the status of the printer itself.
        # This avoids the inconsistency where the user of this class can think that the printer
        # is idle but their list of print jobs says that one is still running on that printer.
        for printer in self.__printer_monitor_list:
            try:
                self.__printer_status_results.append((printer, printer.get_status()))
                job = self.__get_current_job(printer)
                if job is not None:
                    self.__seen_printer_jobs.append((printer, job))
            except:
                logger.exception("printer.get_status raised an exception.")
            watchdog.get_instance().tick()

    def __poll_print_jobs(self) -> None:
        for print_job_identifier in self.__print_job_monitor_list:
            self.__refresh_job_status(print_job_identifier)
            watchdog.get_instance().tick()

    def __get_current_job(self, printer: Printer) -> PrintJob:
        if printer.uuid in self.__printer_current_job_cache:
            return self.__printer_current_job_cache[printer.uuid]
        job = printer.get_current_print_job()
        self.__printer_current_job_cache[printer.uuid] = job
        return job

    def __refresh_job_status(self, print_job_identifier: PrintJobIdentifier) -> None:
        print_job_status = self.__fetch_print_job_status(print_job_identifier)
        print_job_uuid = print_job_identifier.print_job_uuid

        if print_job_status is not None:
            self.__print_job_status_updated_results.append(
                (print_job_identifier.printer, print_job_uuid, print_job_status))
        else:
            self.__print_job_status_unavailable_results.append((print_job_identifier.printer, print_job_uuid))

    def __fetch_print_job_status(self, print_job_identifier: PrintJobIdentifier) -> Optional[PrintJob]:
        external_job_uuid = print_job_identifier.external_uuid
        printer = print_job_identifier.printer
        current_print_job_status = self.__get_current_job(printer)
        if current_print_job_status is not None:
            if current_print_job_status.uuid == external_job_uuid:
                return current_print_job_status

        historical_print_job_status = printer.get_print_job_from_history(external_job_uuid)
        return historical_print_job_status

    def __detect_alien_print_jobs(self) -> None:
        # An alien is simply a print job we saw on a printer which we can't match with an monitored print job.
        known_print_jobs = set((id(tup[2]) for tup in self.__print_job_status_updated_results))

        for (printer, print_job) in self.__seen_printer_jobs:
            if id(print_job) not in known_print_jobs:
                self.__print_job_alien_detected_results.append((printer, print_job))
