import logging
from typing import Any, Dict, Iterable, Optional, Tuple, cast

from app.constants.printer_status import PrinterStatus
from app.libs import watchdog
from app.libs.printer_credentials_store import PrinterCredentialsStore
from app.libs.signal import Signal
from app.libs.ultimaker3.print_job import PrintJob as UM3PrintJob
from app.libs.ultimaker3.status import Status
from app.models.abstraction.abstract_model import AbstractModel
from app.models.configuration import Configuration
from app.models.extruder_train import ExtruderTrain
from app.models.print_job import PrintJob
from app.models.printer import Printer
from app.models.printer_collection import PrinterCollection
from app.services.material_manager import MaterialManager
from app.services.printer_monitor_service import PrinterMonitorService
from um3_commons import thread

logger = logging.getLogger(__name__)

try:
    from um3_commons import dbusif
    import dbus
except Exception:
    dbusif = None
    logger.warning("DBUS not available!")


class PrinterService:
    """
    Printer service.

    Responsibilities:
    - Status and status changes (manual or automatic) of printers
    - LEDS of all connected printers
    - ...
    """
    def __init__(self,
                 collection: PrinterCollection,
                 printer_monitor_service: PrinterMonitorService,
                 material_manager: MaterialManager,
                 application_started_signal: Signal[Tuple],
                 leds_start_active: bool=False) -> None:
        self.__printers = collection  # type: PrinterCollection
        self._printer_monitor_service = printer_monitor_service  # type: PrinterMonitorService
        self._material_manager = material_manager  # type: MaterialManager

        self._printer_monitor_service.printer_status_updated.connect(self.__printer_status_handler)
        self._printer_monitor_service.print_job_status_updated.connect(
            self.__handle_print_job_status_updated)

        application_started_signal.connect(self.__handle_application_started)
        self.__printers.onAdd.connect(self.__handle_printer_added)
        self.__printers.onDelete.connect(self.__handle_printer_removed)

        self.__lock = thread.RLock()
        self._system_dbus = None                # type: Optional[dbusif.RemoteObject]
        self.__set_dbus_system()

        self.printer_status_changed = Signal[Printer]()
        self.printer_is_about_to_be_removed = Signal[Printer]()
        self.printer_enabled_changed = Signal[Printer]()

        self._leds_active = leds_start_active  # We set the leds to active after application start

        self.amount_of_printers_changed = Signal[int]()

    @property
    def printers(self) -> PrinterCollection:
        return self.__printers

    def __handle_printer_status_changed(self, signal_payload: Tuple[Printer, Dict[str, Any]]) -> None:
        printer, properties = signal_payload
        if printer is not None and "status" in properties.keys():
            logger.debug("Printer %s status change: %s", printer, printer.status)

    def __handle_application_started(self, payload: Tuple) -> None:
        # Clear the status field of all the printers because we don't know what the current
        # status is
        for printer in self.__printers.all():
            printer.status = None

        self._leds_active = True
        self.update_all_leds()
        self.__start_monitoring_printers(self.__printers.all())
        self.__check_and_add_self()

    def __start_monitoring_printers(self, printer_list: Iterable[Printer]) -> None:
        for printer in printer_list:
            self._printer_monitor_service.start_monitoring_printer(printer)

    def __printer_status_handler(self, signal_payload: Tuple[Printer, Status]) -> None:
        printer, printer_status = signal_payload
        changed = False
        if printer.is_authenticated():
            if self.__printer_status_update_working(printer, printer_status):
                changed = True
            if self.__printer_status_update_materials_and_extruders(printer, printer_status):
                changed = True
            if changed:
                self.printer_status_changed.emit(printer)
                self.update_leds(printer)
        else:
            # If the printer has no valid credentials, we remove it from the list.
            # It would be better to implement a notification towards the user, but we do not have the notification
            # system at this time.
            # Keeping it in the list with a different status would also add an extra source for bugs.
            logger.warning("Removing printer: [%s] because of authentication no longer being valid." % (printer))
            self.__printers.delete(printer)

    def __printer_status_update_working(self, printer: Printer, printer_status: Status) -> bool:
        """
        Return true if anything has changed
        """
        if printer_status is None:
            status = PrinterStatus.UNREACHABLE
        else:
            status = printer_status.status

        if printer.status != status:
            printer.status = status
            return True
        return False

    def __printer_status_update_materials_and_extruders(self, printer: Printer, printer_status: Status) -> bool:
        """
        Return if anything has changed
        """
        if printer_status is None:
            return False

        configuration = Configuration()

        for head in printer_status.heads:
            for extruder_index, extruder_train in enumerate(head.extruders):
                extruder = ExtruderTrain(
                    extruder_index=extruder_index,
                    material_guid=extruder_train.material.guid,
                    nozzle_diameter=extruder_train.hotend_diameter,
                    print_core_id=extruder_train.hotend_id
                )
                extruder.material = self._material_manager.get(extruder_train.material.guid)

                configuration.update(extruder)

        # update buildplate information
        printer.build_plate = printer_status.build_plate

        for extruder_index in range(printer.machine_variant.extruder_count):
            if printer.configuration.get(extruder_index) != configuration.get(extruder_index):
                printer.configuration = configuration
                logger.debug("Status changed: %s -> %s", printer, printer_status)
                return True
        return False

    def __handle_printer_added(self, printer: Printer) -> None:
        self._printer_monitor_service.start_monitoring_printer(printer)
        self.amount_of_printers_changed.emit(len(self.__printers.all()))

        printer.onChange.connect(self.__update_printer_enabled_status)

    def __handle_printer_removed(self, printer: Printer) -> None:
        logger.debug("Remove printer: %s", printer)

        self.printer_is_about_to_be_removed.emit(printer)
        self.remove_reservation(printer)
        self._printer_monitor_service.stop_monitoring_printer(printer)
        self.amount_of_printers_changed.emit(len(self.__printers.all()))

    def set_reservation(self, printer: Printer, print_job: Optional[PrintJob], user_message: str) -> None:
        # Reservations are set/removed from the scheduler.
        # Nothing else should touch reservations, except for printer/job deletion.
        # TODO: when refactoring models, this should become a responsibility of the model (except the leds?)
        if printer.reserved_by_job:
            self.remove_reservation(printer)
        if print_job.assigned_to_printer:
            self.remove_reservation(print_job.assigned_to_printer)
        printer.set_reserved_by(print_job)
        print_job.set_assigned_to(printer)
        self.update_leds(printer)
        printer.set_display_message(user_message, "OK")

    def remove_reservation(self, printer: Printer) -> None:
        # Reservations are set/removed from the scheduler.
        # Nothing else should touch reservations, except for printer/job deletion.
        # TODO: when refactoring models, this should become a responsibility of the model (except the leds?)
        if printer.reserved_by is None:
            return
        printer.reserved_by_job.set_assigned_to(None)
        printer.set_reserved_by(None)
        self.update_leds(printer)
        printer.clear_display_message()

    @staticmethod
    def _compare_printer_properties_to_dict(printer: Printer, properties: Dict[str, Any]) -> bool:
        for key, value in properties.items():
            existing_value = getattr(printer, key)
            if existing_value != value:
                return True
        return False

    @staticmethod
    def __has_changed(key: str, properties: Dict[str, Any], value: Any) -> bool:
        return (key in properties) and properties[key] != value

    def get(self, printer_uuid: str) -> Optional[Printer]:
        printers = self.__printers.all()
        for printer in printers:
            if printer.uuid == printer_uuid:
                return printer
        return None

    def find_by_ip_address(self, ip_address: str) -> Optional[Printer]:
        for printer in self.__printers.all():
            if printer.ip_address == ip_address:
                return printer
        return None

    def find_by_unique_name(self, name: str) -> Optional[Printer]:
        for printer in self.__printers.all():
            if printer.unique_name == name:
                return printer
        return None

    def __update_printer_enabled_status(self, signal_payload: Tuple[AbstractModel, Dict[str, Any]]) -> None:
        printer = cast(Printer, signal_payload[0])
        properties = signal_payload[1]

        if properties.get("enabled") is not None:
            self.printer_enabled_changed.emit(printer)
            self.update_leds(printer)

    def update_leds(self, printer: Printer) -> None:
        """Update printer leds."""
        if not self._leds_active:   # led updates are not yet active when the system is starting up.
            return
        printer.set_default_led_color()

    def update_all_leds(self) -> None:
        for printer in self.__printers.all():
            self.update_leds(printer)

    def __handle_print_job_status_updated(self, signal_payload: Tuple[Printer, str, UM3PrintJob]) -> None:
        printer, print_job_uuid, print_job_status = signal_payload
        self.update_leds(printer=printer)

    def __set_dbus_system(self) -> None:
        """
        Set self._dbus_system
        """
        # For computers without dbus
        if dbusif is None:
            return
        try:
            self._system_dbus = dbusif.RemoteObject("system")
        except dbus.DBusException:
            logger.warning("no dbus connection to local printer!")
            return

    @property
    def system_dbus(self) -> "Optional[dbusif.RemoteObject]":
        return self._system_dbus

    def __check_and_add_self(self) -> None:
        # For computers without dbus
        if self._system_dbus is None:
            return

        # this cluster has no self so add it!
        try:
            auth = dbusif.RemoteObject("interface/http", "interface/http/auth")
            if auth.hasExclusiveAuth("cluster"):
                logger.info("not adding self to cluster because there is already a cluster that has added this printer")
                return

            dog = watchdog.get_instance()
            dog.delay_monitoring_by(120)
            # somehow this can take about a minute
            unique_name = self._system_dbus.getHostName()
            dog.tick()

            if self.find_by_unique_name(unique_name) is None:  # unique name is safer then ip address
                own_ip_addr = dbusif.RemoteObject("network").getIPv4Address()
                if own_ip_addr is None or own_ip_addr == "":
                    own_ip_addr = "127.0.0.1"
                printer_name = self._system_dbus.getMachineName()
                credentials_store = PrinterCredentialsStore()
                credentials_store.id, credentials_store.key = auth.addNewAuthorization("Cura connect", "cluster", unique_name)
                printer = Printer(
                    ip_address=own_ip_addr,
                    unique_name=unique_name,
                    machine_variant=dbusif.RemoteObject("printer").getProperty("machine_type_name"),
                    friendly_name=printer_name,
                    digest_auth_id=credentials_store.id,
                    digest_auth_key=credentials_store.key,
                    digest_auth_authorized=True
                )
                self.__printers.add(printer)
                logger.info("added self to cluster %s", printer)
        except Exception:
            logger.exception("Failed to add self to cluster!")
