from typing import Optional, Tuple

from app.services.scheduler.abstract_planner import AbstractPlanner
from app.services.scheduler.result_containers.match_results import ConfigurationChangeMatchResult, MatchResult, Severity
from app.services.scheduler.result_containers.match_results import IncompatibleMatchResult
from app.services.scheduler.result_containers.match_results_containers import MatchMatrix, MatchResultListPerPrinter
from app.services.scheduler.result_containers.planner_result_schedule import PlannerResultSchedule


# Plan jobs first come first serve, for each job in the list find a printer that fits this job.
# Never reserve a printer, only send jobs to printers with an exact match.
class FirstComeFirstServePlanner(AbstractPlanner):
    def __init__(self) -> None:
        # This sets the Severity translation function in the base class of the match results, This is only correct as
        # long as we only have one planner active at one time.
        # This was done because the matchresults have use/value outside of the planner and are made outside of the
        # planners control, yet the severity of the match result can only be valued by the planner.
        MatchResult.severity_to_score_translation_function = self.__severity_to_score_translation

    def calculate_plan(self, match_matrix: MatchMatrix) -> PlannerResultSchedule:
        result = PlannerResultSchedule()  # the result of planning is a schedule that you then try to make/keep.

        for job_match in match_matrix.get_match_result_per_job():
            printer_matches = sorted(job_match, key=lambda printer_match: self.__printer_match_score(result, printer_match))

            best_match = None  # type: Optional[MatchResultListPerPrinter]
            if len(printer_matches) > 0:
                best_match = printer_matches[0]
                if best_match.has(IncompatibleMatchResult):
                    best_match = None
                if best_match.has(ConfigurationChangeMatchResult):
                    best_match = None

            # A bit of a hack, if our selected printer is the reserved one, do not switch to any other printer.
            # Unless the best match we picked can directly accept our job.
            if job_match.print_job.assigned_to_printer:
                if best_match is not None and len(best_match) > 0:
                    best_match = job_match.get_results_for(job_match.print_job.assigned_to_printer)

            if best_match is not None:
                # Add the best match to the plan result.
                # Multiple jobs can be added to the same printer, only the first job will be started or reserve
                # the printer if the printer is free.
                # Other jobs are just added to give an indication on if the job can be printed.
                result.add_to_schedule(best_match.printer, job_match.print_job)

        return result

    @staticmethod
    def __severity_to_score_translation(severity: Severity) -> int:
        if severity == Severity.INFO:
            return 0
        elif severity == Severity.FUTURE:
            return 4
        elif severity == Severity.CONFIGURATION_CHANGE:
            return 42
        elif severity == Severity.USER_FIXABLE:
            return 42
        elif severity == Severity.UNFIXABLE:
            return 100
        elif severity == Severity.RESERVATION:
            return 100
        return 100

    @staticmethod
    def __printer_match_score(result_plan: PlannerResultSchedule, printer_match: MatchResultListPerPrinter)\
            -> Tuple[int, int, int, int]:
        return (
            # Put all incompattible printers at the end of the list
            printer_match.count(IncompatibleMatchResult),
            # Next, any printer that has configuration changes required, should at the back of the list as well.
            printer_match.count(ConfigurationChangeMatchResult),
            # Next we want free printers to be at the front.
            result_plan.get_schedule_count(printer_match.printer),
            # Of the free printers, find the printers that are actually free to print on right now
            printer_match.score,
        )
