import logging

from app.constants.print_job_status import PrintJobStatus
from app.constants.printer_status import PrinterStatus
from app.libs import watchdog
from app.models.print_job import PrintJob
from app.models.printer import Printer
from app.services.print_job_service import PrintJobService
from um3_commons.thread import Thread

logger = logging.getLogger(__name__)


# This class is responsible for sending jobs to the printer, it does so in it's own thread.
# Use the join after dispatching and then read the result via was_successful.
class PrintJobDispatcher:
    def __init__(self, print_job_service: PrintJobService, print_job: PrintJob, printer: Printer) -> None:
        self.__print_job_service = print_job_service
        self.__print_job = print_job  # type: PrintJob
        self.__printer = printer  # type: Printer
        self.__thread = Thread(
            name="printjob_dispatcher_{}".format(print_job.uuid),
            target=self.__run
        )

        self.__thread.start()
        self.__failed = False

    def join(self) -> None:
        self.__thread.join()

    def __run(self) -> None:
        watchdog.get_instance().start_monitoring()
        try:
            external_uuid = self.__printer.create_job(self.__print_job.job_name, self.__print_job.filepath)
            if external_uuid is None:
                self.__failed = True
                logger.error("Failed to create job... OH NOES, worlds collide and brains explode.")
                return
            logger.info(
                "Finished sending job %s to printer %s, external uuid %s",
                self.job,
                self.printer,
                external_uuid
            )

            with self.__print_job.atomic_modification:
                self.__print_job.printer_uuid = self.printer.uuid
                self.__print_job.external_uuid = external_uuid
                self.__print_job.status = PrintJobStatus.SENT_TO_PRINTER

            logger.info("print %s started on printer %s", self.__print_job, self.__printer)

            # As we send the job to the printer, it is most likely no longer idle.
            # Ensure we do not handle this printer as idle, so put the status to unknown.
            self.__printer.status = PrinterStatus.PRINTING
        except Exception:
            self.__failed = True
            logger.exception("Failed to create job... OH NOES, worlds collide and brains explode.")
        watchdog.get_instance().stop_monitoring()

    @property
    def was_successful(self) -> bool:
        assert not self.__thread.is_alive(), "end state checked before end."
        return not self.__failed

    @property
    def job(self) -> PrintJob:
        return self.__print_job

    @property
    def printer(self) -> Printer:
        return self.__printer
