from enum import IntEnum
from typing import Optional, Callable

import copy

from app.services.scheduler.match_rule import MatchRule


class MatchResultType(IntEnum):
    INCOMPATIBLE = 0
    FUTURE = 1
    CONFIGURATION_CHANGE = 2


class Severity(IntEnum):
    INFO = 0
    CONFIGURATION_CHANGE = 1
    RESERVATION = 2
    FUTURE = 3
    USER_FIXABLE = 4
    UNFIXABLE = 5

    def __str__(self):
        return self.name


class MatchResult:
    # This is a severity type to score translation Function, the actual translation function must be set by the planner
    # when it's constructed.
    # This was done because the matchresults have use/value outside of the planner and are made outside of the
    # planners control, yet the severity of the match result can only be valued by the planner.
    severity_to_score_translation_function = lambda severity: int(severity)  # type: Callable[[Severity], int]

    def __init__(self,
                 origin: Optional[MatchRule],
                 debug_message: str,
                 translation_key: str,
                 severity: Severity=Severity.UNFIXABLE) -> None:
        self.__origin = origin
        self.__debug_message = debug_message
        self.__translation_key = translation_key
        self.__severity = severity

    @property
    def score(self) -> int:
        return MatchResult.severity_to_score_translation_function(self.__severity)

    @property
    def type(self) -> str:
        raise NotImplementedError()

    @property
    def origin(self) -> Optional[MatchRule]:
        return self.__origin

    @property
    def severity(self) -> Severity:
        return self.__severity

    @property
    def translation_key(self):
        """
        This is a key for the front end to translate this MatchResult which is a impediment to printing to the user.
        These keys are all described in:
        https://docs.google.com/spreadsheets/d/1y_f_67Xw_7aPPDa6RuDXjmZM1jbqUlC5jnk24xj56CA/edit?usp=sharing

        Returns: A string with no spaces, only word characters and underscores.

        """
        return self.__translation_key

    @property
    def debug_message(self) -> str:
        """
        This message is purely meant for debug purposes and should never be exposed to users.
        Returns: A debug string

        """
        return self.__debug_message

    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__, self.debug_message)


class IncompatibleMatchResult(MatchResult):
    def __init__(
            self,
            *,
            origin: MatchRule,
            debug_message: str,
            translation_key: str,
            severity: Severity=Severity.UNFIXABLE) -> None:
        super().__init__(origin, debug_message, translation_key, severity)
        self.__translation_key = translation_key

    @property
    def type(self) -> str:
        return str(MatchResultType.INCOMPATIBLE)


class FutureMatchResult(MatchResult):
    def __init__(
            self,
            *,
            time: float,
            origin: MatchRule,
            debug_message: str,
            translation_key: str,
            severity: Severity=Severity.FUTURE) -> None:
        super().__init__(origin, debug_message, translation_key, severity)
        self.__time = time

    @property
    def type(self) -> str:
        return str(MatchResultType.FUTURE)

    @property
    def time(self):
        return self.__time

    @property
    def debug_message(self):
        return super().debug_message + " until: " + str(self.__time)


class ConfigurationChangeMatchResult(MatchResult):
    class ChangeSet:
        def __init__(
                self,
                *,
                type_of_change: str,
                origin_id: str,
                target_id: str,
                index: Optional[int]=None) -> None:
            assert type_of_change in ["material_change", "material_insert", "print_core_change", "buildplate_change"]
            self.type_of_change = type_of_change
            self.origin_id = origin_id
            self.target_id = target_id
            self.origin_name = None    # type: Optional[str]
            self.target_name = None    # type: Optional[str]
            if index is not None:
                self.index = index

    def __init__(
            self,
            *,
            change: Optional[ChangeSet],
            origin: Optional[MatchRule],
            debug_message: str,
            translation_key: str) -> None:
        super().__init__(origin, debug_message, translation_key, Severity.CONFIGURATION_CHANGE)
        self.__change = change

    @property
    def type(self) -> str:
        return str(MatchResultType.CONFIGURATION_CHANGE)

    @property
    def change(self) -> Optional[ChangeSet]:
        if self.__change is None:
            return None
        return copy.copy(self.__change)


    @property
    def debug_message(self) -> str:
        return super().debug_message + " " + str(self.__change)




