from typing import Optional, Type, List, Iterable, Iterator, Sized

from app.models.print_job import PrintJob
from app.models.printer import Printer
from .match_results import MatchResult


# These classes are all containers for the match matrix build by the match matrix builder.


class MatchResultListPerPrinter(Sized):
    def __init__(self, printer: Printer) -> None:
        self.__printer = printer  # type: Printer
        self.__result_list = []  # type: List[MatchResult]

    @property
    def printer(self) -> Printer:
        return self.__printer

    @property
    def results(self) -> List[MatchResult]:
        return self.__result_list.copy()

    def append(self, result: Optional[MatchResult]) -> None:
        if result is None:
            return
        assert isinstance(result, MatchResult)
        self.__result_list.append(result)

    def count(self, match_type: Type[MatchResult]=None) -> int:
        if match_type:
            return len([n for n in self.__result_list if isinstance(n, match_type)])
        return len(self.__result_list)

    def __len__(self):
        return len(self.__result_list)

    def has(self, match_type: Type[MatchResult]) -> bool:
        if match_type is not None:
            return any(isinstance(n, match_type) for n in self.__result_list)
        return match_type in self.__result_list

    @property
    def score(self) -> int:
        score = 0  # type: int
        for match_result in self.__result_list:
            score += match_result.score

        return score

    def __repr__(self) -> str:
        return "<MatchResultListPerPrinter printer_name='{}' uuid='{}' results='{}'>".format(
            self.__printer.friendly_name, self.__printer.uuid, self.__result_list)


class MatchResultListPerJob(Iterable[MatchResultListPerPrinter]):
    def __init__(self, print_job: PrintJob) -> None:
        self.__print_job = print_job  # type: PrintJob
        self.__result_list = []  # type: List[MatchResultListPerPrinter]

    @property
    def print_job(self) -> PrintJob:
        return self.__print_job

    @property
    def results(self) -> List[MatchResultListPerPrinter]:
        return self.__result_list.copy()

    def append(self, result_per_printer) -> None:
        self.__result_list.append(result_per_printer)

    def count(self, match_type: Type[MatchResult]=None) -> int:
        total = 0  # type: int
        for result in self.__result_list:
            total += result.count(match_type)
        return total

    def has(self, match_type: Type[MatchResult]) -> bool:
        for result in self.__result_list:
            if result.has(match_type):
                return True
        return False

    def get_results_for(self, printer: Printer) -> Optional[MatchResultListPerPrinter]:
        for result in self.__result_list:
            if result.printer == printer:
                return result
        return None

    def __iter__(self) -> Iterator[MatchResultListPerPrinter]:
        return iter(self.__result_list)

    def __repr__(self) -> str:
        return "<MatchResultListPerJob job_name='{}' uuid='{}' results='{}'>".format(
            self.__print_job.name, self.__print_job.uuid, self.__result_list)


class MatchMatrix:
    def __init__(self) -> None:
        self.__results = []  # type: List[MatchResultListPerJob]

    def append(self, per_job_result_list: MatchResultListPerJob) -> None:
        self.__results.append(per_job_result_list)

    def get_match_result_per_job(self) -> List[MatchResultListPerJob]:
        return self.__results.copy()

    def get_match_result(self, printer: Printer, print_job: PrintJob) -> Optional[MatchResultListPerPrinter]:
        for result in self.__results:
            if result.print_job == print_job:
                return result.get_results_for(printer)
        return None

    def __repr__(self) -> str:
        return "<MatchMatrix results='{}'>".format(self.__results)
