import logging
from collections import deque
from typing import Optional

from app.services.material_manager import MaterialManager
from app.services.print_job_service import PrintJobService
from app.services.printer_service import PrinterService
from app.services.scheduler.result_containers.planner_result_schedule import PlannerResultSchedule
from app.services.scheduler.match_matrix_builder import MatchMatrix, MatchMatrixBuilder
from app.services.scheduler.planner.first_come_first_serve_planner_with_reservations import FirstComeFirstServePlannerWithReservations
from app.services.scheduler import rules    # mypy doesn't support wildcard * imports.
from app.services.scheduler.schedule_executor import ScheduleExecutor

logger = logging.getLogger(__name__)
MYPY = False
if MYPY:
    from typing import Deque


# This class collects all the parts of the scheduler and performs the matching, planning
# and executing of the schedule in order
class PrintJobScheduler:
    MAX_HISTORY_LENGTH = 5

    def __init__(self,
                 printer_service: PrinterService,
                 print_job_service: PrintJobService,
                 material_manager: MaterialManager) -> None:
        self._printer_service = printer_service
        self._print_job_service = print_job_service

        self.__match_matrix_builder = MatchMatrixBuilder([
            rules.FitsInBuildVolumeMatchRule(),
            rules.IsPrinterReachable(),
            rules.DoesPrinterMatchTargetFamilyRule(),
            rules.IsAuthenticatedMatchRule(),
            rules.MatchesConstrainedPrinterNameRule(),
            rules.PrinterIsMissingPrintCoresMatchRule(),
            rules.HasNotFailedOnPrinterMatchRule(),
            rules.OverruleConfigurationMismatch(rules.IsHotendCartridgeCompatibleMatchRule(0)),
            rules.OverruleConfigurationMismatch(rules.IsHotendCartridgeCompatibleMatchRule(1)),

            rules.RuleStack(
                rules.IsMaterialPresentMatchRule(0),
                rules.OverruleConfigurationMismatch(rules.IsMaterialCompatibleMatchRule(0))
            ),
            rules.RuleStack(
                rules.IsMaterialPresentMatchRule(1),
                rules.OverruleConfigurationMismatch(rules.IsMaterialCompatibleMatchRule(1))
            ),
            # The IsMaterialPresentMatchRule above are very similar to the `IsMaterialCompatibleMatchRule`s above. They are needed to
            # enforce the constraint that even when a print job is being forced on to a printer we still
            # prevent the job from trying to use a PrintCore which doesn't have *any* material loaded.

            rules.IsBuildplateCompatibleMatchRule(),
            rules.IsPrinterDisabledMatchRule(),
            rules.IsPrinterInMaintenanceMatchRule(),
            rules.IsPrinterPrintingMatchRule(),
            rules.IsPrinterReservedMatchRule(),
            rules.IsPrinterInPrePrintMatchRule(),
            rules.IsPrinterInUnknownStateMatchRule()
        ])

        self.__planner = FirstComeFirstServePlannerWithReservations()
        self.__executor = ScheduleExecutor(self._printer_service, self._print_job_service, material_manager)

        self.__debug_match_matrix_history = deque([], maxlen=self.MAX_HISTORY_LENGTH)  # type: Deque[MatchMatrix]
        self.__debug_schedule_history = deque([], maxlen=self.MAX_HISTORY_LENGTH)  # type: Deque[PlannerResultSchedule]
        self.__current_match_matrix = None  # type: Optional[MatchMatrix]
        self.__current_schedule = None  # type: Optional[PlannerResultSchedule]

    def run(self) -> bool:
        self.__current_match_matrix = self.__match_matrix_builder.build(
            print_jobs=self._print_job_service.print_jobs.queued(),
            printers=self._printer_service.printers.all()
        )

        self.__current_schedule = self.__planner.calculate_plan(self.__current_match_matrix)
        logger.debug("Going to execute plan: %s", self.__current_schedule)

        self.__executor.execute_plan_result(
            schedule=self.__current_schedule,
            match_matrix=self.__current_match_matrix
        )

        if self.__executor.has_started_jobs() or len(self.__debug_match_matrix_history) < 1:
            self.__debug_match_matrix_history.append(self.__current_match_matrix)
            self.__debug_schedule_history.append(self.__current_schedule)
        return self.__executor.requires_reschedule()

    def debug_get_current_match_matrix(self) -> Optional[MatchMatrix]:
        return self.__current_match_matrix

    def debug_get_current_schedule(self) -> Optional[PlannerResultSchedule]:
        return self.__current_schedule

    def debug_get_match_matrix_history(self, index: int=-1) -> Optional[MatchMatrix]:
        try:
            return self.__debug_match_matrix_history[index]
        except IndexError:
            return None

    def debug_get_schedule_history(self, index: int=-1) -> Optional[PlannerResultSchedule]:
        try:
            return self.__debug_schedule_history[index]
        except IndexError:
            return None

    @property
    def debug_history_maxlen(self) -> int:
        return self.__debug_match_matrix_history.maxlen

    def clear_reservations(self) -> None:
        self.__executor.clear_reservations(self._printer_service.printers.all())
