import logging
from collections import OrderedDict
from threading import RLock
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

from app.events.base_event import BaseEvent
from app.libs.signal import Signal
from app.models.abstraction.abstract_collection_model import AbstractCollectionModel
from app.models.abstraction.abstract_model import AbstractModel
from app.models.print_job import PrintJobCreationError
from app.storage.event_store import EventStore

logger = logging.getLogger(__name__)
M = TypeVar('M', covariant=True, bound=AbstractModel)   # Used for specifying the model's type for the manager.


# this class hangs between the event_store and your models, it listens to events and builds up A DB of events it
# (after replay) passes to a Abstact Collection Model, once it's done that it disconects itself from listening to
# the event store but it connects itself to the model's signals These are used to create new events and store new
# changes from then on. This is a base class and the actual events used are supplied by the subclasses such as the
# PrinterEventStorageManager
class AbstractEventStorageManager(Generic[M]):
    def __init__(
            self,
            *,
            event_store: EventStore,
            replay_done: Signal,
            collection: AbstractCollectionModel[M],
            add_event: Type[BaseEvent],
            modify_events: List[Type[BaseEvent]],
            delete_event: Type[BaseEvent],
            model: Type[M],
            move_event: Type[BaseEvent]=None
    ) -> None:
        self._db = OrderedDict()  # type: OrderedDict[str, M]
        self.__event_store = event_store
        self.__collection = collection
        self.__add_event = add_event  # type: Type[BaseEvent]
        self.__move_event = move_event  # type: Optional[Type[BaseEvent]]
        self.__modify_event = modify_events[0]  # type: Type[BaseEvent]
        self.__delete_event = delete_event  # type: Type[BaseEvent]
        self.__model = model  # type: Type[M]
        self.__lock = RLock()
        self.__loading = True

        event_store.subscribe_to(add_event, self.__handle_add_event)
        event_store.subscribe_to(delete_event, self.__handle_delete_event)
        for event in modify_events:
            event_store.subscribe_to(event, self.__handle_modify_event)
        if move_event is not None:
            event_store.subscribe_to(move_event, self.__handle_move_event)

        replay_done.connect(self.__handle_end_of_replay)

    #######################################################
    # Events From EventStore Handing (Only during replay) #
    #######################################################

    def __handle_add_event(self, is_replay: bool, event: BaseEvent) -> None:
        assert self.__loading, "The replay of events should only happen during the loading stage."
        properties = {"uuid": event.uuid}
        if event.timestamp is not None:
            properties.update({"created_at": event.timestamp})
        properties.update(event.to_dict())
        with self.__lock:
            try:
                self._db[event.uuid] = self.__model(**properties)   # type: ignore
                                                                    # Model constructors have many different signatures.
                                                                    # We can't statically check them.
            except PrintJobCreationError:
                logger.warning("Ignored new print_job %s [%s] because the gcode was not there!", event.uuid, properties)

    def __handle_delete_event(self, is_replay: bool, event: BaseEvent) -> None:
        assert self.__loading, "The replay of events should only happen during the loading stage."
        with self.__lock:
            try:
                del self._db[event.uuid]
            except KeyError:
                logger.warning("uuid %s was not found to be deleted in %s", event.uuid, self)

    def __handle_move_event(self, is_replay: bool, event: BaseEvent) -> None:
        assert self.__loading, "The replay of events should only happen during the loading stage."
        new_position = event.kwargs["to_position"]
        if event.uuid not in self._db:
            logger.debug("Ignored move item, the item does not exist. [%s]", event.uuid)
            return
        # We're only supporting position 0 for now
        if new_position != 0:
            logger.warning("Not supported: move print job [%s] to position [%s]", event.uuid, new_position)
            return
        logger.info("Move print job [%s] to position [%s]", event.uuid, new_position)

        with self.__lock:
            self._db.move_to_end(event.uuid, last=False)

    def __handle_modify_event(self, is_replay: bool, event: BaseEvent) -> None:
        assert self.__loading, "The replay of events should only happen during the loading stage."
        with self.__lock:
            item = self._db.get(event.uuid)
            if item is not None:
                with item.atomic_modification:  # should not be necessary as this should not generate events,
                    # but might speed up the process, less unconnected signals, etc
                    for attr, value in event.to_dict().items():
                        item._set_via_ModelField(attr, value)

    ######################################################
    # Handover from event_store control to Model control #
    ######################################################

    def __handle_end_of_replay(self, _: Tuple) -> None:
        with self.__lock:
            # Start to establish links to write back to event_store (have to be connected before add
            # so all models created can be observed
            self.__collection.onAdd.connect(self.__handle_addition_to_list)

            for key, value in self._db.items():
                self.__collection.add(value)

            # Disentangle normal listening to event_store (future events are to be stored, not processed)
            for handler in [self.__handle_add_event,
                            self.__handle_delete_event,
                            self.__handle_modify_event,
                            self.__handle_move_event]:
                self.__event_store.unsubscribe(handler)

            # This has to happen after the
            self.__loading = False

            # complete establishment of links with the object models.
            self.__collection.onDelete.connect(self.__handle_deletion_from_list)
            self.__collection.onMove.connect(self.__handle_move_in_list)
            self._db = None  # clear temporary data build up during event_store replay.

    ##############################
    # Signal From Models Handing #
    ##############################

    def __handle_addition_to_list(self, object_model: AbstractModel) -> None:
        with self.__lock:
            if not self.__loading:
                self.__event_store.append(self.__add_event(**object_model.to_dict()))
            object_model.onPreChange.connect(self.__handle_model_change)

    def __handle_model_change(self, signal_payload: Tuple[AbstractModel, Dict[str, Any]]) -> None:
        object_model, properties = signal_payload
        assert not self.__loading, "The sending of signals should only happen after the loading stage."
        with self.__lock:
            self.__event_store.append(self.__modify_event(uuid=object_model.uuid, **properties))

    def __handle_move_in_list(self, signal_payload: Tuple[AbstractModel, int]) -> None:
        object_model, to_position = signal_payload
        assert not self.__loading, "The sending of signals should only happen after the loading stage."
        with self.__lock:
            self.__event_store.append(self.__move_event(uuid=object_model.uuid, to_position=to_position))

    def __handle_deletion_from_list(self, object_model: AbstractModel) -> None:
        assert not self.__loading, "The sending of signals should only happen after the loading stage."
        with self.__lock:
            self.__event_store.append(self.__delete_event(uuid=object_model.uuid))
            object_model.onPreChange.disconnect(self.__handle_model_change)
