import datetime
import json
import logging
import os
import time
from typing import Callable, Dict, Iterator, List, Optional, Union

import sqlalchemy
import sqlalchemy.sql

from app.libs import watchdog
from app.libs.event_store_exceptions import EventStoreEventClassUnknown
from app.storage.persistent_event import PersistentEvent
from um3_commons import thread

metadata = sqlalchemy.MetaData()
logger = logging.getLogger(__name__)

# The 'events' tables holds the main data regarding an event.
persistentEventTable = sqlalchemy.Table(
                        "events", metadata,
                        sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),

                        # time since system boot in fractional seconds.
                        sqlalchemy.Column("boot_time", sqlalchemy.FLOAT),

                        # Event creation real time. This may be completely incorrect on many
                        # systems, and it may jump around if the RTC is updated from an NTP server.
                        sqlalchemy.Column("timestamp", sqlalchemy.TIMESTAMP),

                        sqlalchemy.Column("event_type", sqlalchemy.Integer, index=True),
                        sqlalchemy.Column("entity_uuid", sqlalchemy.String, index=True),
                        sqlalchemy.Column("properties", sqlalchemy.String))


EventSubscriber = Callable[[bool, PersistentEvent], None]
EventPropertiesDict = Optional[Dict[str, Union[str, int]]]


class EventStore:
    """The EventStore itself.

    Persistent storage is done via SQLite.
    """

    def __init__(self, database_filepath: str) -> None:
        """Constructor

        Arguments:
        database_filepath -- the name of the file to use to store the SQL database in.
        """

        os.makedirs(os.path.dirname(database_filepath), exist_ok=True)

        self._database_filepath = database_filepath
        self._engine = sqlalchemy.create_engine("sqlite:///%s" % (self._database_filepath))
        metadata.create_all(self._engine)
        self._subscribers = []  # type: List[EventSubscriber]
        self._registered_events = {}  # type: Dict
        self._subscribers_by_event = {}  # type: Dict
        self._lock = thread.RLock()
        self._all_events_replayed = False

    def dispose(self) -> None:
        if self._engine:
            self._engine.dispose()
            self._engine = None

    def _get_connection(self):
        return self._engine.connect()

    def create_event(self, event_type: int, entity_uuid: str,
                     props_dict: EventPropertiesDict = None) -> PersistentEvent:
        """Create an event object.

        This doesn't add the event to the EventStore. See `append_event`

        Arguments:
        event_type -- this is an application defined ID for the type of event.
        entity_uuid -- the UUID of the entity this event pertains to. This
                       string may not be longer than 36 characters.
        props_dict -- extra properties to be stored with this event. It is just
                       pairs of keys and values.

        Returns the new event. The ID field of the event is not valid until the
        event has been persisted.
        """
        event = PersistentEvent(id=None, event_type=event_type, entity_uuid=entity_uuid,
                                props_dict=props_dict)
        return event

    def subscribe(self, subscriber: EventSubscriber) -> None:
        """Register a subscriber to receive events.

        Arguments:
        subscriber -- the function to call with events. It is called with two arguements. First
                      is a boolean indicating if the event is an event being replayed. Second
                      is the  event object.
        """
        with self._lock:
            self._subscribers.append(subscriber)

    def register_event(self, event_class, event_type):
        self._registered_events[event_class] = event_type

    def subscribe_to(self, event_class, event_handler):
        event_type = self._registered_events.get(event_class)
        subscribers = self._subscribers_by_event.get(event_type)

        if event_type is None:
            raise EventStoreEventClassUnknown(
                "Programming error, subscribing to an unknown event type: %s" % (event_class))
        if subscribers is None:
            self._subscribers_by_event[event_type] = []

        self._subscribers_by_event[event_type].append(event_handler)

    def unsubscribe(self, event_handler):
        for event_type, subscibers in self._subscribers_by_event.items():
            if event_handler in subscibers:
                subscibers.remove(event_handler)
                return
        if event_handler in self._subscribers:
            self._subscribers.remove(event_handler)

    def append(self, event):    # FIXME event should be of type base_event
        persistent_event = PersistentEvent(
            id=None,
            event_type=self._registered_events.get(type(event)),
            entity_uuid=event.uuid,
            props_dict=event.to_dict()
        )

        self.append_event(persistent_event)

    def append_event(self, event: PersistentEvent) -> None:
        """Append an event to the event store.

        This will persist the event and also notify any subscribers.

        Arguments:
        event -- the event to persist. It's ID field will be filled
                    with its new allocated event ID.
        """
        assert self._all_events_replayed, "All events must have been replayed before new events can be triggered!"
        with self._lock:
            connection = self._get_connection()
            try:
                trans = connection.begin()

                try:
                    insert_stmt = persistentEventTable.insert()

                    event.boot_time = time.monotonic()
                    event.timestamp = datetime.datetime.now()
                    flat_properties = json.dumps(event.props) if event.props else None

                    insert = insert_stmt.values(boot_time=event.boot_time,
                                                timestamp=event.timestamp,
                                                event_type=event.event_type,
                                                entity_uuid=event.entity_uuid,
                                                properties=flat_properties)
                    connection.execute(insert)

                    trans.commit()

                    self.__notify_subscribers(event, False)

                except:
                    trans.rollback()
                    raise
            finally:
                connection.close()

    def __notify_subscribers(self, event, is_replay: bool):
        for sub in self._subscribers:
            sub(is_replay, event)

        for sub in self._subscribers_by_event.get(event.event_type, {}):
            event_class = self.__get_event_by_event_type(event.event_type)
            if event_class is not None:
                deserialized_event = event_class.from_persistent_event(event)
                sub(is_replay, deserialized_event)

    def __get_event_by_event_type(self, event_type):
        if event_type in self._registered_events.values():
            for key, value in self._registered_events.items():
                if value == event_type:
                    return key

    def all_events(self) -> Iterator[PersistentEvent]:
        """Get an iterator over all of the events in the store.

        Returns an iterator which gives PersistentEvent objects.
        """
        with self._lock:
            connection = self._get_connection()
            try:
                event_select = sqlalchemy.sql.select([persistentEventTable]) \
                    .order_by(persistentEventTable.c.id)
                event_results = connection.execute(event_select)
                try:
                    for event_row in event_results:
                        event = PersistentEvent.from_row(event_row)
                        yield event
                finally:
                    event_results.close()
            finally:
                connection.close()

    def replay_all_events(self) -> None:
        with self._lock:
            logger.info("Loading all events")
            start_time = time.perf_counter()
            counter = 0
            for event in self.all_events():
                self.__notify_subscribers(event, True)
                counter += 1
                if counter % 250 == 0:
                    logger.info("Loaded %i events.", counter)
                    watchdog.get_instance().tick()
            end_time = time.perf_counter()
            logger.info("Loading took %f seconds to process %i events.", end_time - start_time, counter)

            self._all_events_replayed = True

    def dump_all_events(self) -> None:
        for event in self.all_events():
            print(repr(event))

    # Function to clear out all the events from the event store.
    # Called from the factory reset.
    def clear_event_store(self):
        # Grabbing the lock here to prevent any more database transactions.
        # Any future transactions will be useless as we are deleting the whole eventstore anyhow.
        with self._lock:
            # We unlink the database file, the event store could still be open, and events could be written,
            # however, due how unix handles files, the actual deletion will happen as soon as the database is closed.
            # As no more active links from the filesystem are available. This makes for a clean and easy delete.
            logger.info("Clearing event store")
            os.unlink(self._database_filepath)
            logger.info("Finished clearing the event store")
