from ..avrisp.avrISP import AvrISP
from ..stk500v2.stk500v2 import Stk500v2

import os
import logging


log = logging.getLogger(__name__.split(".")[-1])

## Multi stage programmer.
#  Uses both the AvrISP programmer and the Stk500v2 programmer.
#  First programms the AVR with the right bootloader in AvrISP mode.
#  Once this is finished, it will connect to the bootloader and program the supplied firmware into the chip with the bootloader.
class MultiStageProgrammer:
    def __init__(self):
        self.__stk500v2 = Stk500v2()
        self.__bootloader_search_paths = [
            "/lib/firmware/avr/bootloader",
            os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "bootloader")),
        ]

    # Connect to the hardware. Returns false if it fails to connect.
    # @param target: The target string is programmer specific, and is used to identify the hardware device used for communication.
    # @return: bool, True if opening the hardware and connecting to the chip was succesful. False if this failed.
    def connect(self, **kwargs):
        if "bootloader" not in kwargs:
            raise TypeError("Avr multi installer requires bootloader argument in from of a single string or a dictionary chip->bootloader mapping")
        if not self.__stk500v2.connect(**kwargs):
            log.info("Couldn't connect the Stk500v2 for multi stage programming. Going to install the bootloader")
            avrisp = AvrISP()
            if not avrisp.connect(**kwargs):
                log.error("Failed to connect the AvrISP, cannot install the bootloader!")
                avrisp.close()
                return False
            bootloader = kwargs["bootloader"]
            if isinstance(bootloader, dict):
                chip_data = avrisp.getChipData()
                if chip_data["name"] not in bootloader:
                    log.error("No suitable bootloader for %s", chip_data["name"])
                    return False
                bootloader = bootloader[chip_data["name"]]
            if not os.path.isfile(bootloader):
                # Bootloader not found yet, search in our default search paths.
                for path in self.__bootloader_search_paths:
                    if os.path.isfile(os.path.join(path, bootloader)):
                        bootloader = os.path.join(path, bootloader)
                        break
            if not os.path.isfile(bootloader):
                log.error("Bootloader not found: %s", bootloader)
                return False
            if not avrisp.program(bootloader):
                log.error("Failed to program the bootloader")
                return False
            avrisp.close()
            if not self.__stk500v2.connect(**kwargs):
                log.info("Couldn't connect the Stk500v2 for multi stage programming, after installing the bootloader! (wrong serial port in bootloader or programmer?)")
                return False
        return True
    
    # Program a hex file into the chip.
    # @param file: The filename of the firmware to program.
    # @return: bool, True if programming was successful, False if it failed.
    def program(self, file):
        return self.__stk500v2.program(file)
    
    # Close the connection to the hardware, and boot the chip into the firmware.
    # Without this call the chip can remain in bootloader/programming mode and never start.
    def close(self):
        self.__stk500v2.close()
    
    # Set a callback function to be called when the program function is making progress.
    # @param callback: A callable which will be called with a single float in the range 0 to 1 to indicate the progress made on updating the firmware, where 1.0 is finished.
    # @param callback: None de-attaches the previously attached callback.
    def setProgressCallback(self, callback):
        self.__stk500v2.setProgressCallback(callback)

    # get the current firmware update progress
    # @return float between 0 and 1 that indicates the update progress
    def getProgress(self):
        return self.__stk500v2.getProgress()


def main(filename):
    avr_isp = MultiStageProgrammer()
    if avr_isp.connect(softspi="/dev/gpio/avr", reset="/dev/gpio/avr_reset", port="/dev/ttyS1", speed="250000", bootloader={"ATMega1280": "bootloader-atmega1280-uart250000-2.hex", "ATMega2560": "bootloader-atmega2560-uart250000-2.hex"}):
        avr_isp.program(filename)
    avr_isp.close()
    return True

if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)-15s %(levelname)-8s %(name)-15s %(message)s", level=logging.DEBUG)

    import sys
    for arg in sys.argv[1:]:
        main(sys.argv[1])
