import struct
import sys
import time

from serial import Serial   # type: ignore
from serial import SerialException
from serial import SerialTimeoutException

import logging

log = logging.getLogger(__name__.split(".")[-1])


## Low level STK500v2 implementation.
#  Note that this was partially copied from the implementation found in the original Cura, but cleaned up and documented.
class LowLevelStk500v2:
    #*****************[ STK message constants ]***************************
    MESSAGE_START                     = 0x1B
    TOKEN                             = 0x0E

    #*****************[ STK general command constants ]**************************
    CMD_SIGN_ON                       = 0x01
    CMD_SET_PARAMETER                 = 0x02
    CMD_GET_PARAMETER                 = 0x03
    CMD_SET_DEVICE_PARAMETERS         = 0x04
    CMD_OSCCAL                        = 0x05
    CMD_LOAD_ADDRESS                  = 0x06
    CMD_FIRMWARE_UPGRADE              = 0x07


    #*****************[ STK ISP command constants ]******************************
    CMD_ENTER_PROGMODE_ISP            = 0x10
    CMD_LEAVE_PROGMODE_ISP            = 0x11
    CMD_CHIP_ERASE_ISP                = 0x12
    CMD_PROGRAM_FLASH_ISP             = 0x13
    CMD_READ_FLASH_ISP                = 0x14
    CMD_PROGRAM_EEPROM_ISP            = 0x15
    CMD_READ_EEPROM_ISP               = 0x16
    CMD_PROGRAM_FUSE_ISP              = 0x17
    CMD_READ_FUSE_ISP                 = 0x18
    CMD_PROGRAM_LOCK_ISP              = 0x19
    CMD_READ_LOCK_ISP                 = 0x1A
    CMD_READ_SIGNATURE_ISP            = 0x1B
    CMD_READ_OSCCAL_ISP               = 0x1C
    CMD_SPI_MULTI                     = 0x1D

    CMD_ULTI_CHECKSUM                 = 0xEE

    #*****************[ STK PP command constants ]*******************************
    CMD_ENTER_PROGMODE_PP             = 0x20
    CMD_LEAVE_PROGMODE_PP             = 0x21
    CMD_CHIP_ERASE_PP                 = 0x22
    CMD_PROGRAM_FLASH_PP              = 0x23
    CMD_READ_FLASH_PP                 = 0x24
    CMD_PROGRAM_EEPROM_PP             = 0x25
    CMD_READ_EEPROM_PP                = 0x26
    CMD_PROGRAM_FUSE_PP               = 0x27
    CMD_READ_FUSE_PP                  = 0x28
    CMD_PROGRAM_LOCK_PP               = 0x29
    CMD_READ_LOCK_PP                  = 0x2A
    CMD_READ_SIGNATURE_PP             = 0x2B
    CMD_READ_OSCCAL_PP                = 0x2C    

    CMD_SET_CONTROL_STACK             = 0x2D

    #*****************[ STK HVSP command constants ]*****************************
    CMD_ENTER_PROGMODE_HVSP           = 0x30
    CMD_LEAVE_PROGMODE_HVSP           = 0x31
    CMD_CHIP_ERASE_HVSP               = 0x32
    CMD_PROGRAM_FLASH_HVSP            = 0x33
    CMD_READ_FLASH_HVSP               = 0x34
    CMD_PROGRAM_EEPROM_HVSP           = 0x35
    CMD_READ_EEPROM_HVSP              = 0x36
    CMD_PROGRAM_FUSE_HVSP             = 0x37
    CMD_READ_FUSE_HVSP                = 0x38
    CMD_PROGRAM_LOCK_HVSP             = 0x39
    CMD_READ_LOCK_HVSP                = 0x3A
    CMD_READ_SIGNATURE_HVSP           = 0x3B
    CMD_READ_OSCCAL_HVSP              = 0x3C

    #*****************[ STK status constants ]***************************

    #Success
    STATUS_CMD_OK                     = 0x00

    #Warnings
    STATUS_CMD_TOUT                   = 0x80
    STATUS_RDY_BSY_TOUT               = 0x81
    STATUS_SET_PARAM_MISSING          = 0x82

    #Errors
    STATUS_CMD_FAILED                 = 0xC0
    STATUS_CKSUM_ERROR                = 0xC1
    STATUS_CMD_UNKNOWN                = 0xC9

    #*****************[ STK parameter constants ]***************************
    PARAM_BUILD_NUMBER_LOW            = 0x80
    PARAM_BUILD_NUMBER_HIGH           = 0x81
    PARAM_HW_VER                      = 0x90
    PARAM_SW_MAJOR                    = 0x91
    PARAM_SW_MINOR                    = 0x92
    PARAM_VTARGET                     = 0x94
    PARAM_VADJUST                     = 0x95
    PARAM_OSC_PSCALE                  = 0x96
    PARAM_OSC_CMATCH                  = 0x97
    PARAM_SCK_DURATION                = 0x98
    PARAM_TOPCARD_DETECT              = 0x9A
    PARAM_STATUS                      = 0x9C
    PARAM_DATA                        = 0x9D
    PARAM_RESET_POLARITY              = 0x9E
    PARAM_CONTROLLER_INIT             = 0x9F

    #*****************[ STK answer constants ]***************************
    ANSWER_CKSUM_ERROR                = 0xB0


    def __init__(self, port, speed=115200, reset=None, **kwargs):
        assert port is not None
        assert int(speed) > 0
        
        self.__serial = None
        self.__serial_port_name = port
        self.__serial_port_speed = int(speed)
        self.__reset_gpio = reset
        self.__sequence_number = 1

    ## Enter programming mode.
    #  @return True if entering programming mode was succesful, false if otherwise.
    def enterProgrammingMode(self):
        if self.__serial:
            self.__serial.close()
        self.__serial = Serial(self.__serial_port_name, self.__serial_port_speed, timeout=1, writeTimeout=10000)
        self.__setReset(True)
        time.sleep(0.010)
        self.__setReset(False)
        time.sleep(0.100)   # Wait at least 100ms, the AVR startup fuse settings can contain a boot delay of 64ms, so we need to ensure that this time has past and the bootloader code is running.
        
        self.__serial.flushInput()
        self.__serial.flushOutput()

        self.__sequence_number = 1

        self.__serial.timeout = 0.1
        if self.__sendMessageAndGetReply(bytes([self.CMD_ENTER_PROGMODE_ISP, 0xc8, 0x64, 0x19, 0x20, 0x00, 0x53, 0x03, 0xac, 0x53, 0x00, 0x00])) != bytes([self.CMD_ENTER_PROGMODE_ISP, self.STATUS_CMD_OK]):
            log.warning("Failed to enter programming mode")
            time.sleep(1)
            return False
        self.__serial.timeout = 5
        return True

    ## Leave programming mode after we entered programming mode previously by enterProgrammingMode()
    def leaveProgrammingMode(self):
        if self.__serial:
            if self.__sendMessageAndGetReply(bytes([self.CMD_LEAVE_PROGMODE_ISP])) != bytes([self.CMD_LEAVE_PROGMODE_ISP, self.STATUS_CMD_OK]):
                log.warning("Failed to leave programming mode")
                self.__setReset(True)
                time.sleep(0.010)
                self.__setReset(False)
                time.sleep(0.100)
            self.__serial.close()
        self.__serial = None

    ## Read the signature bytes
    #  @return bytes object of the length of 3. The signature bytes indicate which chip is connected to the programmer.
    def readSignature(self):
        data0 = self.__sendMessageAndGetReply(bytes([self.CMD_READ_SIGNATURE_ISP, 0, 0, 0, 0]))
        data1 = self.__sendMessageAndGetReply(bytes([self.CMD_READ_SIGNATURE_ISP, 0, 0, 0, 1]))
        data2 = self.__sendMessageAndGetReply(bytes([self.CMD_READ_SIGNATURE_ISP, 0, 0, 0, 2]))
        if data0 is None or data1 is None or data2 is None:
            log.warning("Failed to get signature bytes: %s %s %s", data0, data1, data2)
            return None
        return bytes([data0[2], data1[2], data2[2]])

    ## Read the lock byte
    #  @return integer containing the lock bits specific for this chip.
    def readLock(self):
        data = self.__sendMessageAndGetReply(bytes([self.CMD_READ_LOCK_ISP]))
        if data is None:
            return None
        return data[2]

    ## Read the chip fuse bytes. The fuse bytes configure clock source, as well as bootloader type, and are chip specific.
    #  @return bytes object with the length of 3.
    def readFuse(self):
        # As we cannot program the fuses, we return None here to indicate that we have no fuse access.
        return None

    ## Write the fuse bytes, if required. This first reads the fuse bytes to check if writing is needed.
    #  @param fuses: bytes object containing the fuse data for the fuses.
    #  @return True if the fuses where programmed succesfully.
    def writeFuse(self, fuses):
        raise NotImplementedError("We cannot write fuses from the stk500v2 bootloader access")

    ## Read the memory between the low and high address, excluding the high address itself.
    #  @param low_addr: Integer containing the low address
    #  @param high_addr: Integer containing the high address
    #  @return bytes object containing the flash data from the low to high address range.
    def readMemory(self, low_addr, high_addr):
        if self.__sendMessageAndGetReply(struct.pack(">BI", self.CMD_LOAD_ADDRESS, low_addr >> 1)) is None:
            return None
        data = self.__sendMessageAndGetReply(struct.pack(">BH", self.CMD_READ_FLASH_ISP, high_addr - low_addr))
        if data is None or len(data) < 2:
            return None
        return data[2:-1]

    ## Write memory page
    #  @param address: integer, containing the absolute flash address of this page.
    #  @param page_data: bytes object containing the data for this page to write. Should be the exact lenght of a flash page for this chip.
    #  @return boolean, True if writing the page was succesful.
    def writePage(self, address, page_data):
        if self.__sendMessageAndGetReply(struct.pack(">BI", self.CMD_LOAD_ADDRESS, address >> 1)) is None:
            return False
        result = self.__sendMessageAndGetReply(struct.pack(">BHBBBBBBB", self.CMD_PROGRAM_FLASH_ISP, len(page_data), 0xc1, 0x0a, 0x40, 0x4c, 0x20, 0x00, 0x00) + page_data)
        if result is None:
            return False
        if result[1] != self.STATUS_CMD_OK:
            return False
        return True

    ## Get a checksum of an address range.
    #  @param low_addr: Integer containing the low address
    #  @param high_addr: Integer containing the high address
    #  @return None if this function is not supported by the bootloader.
    #  @return Integer in the range 0 to 0xffff, which is the sum of all the bytes between the low and high address (including the low, excluding the high address)
    def getChecksum(self, low_addr, high_addr):
        low_checksum = self.__getChecksum(low_addr)
        if low_checksum is None:
            return None
        high_checksum = self.__getChecksum(high_addr)
        return (high_checksum - low_checksum) & 0xFFFF

    ## Erase the chip
    #  After this function the chip will contain no data.
    #  Note that the bootloader can never fully erase the chip, as it will protect the bootloader.
    #   On some bootloaders this is a NO-OP.
    def eraseChip(self):
        # We currently skip this command, it takes 13 seconds to execute this command on the ATMega2560 with the arduino bootloader.
        # And the end result is that the EEPROM is erased, but the flash is untouched.
        pass
        # self.__serial.timeout = 30
        # self.__sendMessageAndGetReply(struct.pack(">B", self.CMD_CHIP_ERASE_ISP))
        # self.__serial.timeout = 1

    ## Set the reset line.
    #  @param enable: boolean, True: The chip will be in reset, False, the chip will not be in reset and run the main program.
    def __setReset(self, enable):
        if self.__reset_gpio is None:
            self.__serial.setDTR(enable)
        else:
            with open(self.__reset_gpio, "w") as f:
                if enable:
                    f.write("0")
                else:
                    f.write("1")

    ## Send a checksum request and get a reply.
    #  The checksum is done from address 0 up to the given address.
    #  @param address: End address for the checksum, in bytes. Checksum is calculated up to this address, but not including the address.
    #  @return None if the checksum function is not supported by the bootloader (example Arduino bootloader)
    #  @return integer value between 0 and 0xFFFF which is a sum of all the flash bytes between address 0 and the given address.
    def __getChecksum(self, address):
        if self.__sendMessageAndGetReply(struct.pack(">BI", self.CMD_LOAD_ADDRESS, address >> 1)) is None:
            return None
        checksum_reply = self.__sendMessageAndGetReply(bytes([self.CMD_ULTI_CHECKSUM]))
        if checksum_reply is None:
            return None
        if len(checksum_reply) == 4 and checksum_reply[1] == self.STATUS_CMD_OK:
            return struct.unpack("<H", checksum_reply[2:4])[0]
        return None

    ## Send a message to the bootloader, and capture the reply.
    #  Encode the message into the bootloader protocol.
    #  @param data: Bytes object containing the payload to be send.
    #  @return bytes object containing the reply, or None on a timeout.
    def __sendMessageAndGetReply(self, data):
        message = struct.pack(">BBHB", self.MESSAGE_START, self.__sequence_number, len(data), self.TOKEN)
        message += data
        checksum = 0
        for c in message:
            checksum ^= c
        message += struct.pack(">B", checksum)
        try:
            self.__serial.write(message)
        except SerialTimeoutException:
            raise ispBase.IspError("Serial send timeout")
        sequence_number = self.__sequence_number
        self.__sequence_number = (self.__sequence_number + 1) & 0xFF
        return self.__receiveMessage(sequence_number)

    ## Read the response from the bootloader.
    #  @return bytes object containing the reply, or None on a timeout.
    def __receiveMessage(self, wanted_sequence_number):
        state = "Start"
        checksum = 0
        while True:
            buffer = self.__serial.read()
            if len(buffer) < 1:
                return None
            for b in buffer:
                checksum ^= b
                if state == "Start":
                    if b == self.MESSAGE_START:
                        state = "GetSeq"
                        checksum = self.MESSAGE_START
                elif state == "GetSeq":
                    sequence_number = b
                    state = "MsgSize1"
                elif state == "MsgSize1":
                    msg_size = b << 8
                    state = "MsgSize2"
                elif state == "MsgSize2":
                    msg_size |= b
                    state = "Token"
                elif state == "Token":
                    if b != self.TOKEN:
                        state = "Start"
                    else:
                        state = "Data"
                        data = bytearray()
                elif state == "Data":
                    data.append(b)
                    if len(data) == msg_size:
                        state = "Checksum"
                elif state == "Checksum":
                    if checksum != 0:
                        state = "Start"
                    elif wanted_sequence_number == sequence_number:
                        return data
