from ..common import chipDB
from ..common.intelHex import IntelHex

import binascii
import logging

log = logging.getLogger(__name__.split(".")[-1])


class FirmwareUpdateBase:
    def __init__(self):
        self._low_level = None
        self.__progress_amount = 0.0
        self.__progress_callback = None
        self.__has_checksum_function = False

    # Connect to the hardware. Returns false if it fails to connect.
    # @param target: The target to program. In the form of "[arg]:[value] [arg]:[value]"
    # @return: bool, True if opening the hardware and connecting to the chip was succesful. False if this failed.
    def connect(self, **kwargs):
        self._low_level = self._connectLowLevel(**kwargs)
        if self._low_level is None:
            return False
        return self.__tryToEnterProgrammingMode()

    def _connectLowLevel(self, target):
        raise NotImplementedError("Subclass should implement the _connectLowLevel function and return the low level communication")

    ## Try to enter the programming mode, this retries a few times to increase success rate.
    #  @return bool: True if entered programming mode, False if failed.
    def __tryToEnterProgrammingMode(self):
        for n in range(3):  # Do 3 retries when trying to enter programming mode. Because the first time sometimes fails. And 3 times is the charm.
            if self._low_level.enterProgrammingMode():
                return True
            self._low_level.leaveProgrammingMode()
        return False
    
    ## Get the chip data from the chip database.
    #  @return None if the chip is not found or we cannot read the signature.
    #  @return dictionary from chipDB.json containing the chip data.
    def getChipData(self):
        # First, read the signature to get our chip information from the database.
        signature = self._low_level.readSignature()
        if signature is None:
            log.error("Failed to get chip signature.")
            return None
        chip_info = chipDB.get(signature)
        if chip_info is None:
            log.error("Could not find chip in database: %s", signature)
            return None
        return chip_info
            
    # Program a hex file into the chip.
    # @param file: The filename of the firmware to program.
    # @return: bool, True if programming was successful, False if it failed.
    def program(self, file, *, fuses=None):
        if not isinstance(file, IntelHex):
            file = IntelHex(file)

        # First, read the signature to get our chip information from the database.
        chip_info = self.getChipData()
        if chip_info is None:
            return False

        # Check if any lockbits are programmed, and erase those.
        if self._low_level.readLock() != 0xFF:
            log.info("Chip is locked, erasing it, so it is unlocked again.")
            self._low_level.eraseChip()
            if self._low_level.readLock() != 0xFF:
                log.error("Chip still locked after erase.")
                return False

        # Get the page size in bytes, from the page size in words from the chip database.
        page_size = chip_info["pageSize"] * 2
        
        # Check if we have the checksum function, this allows for quicker validation of the installed firmware.
        # And thus also has impact on our progress calculation.
        self.__has_checksum_function = self._low_level.getChecksum(0, 0) != None
        
        chip_fuses = self._low_level.readFuse()
        if chip_fuses is not None:
            if fuses is None:
                fuses = binascii.unhexlify(chip_info["default_fuses"])
            log.debug("Setting fuses to: %s", fuses)
            if not self._low_level.writeFuse(fuses):
                log.error("Failed to set the fuses.")
                return False

        # Check if we need to program, else we can just skip the whole process.
        if self.__verifyMemory(file, page_size, report_progress=False):
            log.debug("Firmware matched, no programming needed.")
            return True

        self.__updateProgress(0.0)
        # Erase the chip before writing.
        self._low_level.eraseChip()
        # Write the flash per page.
        for address, block in file:
            if address % page_size != 0:
                log.warning("Start address for block is not page aligned: %x", address)
                offset = address % page_size
                address -= offset
                block = b'\x00' * offset + block
            page_count = (len(block) + page_size - 1) // page_size
            for page_nr in range(page_count):
                if self.__has_checksum_function:
                    self.__updateProgress(page_nr / page_count)
                else:
                    self.__updateProgress(page_nr / page_count * 0.5)   # If we don't have the checksum function, then verfication takes as much time as writting, so progess is halfed.
                page_data = block[page_nr * page_size:page_nr * page_size + page_size]
                if len(page_data) < page_size:
                    page_data += b'\x00' * (page_size - len(page_data))
                log.debug("Writing page %d/%d", page_nr, page_count)
                if not self._low_level.writePage(address + page_nr * page_size, page_data):
                    log.debug("Page write failed!")
                    return False

        # Check the final result.
        if not self.__verifyMemory(file, page_size):
            log.debug("Verification failed!")
            self._low_level.eraseChip()
            self.__updateProgress(1.0)
            return False
        self.__updateProgress(1.0)
        log.debug("Programming finished.")
        return True

    def __verifyMemory(self, file, page_size, *, report_progress=True):
        if self.__has_checksum_function:
            # Use checksum verification (faster)
            for address, block in file:
                low_address = address
                high_address = low_address + len(block)
                log.debug("Checksum 0x%x-0x%x for verification", low_address, high_address)
                checksum = self._low_level.getChecksum(low_address, high_address)
                if checksum != (sum(block) & 0xFFFF):
                    log.debug("Checksum mismatch: %x %x", checksum, (sum(block) & 0xFFFF))
                    return False
        else:
            # Use flash read verification (slower)
            for address, block in file:
                page_count = (len(block) + page_size - 1) // page_size
                for page_nr in range(page_count):
                    if report_progress:
                        self.__updateProgress(0.5 + page_nr / page_count * 0.5)   # If we don't have the checksum function, then verfication takes as much time as writting, so progess is halfed.

                    this_size = min(page_size, len(block) - page_nr * page_size)
                    block_data = block[page_nr * page_size: page_nr * page_size + this_size]

                    low_address = address + page_nr * page_size
                    high_address = low_address + this_size

                    log.debug("Reading 0x%x-0x%x for verification", low_address, high_address)
                    read_data = self._low_level.readMemory(low_address, high_address)
                    if read_data != block_data:
                        log.debug("Mismatch on page %d", page_nr)
                        return False
        return True

    # Close the connection to the hardware, and boot the chip into the firmware.
    # Without this call the chip can remain in bootloader/programming mode and never start.
    def close(self):
        if self._low_level:
            self._low_level.leaveProgrammingMode()

    # Set a callback function to be called when the program function is making progress.
    # @param callback: A callable which will be called with a single float in the range 0 to 1 to indicate the progress made on updating the firmware, where 1.0 is finished.
    # @param callback: None de-attaches the previously attached callback.
    def setProgressCallback(self, callback):
        self.__progress_callback = callback

    # get the current firmware update progress
    # @return float between 0 and 1 that indicates the update progress
    def getProgress(self):
        return self.__progress_amount

    # Update the current progress to the set amount. This updates the internal value, as well as calling the callback if required.
    # @param amount: Progress amount beteen 0.0 and 1.0
    def __updateProgress(self, amount):
        self.__progress_amount = amount
        if self.__progress_callback:
            self.__progress_callback(amount)
