"""
Module to read intel hex files into binary data blobs.
IntelHex files are commonly used to distribute firmware
See: http://en.wikipedia.org/wiki/Intel_HEX
"""
import binascii
import struct

class IntelHex:
    # Open and read an intel hex file.
    # Reads the hex file in data blocks, each data block is a continues piece of data from a start address.
    def __init__(self, filename):
        self.__data_blocks = []
        
        self.__file = open(filename, "r")
        self.__readFile()
        self.__file.close()
    
    ## Read a single line out of the intel hex file.
    #  Throws exceptions when the file is in an incorrect format.
    #  @return a tuple containing the record type number, the address, and the data for this record.
    #  @return a tuple containing None if the end of file was reached.
    def __readNextLine(self):
        line = self.__file.readline()
        line = line.strip()
        if len(line) < 1:
            return None, None, None
        if line[0] != ":":
            raise Exception("Hex file has a line not starting with ':'")
        data = binascii.unhexlify(line[1:])
        record_length, address, record_type = struct.unpack(">BHB", data[0:4])
        if len(data) != record_length + 5:
            raise Exception("Error in hex file: " + line)
        check_sum = sum(data) & 0xFF
        if check_sum != 0:
            raise Exception("Checksum error in hex file: " + line)
        return record_type, address, data[4:-1]

    ## Read the whole file contents
    #  Uses __readNextLine to parse the intel hex file.
    def __readFile(self):
        block_start = None
        block_data = None
        extra_addr = 0  #< extended address set py record type 2 or 4
        while True:
            record_type, address, data = self.__readNextLine()
            if record_type is None:
                break

            if record_type == 0:    # Data record
                address += extra_addr
                if block_data is None or block_start + len(block_data) != address:
                    block_data = bytearray()
                    block_start = address
                    self.__data_blocks.append((block_start, block_data))
                block_data[address-block_start:address-block_start+len(data)] = data
            elif record_type == 1:  # End Of File record
                break
            elif record_type == 2:  # Extended Segment Address Record
                extra_addr = struct.unpack(">H", data)[0] * 16
            elif record_type == 3:  # Start Segment Address
                pass    # We ignore the start address, as we no use for it.
            elif record_type == 4:  # Extended Linear Address Record
                extra_addr = struct.unpack(">H", data)[0] << 16
            else:
                raise Exception("Unknown record type in intel hex file: %02x" % (record_type))

    def __iter__(self):
        return iter(self.__data_blocks)

if __name__ == "__main__":
    import sys
    for arg in sys.argv[1:]:
        print("File: %s" % (arg))
        for start, block in IntelHex(arg):
            print("Block: %08x-%08x" % (start, start+len(block)))
