#!/bin/sh
#
# Copyright (C) 2017-2018 Ultimaker B.V.
# Copyright (C) 2017-2018 Olliver Schinagl <oliver@schinagl.nl>
#
# SPDX-License-Identifier: LGPL-3.0+

set -eu

UPDATE_ROOT_DIR="${UPDATE_ROOT_DIR:-${1}}"
if [ ! -d "${UPDATE_ROOT_DIR}" ]; then
	echo "${0} needs the location of the extracted tarball supplied."
	exit 1
fi

parse_cmdline()
{
       if [ ! -r "/proc/cmdline" ]; then
               echo "Unable to read /proc/cmdline to parse boot arguments."
               exit 1
       fi

       for cmd in $(cat /proc/cmdline); do
               case "${cmd}" in
               root=*)
                       local _root="${cmd#*=}"
                       local _prefix="${_root%%=*}"

                       if [ "${_prefix}" = "UUID" ] || \
                          [ "${_prefix}" = "PARTUUID" ] || \
                          [ "${_prefix}" = "LABEL" ] || \
                          [ "${_prefix}" = "PARTLABEL" ]; then
                               root=$(findfs "${_root}")
                       else
                               root="${cmd#*=}"
                       fi
               ;;
               rootflags=*)
                       rootflags="${cmd#*=}"
               ;;
               rootfstype=*)
                       rootfstype="${cmd#*=}"
               ;;
               esac
       done
}

echo "Replacing kernel and boot parameters ... "
parse_cmdline

TARGET_BOOT_DIR=$(mktemp -t -d um_updater_bootdisk.XXXXXX)

BOOTDEV=""
BOOTDISK=""

# BOOTDISK
# To obtain the bootdisk from the rootdev we need to find the parent of the
# rootdev in sysfs.
# A bootdisk would look like /dev/mmcblk0. To obtain this from our previously
# found root, we first have to turn it into a sysfs node via a regular
# expression by replacing "dev" with "sys/class/block" once.
# The obtained symlink goes to the actual physical sysfs node. Herein the
# root partition is a child (subdirectory) of the actual disk. dirname
# resolves this for us and by dropping the previous paths with basename
# we are left with the disk "mmcblk0". A simple prefix of "/dev/" brings us to
# our bootdisk of "/dev/mmcblk0".
# Note, that this makes the assumption that our bootloader is located on the
# same disk as our rootfs.
BOOTDISK="/dev/"$(basename $(dirname $(readlink $(echo "${root}" | sed -e 's/dev/sys\/class\/block/1'))))

# BOOTDEV
# The bootloader files are on the ${BOOTDEV} partition, which is assumed to be
# the first disk of our ${BOOTDISK}.
BOOTDEV="${BOOTDISK}p1"

UBOOT_LOC=8192
UBOOTENV_LOC=557056
UBOOTENV="um_u-boot.env.bin"
UBOOTBIN="u-boot-sunxi-with-spl.bin"

KERNELBIN=uImage-sun7i-a20-opinicus_v1

# U-Boot doesn't like if its files are split over multiple extends. Reformat
# during upgrade if extents where used, since we cannot force defrag with
# e4defrag nor disable the feature with tune2fs.
umount "${BOOTDEV}" > /dev/null 2>&1 || true
if tune2fs -l "${BOOTDEV}" | grep extents 2>&1 > /dev/null; then
    mkfs.ext4 -F -L boot -O ^extents,^64bit "${BOOTDEV}"
fi
mount "${BOOTDEV}" "${TARGET_BOOT_DIR}"

echo "Copying files from ${UPDATE_ROOT_DIR}/tmp to ${TARGET_BOOT_DIR}"
cp -aT "${UPDATE_ROOT_DIR}/boot/" "${TARGET_BOOT_DIR}/"

if [ "${rootfstype}" = "f2fs" ]; then
    # Roofs type is f2fs, so we must be booting from eMMC.
    cp "${UPDATE_ROOT_DIR}/boot/boot_emmc.scr" "${TARGET_BOOT_DIR}/boot.scr"
elif [ "${rootfstype}" = "ext4" ]; then
    # We are likely running ext4, which is the case when we are running from mmc.
    cp "${UPDATE_ROOT_DIR}/boot/boot_mmc.scr" "${TARGET_BOOT_DIR}/boot.scr"
else
    echo "Unsupported rootfstype: ${rootfstype}."
    exit 1
fi

echo "Unmounting ${TARGET_BOOT_DIR}"
umount "${TARGET_BOOT_DIR}"

# Check if u-boot needs to be updated.
TMPUBOOT=$(mktemp -t um_updater_cur_uboot.XXXXXX)
dd if="${BOOTDISK}" of="${TMPUBOOT}" bs=1 skip="${UBOOT_LOC}" count=$(stat -c "%s" "${UPDATE_ROOT_DIR}/boot/${UBOOTBIN}") 1> /dev/null 2>&1
if ! cmp "${TMPUBOOT}" "${UPDATE_ROOT_DIR}/boot/${UBOOTBIN}"; then
    echo "Installing new u-boot"
    dd if="${UPDATE_ROOT_DIR}/boot/${UBOOTBIN}" of="${BOOTDISK}" bs=1024 seek=$((${UBOOT_LOC} / 1024)) count=1024 conv=notrunc,fsync
fi

# Check if u-boot environment needs to be updated.
if [ -r "${UPDATE_ROOT_DIR}/boot/${UBOOTENV}" ]; then
    TMPUBOOTENV=$(mktemp -t um_updater_cur_ubootenv.XXXXXX)
    dd if="${BOOTDISK}" of="${TMPUBOOTENV}" bs=1 skip="${UBOOTENV_LOC}" count=$(stat -c "%s" "${UPDATE_ROOT_DIR}/boot/${UBOOTENV}") 1> /dev/null 2>&1
    if ! cmp "${TMPUBOOTENV}" "${UPDATE_ROOT_DIR}/boot/${UBOOTENV}"; then
        echo "Installing new u-boot environment."
	dd if="${UPDATE_ROOT_DIR}/boot/${UBOOTENV}" of="${BOOTDISK}" bs=1024 seek=$((${UBOOTENV_LOC} / 1024)) count=128 conv=notrunc,fsync
    fi
fi

echo "Update of bootloader and kernel complete."
