import re
import os
import os.path
import logging
import time

from typing import Optional, Dict
from griffin.devblock import rfkill
from griffin.io.fileHelper import FileHelper

log = logging.getLogger(__name__.split(".")[-1])

class DevBlocker:

    # TODO: The following configuration has been put here because we lack a central place to *read* machine config.
    # As a temp workaround we considered getting the config trough the dbus printer-service's but
    #  because of technicalities we won't do that.
    __BLOCKABLE_DEVICES = {
        "wifi-usb": {
            "idVendor": "0cf3",  # [hex]
            "idProduct": "9271"  # [hex]
        }
    }

    # A mapping from sys device nodes to usbport-rfkill indices
    # Key: a regex to match the usb device node path
    # Value: a rfkill index
    __SYSDEV_RFKILL_MAP = {
        r"^/sys/devices/platform/soc@1c00000/1c13000\.usb/.*$": 0,  # <static-path>1c13000<dot>usb<anything>
        r"^/sys/devices/platform/soc@1c00000/1c14[04]00\.usb/.*$": 1, # <static-path>(1c14400 OR 1c14000)<dot>usb<anything>
        r"^/sys/devices/platform/soc@1c00000/1c1c[04]00\.usb/.*$": 2 # <static-path>(1c1c400 OR 1c1c000)<dot>usb<anything>
    }

    def __init__(self):
        self.__device_rfkill_idx_cache = {}  # type: Dict[str, int]


    def discoverDevices(self) -> None:
        #first enable all usb ports, which are of the "fm" type, which in our case means a usb port
        ports_to_unblock = [info for info in rfkill.list() if "index" in info and "type" in info and info["type"] == "fm"]

        for rfkill_info in ports_to_unblock:
            rfkill.unblock(int(rfkill_info["index"]))

        time.sleep(3)

        #discover the devices by using the side effect that caches the devicename -> rfkill-idx relation
        for name in self.__BLOCKABLE_DEVICES:
            self.__findDevRFKillIdx(name)


    def blockDev(self, devname : str) -> None:
        rfkill_idx = self.__findDevRFKillIdx(devname)

        if rfkill_idx is None:
            return

        rfkill.block(rfkill_idx)

    def unblockDev(self, devname : str) -> None:
        rfkill_idx = self.__findDevRFKillIdx(devname)

        if rfkill_idx is None:
            return

        rfkill.unblock(rfkill_idx)

    def canBlockDev(self, devname : str) -> bool:
        rfkill_idx = self.__findDevRFKillIdx(devname)

        return rfkill.canBlock(rfkill_idx) if rfkill_idx is not None else False

    def __findDevRFKillIdx(self, devname : str) -> Optional[int]:
        if devname in self.__device_rfkill_idx_cache:
            return self.__device_rfkill_idx_cache[devname]

        devinfo = self.__BLOCKABLE_DEVICES[devname]
        idVendor = devinfo["idVendor"]
        idProduct = devinfo["idProduct"]

        usb_path = FileHelper.searchDirectory("/sys/devices/platform",
                                              lambda path: self.__matchesUSBDir(path, idVendor, idProduct))

        if usb_path is None:
            return None

        rfkill_idx = self.__RFKillIdxOfUSBPath(os.path.join("/sys/devices/platform", usb_path))

        if rfkill_idx is not None:
            self.__device_rfkill_idx_cache[devname] = rfkill_idx

        return rfkill_idx

    def __matchesUSBDir(self, path : str, idVendor : str, idProduct : str) -> bool:
        try:
            with open(os.path.join(path, "idVendor")) as f:
                if f.read().strip().lower() != idVendor.lower():
                    return False

            with open(os.path.join(path, "idProduct")) as f:
                if f.read().strip().lower() != idProduct.lower():
                    return False
        except IOError:
            return False
        except FileNotFoundError:
            return False

        return True

    def __RFKillIdxOfUSBPath(self, path) -> Optional[int]:
        matches = [i[1] for i in self.__SYSDEV_RFKILL_MAP.items() if re.fullmatch(i[0], path)]
        return int(matches[0]) if len(matches) > 0 else None


    __instance = None   # type: Optional["DevBlocker"]

    @classmethod
    def getInstance(cls) -> Optional["DevBlocker"]:
        if cls.__instance is None:
            cls.__instance = DevBlocker()
            cls.__instance.discoverDevices()

        return cls.__instance