# New record types (inheriting from Base) must be added here as well
from griffin.history.alchemy.printerSettingRecord import PrinterSettingRecord
from griffin.history.alchemy.database_link import *

from griffin.history.alchemy.printerSettings import PrinterSettings
from sqlalchemy.engine import reflection

from .migrations import v2
from .migrations import v3

import logging


log = logging.getLogger(__name__.split(".")[-1])


## @brief Create the tables in the database if they don't exist already
# If the database is already exist then according to the recent version will be applied database migration
class DatabaseService():
    __DATABASE_VERSION_MIGRATIONS = {
        2: v2.migrateDatabase,
        3: v3.migrateDatabase,
    }

    def __init__(self):
        # Set up tables for all the imported Record-classes

        self.settings = PrinterSettings()

        try:
            self.__runDatabaseInit()

        except Exception as err:
            log.exception("Unexpected error while initializing the database. Reason: %s", err)

    ## @brief Run database initialization
    def __runDatabaseInit(self):
        # load information from the database file located on the printer
        insp = reflection.Inspector.from_engine(engine)
        inspect_table_names = insp.get_table_names()

        # if the length is 0 the the database was not created earlier
        if len(inspect_table_names) == 0:
            self.__createInitialDatabase()

        current_db_version = self.__getCurrentDatabaseVersion()
        latest_db_version = self.settings.getDatabaseVersionValue()

        # Run each migration up to the version we require for this code base.
        # If the database is a higher version then the latest known version, we do nothing.
        while current_db_version < latest_db_version:
            next_version = current_db_version + 1
            if next_version in self.__DATABASE_VERSION_MIGRATIONS:
                # Execute the migration in a session, this makes the migration 100% complete or 100% failed
                # Partial migrations are hell.
                with session_scope() as session:
                    log.info("Going to migrate the database to version %d", next_version)
                    self.__DATABASE_VERSION_MIGRATIONS[next_version](session)

                    # Store the update of the database version field in the same session so it gets stored
                    # with the rest of the changes on the database.
                    version_record = session.query(PrinterSettingRecord).get(self.settings.getDatabaseVersionName())
                    version_record.setting_value = next_version
            current_db_version = next_version


    ## @brief get installed database version on the printer
    #  @return database version. If the version not available then return 'None'
    def __getCurrentDatabaseVersion(self):
        # if the Settings table doesn't exist then create it, and fill it with a version 1, as we come from a time
        # before this table exists, which is version 0.
        if not engine.dialect.has_table(engine, PrinterSettingRecord.get_table_name()):
            PrinterSettingRecord.__table__.create(engine)
            with session_scope() as session:
                record = PrinterSettingRecord(name=self.settings.getDatabaseVersionName(), setting_value=1)
                session.add(record)

        try:
            with session_scope() as session:
                db_param_key = self.settings.getDatabaseVersionName()
                db_setting = session.query(PrinterSettingRecord).get(db_param_key)
                if db_setting is not None:
                    return db_setting.setting_value
                log.error("Error, no database version number found.")
        except:
            log.exception("Error while getting database version number.")

        # If we end up here, we could not get which database version number we have.
        # Do the only safe thing we can do. Nuke from orbit and start clean.
        Base.metadata.drop_all()
        self.__createInitialDatabase()
        return self.__getCurrentDatabaseVersion()

    ## Create the database
    #  And initialize database settings with the default value for the database version number.
    def __createInitialDatabase(self):
        Base.metadata.create_all()

        # Store are current database revision number, as we do not need any migrations on a fresh database.
        with session_scope() as session:
            record = PrinterSettingRecord(name=self.settings.getDatabaseVersionName(),
                                          setting_value=self.settings.getDatabaseVersionValue())
            session.add(record)
