import logging
from griffin import dbusif

from griffin.datatypes.printJob import PrintJob
import uuid
import dbus

from griffin.history.jobHistoryBackend import JobHistoryBackend

log = logging.getLogger(__name__.split(".")[-1])

## @brief Signature of a DBus method dealing with PrintJob
# This service will often deal with this datatype and thus it makes sense to centralize the definition
# It is a dictionary, in DBus represented as a list of key-value pairs, where each key is a string and the value a variant
PRINT_JOB_OBJECT_SIGNATURE = "a{sv}"


## @brief Interface to the history of PrintJobs on this machine
class JobHistoryService(dbusif.ServiceObject):
    def __init__(self):
        super().__init__("job_history")

        self._backend = JobHistoryBackend()

        ## Add callback to this job_history service which implements the call to the resetToFactorySettings
        dbusif.RemoteObject("system").addFactoryResetCallback("job_history", "resetSettings")

    ## @brief The callback function to register which needs to be called when a factory reset is issued
    #  @param reset_type One of the defined values from FactoryReset which would indicate a hard or soft reset
    @dbusif.method("s", "")
    def resetSettings(self, reset_type):
        log.debug("Removing all jobs, for reset_type {reset_type}".format(reset_type=reset_type))
        self._backend.removeAllJobs()

    ## @brief Add a DBus-packed PrintJob to the history
    # @param printJobDict, dict, representing the PrintJob to be added.
    # On the client side, use PrintJob.toDbus() to represent a PrintJob to this method
    @dbusif.method(in_signature=PRINT_JOB_OBJECT_SIGNATURE)
    def addPrintJob(self, printJobDict):
        log.debug("Adding new PrintJob: {job}".format(job=printJobDict))
        job = PrintJob.fromDict(printJobDict)
        self._addPrintJob(job)

    ## @brief add a native PrintJob to history
    # @param job, PrintJob tha printjob to be added
    def _addPrintJob(self, job):
        self._backend.addPrintJob(job)

    ## @brief Get a PrintJob by it UUID
    # @param uuid_str, str,  the PrintJob's UUID, represented as a str
    # @returns dict with the PrintJob with that UUID, represented as a DBus dictionary
    @dbusif.method("s", out_signature=PRINT_JOB_OBJECT_SIGNATURE)
    def getJobByUuid(self, uuid_str):
        _uuid = uuid.UUID(uuid_str)
        job = self._backend.getJobByUuid(_uuid)
        return dbus.Dictionary({}, signature="sv") if job is None else job.toDbus()

    ## @brief get all jobs (but with a size limit)
    # @param offset, int: the offset from the end of the job list to start returning jobs. 0 is include the latest job.
    # @param count, int the maximum number of jobs to return (e.g. do not just return 10000 jobs unless really asked for)
    @dbusif.method(in_signature="ii", out_signature="a"+PRINT_JOB_OBJECT_SIGNATURE)
    def getAllJobs(self, offset, count):
        return [job.toDbus() for job in self._backend.getAllJobs(offset, count)]