from griffin.interface.http.exposedItems.httpExposedObject import HttpExposedObject
from griffin.interface.http.flaskResponses import FlaskResponses
from griffin.datatypes import eventType

import flask
import datetime
from griffin import dbusif

import logging

log = logging.getLogger(__name__.split(".")[-1])


## @brief Expose a list of events to HTTP as JSON objects
class EventHistory(HttpExposedObject):

    ## How many Events are maximally shown by default
    __DEFAULT_EVENT_COUNT = 50

    def __init__(self, *args, **kwargs):
        kwargs.update({"allowed_request_methods": ["GET", "POST"] })
        super().__init__(*args, **kwargs)
        self.__event_history_service = dbusif.RemoteObject("event_history")

    ## @brief get a list of events in history.
    # These are sorted by latest first
    def get(self):
        count = flask.request.args.get("count", self.__DEFAULT_EVENT_COUNT, type=int)
        offset = flask.request.args.get("offset", 0, type=int)
        type_id = flask.request.args.get("type_id", 0, type=int)  # This acts as a filter

        # Every event is now a tuple containing: timestamp, type_id, list(parameters)
        if type_id:
            all_events = self.__event_history_service.getEventsByType(type_id, offset, count)
        else:
            all_events = self.__event_history_service.getEvents(offset, count)

        json_result = []
        for timestamp, event_id, parameters in all_events:
            event_type = eventType.event_by_id.get(event_id, None)
            if event_type:
                try:
                    message = event_type.message.format(*parameters)
                except IndexError:
                    message = "Formatting error ID: %x:%s:%s" % (event_id, event_type.message, parameters)
            else:
                message = "Unknown ID: %x" % (event_id)
            json_result.append({
                "time": datetime.datetime.utcfromtimestamp(timestamp).replace(microsecond=0).isoformat(),
                "type_id": event_id,
                "message": message,
                "parameters": parameters
            })

        return FlaskResponses.createSuccess(json=json_result)

    def post(self, received_data):
        if received_data.json is None and not isinstance(received_data.json, dict):
            return FlaskResponses.createFailed("no json data")
        if not "type_id" in received_data.json:
            return FlaskResponses.createFailed("no type_id in data")

        type_id = received_data.json.get("type_id")
        if not type_id in eventType.USER_PROVIDED_EVENTS:
            return FlaskResponses.createFailed(
                "Post not allowed for event type_id {}".format(type_id))

        if type_id == eventType.SYSTEM_MAINTENANCE.type_id:
            parameters = received_data.json.get("parameters")
            if not parameters or len(parameters) < 1:
                return FlaskResponses.createFailed("not enough parameters")

            action, mechanic = parameters[0],parameters[1] if len(parameters) > 1 else eventType.SYSTEM_MAINTENANCE_DEFAULT_MECHANIC
            if not action in eventType.SYSTEM_MAINTENANCE_ACTIONS:
                return FlaskResponses.createFailed("The action parameter is not one of the known Actions: {}".format(str(eventType.SYSTEM_MAINTENANCE_ACTIONS)))

            eventType.postEvent(eventType.SYSTEM_MAINTENANCE, [action, mechanic])

            return FlaskResponses.createSuccess("Action logged")
