from griffin.interface.http.exposedItems.httpExposedObject import HttpExposedObject
from griffin.interface.http.flaskResponses import FlaskResponses

from griffin import dbusif

import time

import logging

log = logging.getLogger(__name__.split(".")[-1])


# Handle the API for pre-heating of the bed.
class PreHeatBed(HttpExposedObject):
    ## Constraints on temperature
    __TEMPERATURE_MIN_MAX = [0, 115]
    ## Default timeout of 5 minutes
    __DEFAULT_TIMEOUT = 5 * 60
    ## Constraints on timeout: minimal 1 minute, maximal 1 hour
    __TIMEOUT_MIN_MAX = [60, 3600]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, allowed_request_methods=["GET", "PUT"], **kwargs)
        self.__printer_service = dbusif.RemoteObject("printer")
        PreHeatBed.__TEMPERATURE_MIN_MAX[1] = self.__printer_service.getMaxBedTemperature()

    def put(self, received_data):
        if received_data is None or "temperature" not in received_data:
            return FlaskResponses.createFailed("Temperature was not specified")
        temperature = int(received_data["temperature"])
        if not self.__isInRange(temperature, self.__TEMPERATURE_MIN_MAX):
            return FlaskResponses.createFailed("Invalid Temperature specified")
        if "timeout" not in received_data:
            timeout = self.__DEFAULT_TIMEOUT
        else:
            timeout = int(received_data["timeout"])
        if not self.__isInRange(timeout, self.__TIMEOUT_MIN_MAX):
            return FlaskResponses.createFailed("Invalid timeout specified")

        preheat_arguments = {
            "target": temperature,
            "timeout": timeout
        }

        ## Send abort to the pre-heat, in case it is running
        self.__printer_service.messageProcedure("PRE_HEAT_BED", "ABORT")
        # We sleep here to wait for the PRE_HEAT_BED procedure to be aborted.
        # This isn't the best thing to do in an API call, as we block the whole webserver. But doing it differently right now
        # would take a lot more time.
        time.sleep(0.2)

        if not self.__printer_service.startProcedure("PRE_HEAT_BED", preheat_arguments):
            return FlaskResponses.createFailed("Pre-heating of bed failed")
        if temperature > 0:
            return FlaskResponses.createCreated("Pre-heating of bed started")
        else:
            return FlaskResponses.createCreated("Pre-heating of bed stopped")

    ## @brief Compares if the value is in the range of minimum and maximum values (min max included)
    #  @param value The value to verify
    #  @param min_max The list of minimum value [0] and maximum value[1]
    #  @return Returns True if the value is within the range, otherwise False
    def __isInRange(self, value, min_max):
        return value >= min_max[0] and value <= min_max[1]

    ## Get the information about pre-heating.
    #  This returns if pre-heating is active or not. And what the remaining timeout is.
    #  (The current pre-heating temperature can already be read from the target bed temperature.
    # @return Flask response containing json data.
    def get(self):
        meta_data = self.__printer_service.getProcedureMetaData("PRE_HEAT_BED")
        data = {
            "active": bool(meta_data.get("ACTIVE", False))
        }
        if data["active"]:
            data["remaining"] = meta_data.get("REMAINING", 0.0)
        return FlaskResponses.createSuccess(json=data)
