from griffin.interface.http.exposedItems.httpExposedObject import HttpExposedObject
from griffin.interface.http.flaskResponses import FlaskResponses

import flask
import werkzeug
from griffin import dbusif
import os

import logging

log = logging.getLogger(__name__.split(".")[-1])


# PrintJob is a special case of an HttpExposedObject.
class PrintJob(HttpExposedObject):
    ## The location where the uploaded files will be saved (tmpfs)
    __STORAGE_PATH = "/tmp/"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__printer_service = dbusif.RemoteObject("printer")

    # This is a hack due to the inability of flask to delete routes.
    # We fake the correct behaviour by setting the deleted state to True if there is no active step in the print sequence.
    def isDeleted(self):
        if self.__printer_service.getProperty("state") != "printing":
            return True
        return False

    def post(self, received_data):
        # If the state is not idle, we cannot start a print job, because a print job is already running, the printer is not ready yet, or maintenance actions are done.
        if self.__printer_service.getProperty("state") != "idle":
            return FlaskResponses.createMethodNotAllowed()
        else:
            received_file = received_data.files.get("file")
            if not received_file:
                log.warning("No file received")
                return FlaskResponses.createFailed("No file received")

            log.info("Saving received file %s", received_file.filename)

            ## Make sure the filename is not trying to access location on FS where it should not be
            filename = werkzeug.secure_filename(received_file.filename)
            file_path = os.path.join(PrintJob.__STORAGE_PATH, filename)
            while os.path.isfile(file_path):
                filename = "_" + filename
                file_path = os.path.join(PrintJob.__STORAGE_PATH, filename)

            if received_data.json is not None and "jobname" in received_data.json:
                jobname = received_data.json["jobname"]
            else:
                jobname = werkzeug.secure_filename(received_file.filename)
                # Remove any possible extension
                file_types = self.__printer_service.getSupportedFileTypes()
                for ext in file_types:
                    if jobname.endswith(ext):
                        jobname = jobname[:-len(ext)]

            # Get the current user from the authentication controller.
            # This could be None in the rare case that authentication is disabled.
            user_data = self._server.getAuthenticationController().getAuthenticatedUserData()
            # user_data is either None or a dictionary with the key "application" and "user"

            try:
                received_file.save(file_path)
                log.info("File saved as %s", file_path)
                scheme = "upload://"
                print_arguments = {"jobname": jobname, "url": (scheme + file_path), "origin": "WEB_API"}

                if user_data is not None:
                    for key, value in user_data.items():
                        print_arguments["origin_%s" % (key)] = str(value)  # all values are strings but some non string values are being added to the user store such as the local flag
                if not self.__printer_service.startProcedure("PRINT", print_arguments):
                    log.warning("Print of %s could not be started", jobname)
                    os.remove(file_path)
                    return FlaskResponses.createFailed("Print job not started")

                log.info("Print of %s started", jobname)

                # The file not removed, as the printer service is using the file. The printer service will delete the file when it is done with it.
                metadata = self.__printer_service.getProcedureMetaData("PRINT")
                return FlaskResponses.createCreated(json={"message": "Print job started", "uuid": metadata.get("uuid")})
            except:
                return FlaskResponses.createFailed("Unable to write file %s", file_path)

    # Another hack due to the inability of flask to delete routes.
    # If there is no print job, the POST method -is- allowed.
    def handleHttpRequest(self):
        if flask.request.method == "POST" and self.isDeleted():
            return self.post(received_data = flask.request)
        return super().handleHttpRequest()
