from griffin.interface.http.exposedItems.httpExposedObject import HttpExposedObject
from griffin.interface.http.flaskResponses import FlaskResponses
from griffin.datatypes.printJob import PrintJob

import flask
from griffin import dbusif

import logging

log = logging.getLogger(__name__.split(".")[-1])


## @brief Expose a list of past, completed or aborted PrintJobs to HTTP as JSON objects
# Jobs can also be accessed directly by the UUID
class PrintJobHistory(HttpExposedObject):

    ## How many PrintJobs are maximally shown by default
    __DEFAULT_JOB_COUNT = 50

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__job_history_service = dbusif.RemoteObject("job_history")

    ## @brief Register sub-endpoint for direct query by uuid
    def register(self):
        super().register()
        self._server.add_url_rule(self.getAbsolutePath() + "/<uuid>", view_func=self.getJobByUuid, methods=["GET"])

    ## @brief get a list for jobs in history.
    # These are sorted by latest first
    def get(self):
        count = flask.request.args.get("count", self.__DEFAULT_JOB_COUNT, type=int)
        offset = flask.request.args.get("offset", 0, type=int)

        # getAllJobs returns a list of dictionaries. The values are DBus types under the hood, that mostly act like builtin types
        all_jobs_dbus = self.__job_history_service.getAllJobs(offset, count)

        # But the DBus types aren't really DBus types. Therefore, we first deserialize the dictionaries to a normal Python object
        all_jobs_native = [PrintJob.fromDict(job) for job in all_jobs_dbus]  # I tried a map instead of list comprehension, but flask.json.dumps won't grok it

        # Which are then serialized to JSON (which also uses a dict, as an implementation detail)
        all_jobs_json = [job.toJson() for job in all_jobs_native]
        return FlaskResponses.createSuccess(json=all_jobs_json)

    ## @brief get a job by its UUID
    # @param uuid, str, the str-representation of a PrintJob's UUID.
    # @returns a JSON object of the job with that UUID, if any such exists. Otherwise an empty list is returned.
    def getJobByUuid(self, uuid):
        job_dbus = self.__job_history_service.getJobByUuid(uuid)
        if not job_dbus:
            return FlaskResponses.createNotFound()
        job_native = PrintJob.fromDict(job_dbus)
        return FlaskResponses.createSuccess(json=job_native.toJson())
