import logging
from griffin import dbusif

log = logging.getLogger(__name__.split(".")[-1])


# A dbus interface for the http api because the display should never be a dbus service and there is data to
#  communicate directly between the Http api and display.
class MessageService(dbusif.ServiceObject):
    def __init__(self):
        super().__init__("message")

        self.__led_service = dbusif.RemoteObject("led")
        self.__display_message = ""
        self.__display_button_caption = ""
        self.__glow_frame_light = False

    # Set a new message on the display.
    # @param message: string, the message to display to the user.
    # @param button_caption: string, the text to display on the dismissal button.
    # @param initiate_glow: boolean, True if the frame lights need to glow when this message is on the screen.
    def setDisplayMessage(self, message, button_caption, *, initiate_glow=True):
        self.__display_message = message
        self.__display_button_caption = button_caption
        self.__glow_frame_light = initiate_glow

        self.__led_service.setMainLightingRuntimeFlag("message", initiate_glow)
        self.onDisplayMessageFromUser(message, button_caption)

    # Remove any message currently shown on the display.
    def clearDisplayMessage(self):
        self.__display_message = None
        self.__display_button_caption = None
        self.__glow_frame_light = False

        self.__led_service.setMainLightingRuntimeFlag("message", False)
        self.onClearMessageFromUser()

    # @brief This signal indicates a message and possibly button caption are available for an interested party.
    # For example, the display service may subscribe and display this on the screen.
    # @param message The message to be displayed
    # @param button_caption The button caption
    @dbusif.signal("ss")
    def onDisplayMessageFromUser(self, message, button_caption):
        log.info("onDisplayMessageFromUser(%s, %s)", message, button_caption)
        return True

    # @brief This signal indicates a message and possibly button caption are available for an interested party.
    # For example, the display service may subscribe and display this on the screen.
    @dbusif.signal("")
    def onClearMessageFromUser(self):
        log.info("onClearMessageFromUser()")
        return True

    # Get the current display message
    # @return a dictionary containing information about the currently displayed message.
    #       If no message is displayed it returns an empty dictionary.
    #       Else it contains the following keys:
    #           "message": string, is the message to be displayed to the user.
    #           "button_caption": string, the caption to be displayed on the dismiss button.
    #           "glow_frame_light": boolean, True if the frame lights need to glow while this message is displayed.
    @dbusif.method("", "a{sv}")
    def getMessageData(self):
        if self.__display_message is None:
            return {}
        return {
            "message": self.__display_message,
            "button_caption": self.__display_button_caption,
            "glow_frame_light": self.__glow_frame_light
        }

    # Dismiss the currently displayed message on the screen.
    # Called when the user presses the button to remove the message. And will emit the signal that the message
    # should be removed.
    @dbusif.method("", "")
    def dismissMessage(self):
        self.clearDisplayMessage()
