from ..hardware import abstractHardware
from griffin import thread
import queue


## Base class for all the led controllers.
#  A led controller uses system state and local information to decide what effects to display on a hardware led.
#  The main interface for the subclasses is the _queueEffect function, which is used to change the led behaviour.
class AbstractController:
    ## Create a new led controller.
    #  @param hardware: a subclass of griffin.led.hardware.abstract.Abstract that provides the hardware interface to output LEDs
    def __init__(self, hardware):
        assert isinstance(hardware, abstractHardware.AbstractHardware)
        self.__hardware = hardware
        self.__current_effect = None
        self.__effect_queue = queue.Queue()
        thread.Thread("LedController(%s)" % (self), self.__run).start()

    ## Queue a new effect. This effect will be executed as soon as possible.
    #  The previous effect will be aborted but might want to finish up.
    #  @param effect: A subclass of griffin.led.effects.abstractEffect.AbstractEffect
    def _queueEffect(self, effect):
        self.__effect_queue.put(effect)

    # @return boolean: True if there are effects in the queue, false if not.
    def _hasEffectInQueue(self):
        return self.__effect_queue.qsize() > 0

    ## Private run function,
    #  This is ran in a thread to handle the execution of new effects.
    def __run(self):
        while True:
            next_effect = self.__effect_queue.get()
            if self.__current_effect is not None:
                self.__current_effect.abort()
                self.__current_effect.join()
            self.__current_effect = next_effect
            self.__current_effect.setHardware(self.__hardware)
            self.__current_effect.start()
