from . import abstractController
from ..effects import staticEffect
from ..effects import fadeEffect
from ..effects import blinkEffect
from ..effects import glowEffect
from .. import colorTheme

from griffin.preferences import registryFile
from griffin.math import hsvColor
from griffin import dbusif
from griffin import timer

import random
import logging

log = logging.getLogger(__name__.split(".")[-1])


## The MainLighting controller
#  This LED controller has the following features:
#  * A HSV color can be set, and each of the HSV parameters can be set separately.
#  * It has a control mode, which defines the relation to the LEDs and the system state.
#  * It can be asked to blink once to notify the user.
#  * A non volatile maximum brightness can be configured.
#  * At boot up, the LEDs fade from off to the requested state.
class MainLightingController(abstractController.AbstractController):
    __MAIN_LIGHTING_CONTROLLER_PREFERENCES_FILE = "main_lighting.json"

    ## Mode to glow the printer on/off when the print is finished and the printer needs to be emptied.
    __MODE_FLAG_GLOW_WHEN_FINISHED = "glow_when_print_is_finished"
    ## Mode to have the lighting only on when printing, and off while idle.
    __MODE_FLAG_ON_WHEN_PRINTING = "on_when_printing"
    ## Party mode, cycle between colors.
    __MODE_FLAG_PARTY = "party"

    ## Runtime flag to glow the printer on/off during authentication request.
    RUNTIME_FLAG_AUTHENTICATING = "authenticating"
    ## Runtime flag to glow when we have a API driven message on the screen.
    RUNTIME_FLAG_MESSAGE = "message"

    __ALL_MODE_FLAGS = (__MODE_FLAG_GLOW_WHEN_FINISHED, __MODE_FLAG_ON_WHEN_PRINTING, __MODE_FLAG_PARTY)
    __ALL_RUNTIME_FLAGS = (RUNTIME_FLAG_AUTHENTICATING, RUNTIME_FLAG_MESSAGE)

    # Default value for each of the modes. A fresh setup forces the modes to these values.
    __DEFAULT_MODE_SETTINGS = {
        __MODE_FLAG_GLOW_WHEN_FINISHED: True,
        __MODE_FLAG_ON_WHEN_PRINTING: False,
        __MODE_FLAG_PARTY: False,
    }

    __STARTUP_FADE_TIME = 6.0
    __DEFAULT_GLOW_FREQUENCY = 0.1
    __DARK_COLOR_BRIGHTNESS_FACTOR = 0.3

    def __init__(self, hardware):
        super().__init__(hardware)

        # Configuration data (non volatile)
        self.__preferences = registryFile.RegistryFile(self.__MAIN_LIGHTING_CONTROLLER_PREFERENCES_FILE)
        for key in self.__ALL_MODE_FLAGS:
            if not self.__preferences.has(key):
                self.__preferences.setAsBoolean(key, self.__DEFAULT_MODE_SETTINGS[key])
        if not self.__preferences.has("user_brightness"):
            self.__preferences.setAsFloat("user_brightness", 100.0)

        # Runtime data (volatile)
        self.__main_color = colorTheme.WHITE.copy()
        self.__runtime_flags = {}

        printer = dbusif.RemoteObject("printer")
        self.__printer_state = printer.getProperty("state")
        self.__job_state = printer.getProperty("job_state")
        printer.connectSignal("propertyChanged", self.__onPrinterPropertyChanged)

        # Start up by fading from black to the main color in
        self._queueEffect(fadeEffect.FadeEffect(self.__getColor(), self.__STARTUP_FADE_TIME))

    ## @brief Resets to default values
    def resetToDefaultValues(self):
        for key in self.__ALL_MODE_FLAGS:
            self.__preferences.setAsBoolean(key, self.__DEFAULT_MODE_SETTINGS[key])
        self.__preferences.setAsFloat("user_brightness", 100.0)
        self.__preferences.forceSave()

    def setMainColorHue(self, hue):
        self.__main_color.hue = hue
        self.__update()

    def setMainColorSaturation(self, saturation):
        self.__main_color.saturation = saturation
        self.__update()

    def setMainColorBrightness(self, brightness):
        self.__main_color.value = brightness
        self.__update()

    def getMainColorHue(self):
        return self.__main_color.hue

    def getMainColorSaturation(self):
        return self.__main_color.saturation

    def getMainColorBrightness(self):
        return self.__main_color.value

    def blink(self, frequency=1.0, count=1):
        self._queueEffect(blinkEffect.BlinkEffect(self.__getColor(), frequency, count))

    ## Set a boolean mode flag for the main lighting controller.
    #  @param flag_name: string, from the __ALL_MODE_FLAGS list.
    #  @param new_value: bool, True if this mode needs to be enabled, false if not.
    def setModeFlag(self, flag_name, new_value):
        assert flag_name in self.__ALL_MODE_FLAGS
        self.__preferences.setAsBoolean(flag_name, new_value)
        self.__update()

    ## Set a boolean mode flag for the main lighting controller.
    #  @param flag_name: string, from the __ALL_MODE_FLAGS list.
    #  @return bool: True if this mode is enabled, false if not.
    def getModeFlag(self, flag_name):
        assert flag_name in self.__ALL_MODE_FLAGS
        return self.__preferences.getAsBoolean(flag_name)

    ## Set a boolean mode flag for the main lighting controller.
    #  @param flag_name: string, from the __ALL_RUNTIME_FLAGS list.
    #  @param new_value: bool, True if this mode needs to be enabled, false if not.
    def setRuntimeFlag(self, flag_name, new_value):
        assert flag_name in self.__ALL_RUNTIME_FLAGS
        self.__runtime_flags[flag_name] = new_value
        self.__update()

    ## Set a boolean mode flag for the main lighting controller.
    #  @param flag_name: string, from the __ALL_RUNTIME_FLAGS list.
    #  @return bool: True if this mode is enabled, false if not.
    def getRuntimeFlag(self, flag_name):
        assert flag_name in self.__ALL_RUNTIME_FLAGS
        return self.__runtime_flags.get(flag_name, False)

    # Set the user configured brightness
    # @param user_brightness: Float in the range 0 to 100
    def setUserBrightness(self, user_brightness):
        user_brightness = max(0.0, min(user_brightness, 100.0))
        self.__preferences.setAsFloat("user_brightness", user_brightness)
        self.__update()

    # @return user_brightness: Float in the range 0 to 100
    def getUserBrightness(self):
        return self.__preferences.getAsFloat("user_brightness")

    # Function called by the onPropertyChanged signal of the printer dbus remoteobject.
    # @param property_key, key of the changed property
    # @param value, new value of the property
    def __onPrinterPropertyChanged(self, property_key, value):
        # Check if one of our monitored properties change and
        if property_key == "state":
            self.__printer_state = value
            self.__update()
        if property_key == "job_state":
            self.__job_state = value
            self.__update()

    ## Update the leds according to the state of the printer.
    def __update(self):
        if self.getModeFlag(self.__MODE_FLAG_PARTY):
            if not self._hasEffectInQueue():
                self._queueEffect(fadeEffect.FadeEffect(hsvColor.HsvColor(random.uniform(0, 360), 100, self.__preferences.getAsFloat("user_brightness")), 2.0))
                timer.Timer("PARTY!", 2.0, self.__update).start()
        elif self.getRuntimeFlag(self.RUNTIME_FLAG_AUTHENTICATING) or self.getRuntimeFlag(self.RUNTIME_FLAG_MESSAGE):
            self._queueEffect(glowEffect.GlowEffect(self.__getColor(), self.__getDarkColor(), self.__DEFAULT_GLOW_FREQUENCY))
        elif self.__stateIsWaitingForCleanup() and self.getModeFlag(self.__MODE_FLAG_GLOW_WHEN_FINISHED):
            self._queueEffect(glowEffect.GlowEffect(self.__getColor(), self.__getDarkColor(), self.__DEFAULT_GLOW_FREQUENCY))
        elif self.__stateIsPrinting() or not self.getModeFlag(self.__MODE_FLAG_ON_WHEN_PRINTING):
            self._queueEffect(staticEffect.StaticEffect(self.__getColor()))
        else:
            self._queueEffect(staticEffect.StaticEffect(colorTheme.BLACK))

    def __stateIsPrinting(self):
        return self.__printer_state == "printing"

    def __stateIsWaitingForCleanup(self):
        return self.__printer_state == "printing" and (self.__job_state == "wait_cleanup" or self.__job_state == "none")

    ## Get the current color requested by the user and limited by the configured maximum brightness.
    #  @return HsvColor object.
    def __getColor(self):
        return hsvColor.HsvColor(self.__main_color.hue, self.__main_color.saturation, self.__main_color.value * self.__preferences.getAsFloat("user_brightness") / 100.0)

    # @return HsvColor object which is a darker version of the user set color returned by __getColor.
    def __getDarkColor(self):
        color = self.__getColor()
        color.value *= self.__DARK_COLOR_BRIGHTNESS_FACTOR
        return color