import math
import colorsys
from griffin.math import hsvColor
import logging
from .. import colorTheme

log = logging.getLogger(__name__.split(".")[-1])


## Abstract hardware LED class
#  This serves as a base class for all LED hardware types.
#  Subclasses need to implement setRGB to set the hardware output, and may implement startHardwareEffect/endHardwareEffect to have hardware driven effects.
class AbstractHardware:
    def __init__(self):
        self.__current_color = colorTheme.BLACK

    ## Set the output to the HSV color.
    #  this function applies brightness correction and then calculates the RGB values to output to the LEDs.
    #  This function needs to be called from the Effect objects to output the color to the hardware.
    #  @param hsv: HsvColor instance.
    def setColor(self, hsv):
        assert isinstance(hsv, hsvColor.HsvColor)

        self.__current_color = hsv.copy()

        # First apply brightness correction, so we get a more linear brightness response from our LEDs.
        value = self.__correctBrightness(hsv.value)

        # Next convert HSV values to RGB values
        r, g, b = colorsys.hsv_to_rgb(hsv.hue / 360.0, hsv.saturation / 100.0, value / 100.0)

        # Give the RGB values to the setRGB function, the subclass implementation only needs to implement setRGB to work.
        self.setRGB(r, g, b)

    ## Get the last color set to this hardware output.
    #  @return HsvColor object
    def getColor(self):
        return self.__current_color.copy()

    ## Apply brightness correction. Our eyes are non-linear, so we need to apply a curve to make them look more linear.
    #  This gives a better response when fading from black to white.
    #  @param brightness: Float in the range 0 to 100
    #  @return: Float in the range 0 to 100
    def __correctBrightness(self, brightness):
        # f(x) ~ (exp(sin(x)) - 1/e) * 100/(e-1/e)
        # -1/e = -0.36787944117144233, and 100/(e - 1/e) = 42.54590641196608 are constants, and can be pre-calculated.
        # pi/2 is there to make sure the sinusoid is applied to the right range
        max_output = 100.0 # Our brightness is in percentage.
        return (math.exp(math.sin((brightness - max_output / 2) * math.pi / max_output)) - 0.36787944117144233) * 42.54590641196608

    ## Set the leds to a fixed color.
    #  This needs to be implemented by every subclass.
    #  @param red: float - Red output in range 0.0% to 100.0%
    #  @param green: float - Green output in range 0.0% to 100.0%
    #  @param blue: float - Blue output in range 0.0% to 100.0%
    def setRGB(self, red, green, blue):
        raise NotImplementedError

    ## Try to set a hardware controlled effect.
    #  This allows specific hardware implementations to handle effects without
    #  using the effect classes.
    #  @param effect_name: String, name of the effect to be applied.
    #  @param parameters: dictionary containing parameters for the effect. This depends on the effect name.
    #  @return bool, true if the hardware is now executing this effect. False if this hardware implementation cannot handle this effect.
    def startHardwareEffect(self, effect_name, parameters):
        return False

    ## Should be called after startHardwareEffect to end this hardware effect.
    #  @param effect_name: Same effect name as given to startHardwareEffect
    def endHardwareEffect(self, effect_name):
        pass
