from . import abstractHardware
from .. import colorTheme
from griffin.math import hsvColor
from griffin import dbusif
import logging
import time
import math

log = logging.getLogger(__name__.split(".")[-1])


## This class controls the leds that are in the print head.
class HeadSlotHardware(abstractHardware.AbstractHardware):
    def __init__(self, slot_index):
        super().__init__()
        self.__slot_index = slot_index
        self.__printer = dbusif.RemoteObject("printer")
        self.__color_balance = [1.0, 0.3, 0.275]
        self.__last_rgb_output = None

    ## Set the leds to a fixed color.
    #  This needs to be implemented by every subclass.
    #  @param red: float - Red output in range 0.0 to 1.0
    #  @param green: float - Green output in range 0.0 to 1.0
    #  @param blue: float - Blue output in range 0.0 to 1.0
    def setRGB(self, red, green, blue):
        # Apply some white balancing
        red *= self.__color_balance[0]
        green *= self.__color_balance[1]
        blue *= self.__color_balance[2]

        rgb = (int(red * 255), int(green * 255), int(blue * 255))
        if self.__last_rgb_output != rgb:
            self.__last_rgb_output = rgb
            while not self.__printer.debugSendGCodeNoWait("M143 T%d P0 r%d g%d b%d" % ((self.__slot_index,) + rgb)):
                time.sleep(0.1)

    ## Try to set a hardware controlled effect.
    #  This allows specific hardware implementations to handle effects without
    #  using the effect classes.
    #  @param effect_name: String, name of the effect to be applied.
    #  @param parameters: dictionary containing parameters for the effect. This depends on the effect name.
    #  @return bool, true if the hardware is now executing this effect. False if this hardware implementation cannot handle this effect.
    def startHardwareEffect(self, effect_name, parameters):
        if effect_name == "glow" and parameters["first_color"] == colorTheme.BLACK:
            # Hardware glow effect, currently only working for glowing from black to a color.

            # First we set the actual color, before enabling the glow.
            self.setColor(parameters["second_color"])

            # Marlin code calls glow "pulse", but glow is the better name.
            # M142 Px -> P0: disable pulse
            # M142 Px -> P1: enable pulse dual color
            # M142 Px -> P2: enable pulse single color
            # M142 Bx -> x: integer value of milliseconds between each pulse step, there are 255 pulse steps from off to on.

            step_size = math.ceil(parameters["cycle_time"] * 1000.0 / 255.0 / 2.0)
            while not self.__printer.debugSendGCodeNoWait("M143 T%d P%d B%d" % (self.__slot_index, 2, step_size)):
                time.sleep(0.1)
            return True
        return False

    def endHardwareEffect(self, effect_name):
        if effect_name == "glow":
            while not self.__printer.debugSendGCodeNoWait("M143 T%d P%d B3" % (self.__slot_index, 0)):
                time.sleep(0.1)
