from . import abstractHardware
import logging
import os

from .drivers.linuxLEDClass import LinuxLEDClass
from .drivers.marlinLEDDebug import MarlinLEDDebug

log = logging.getLogger(__name__.split(".")[-1])


## This class controls the LED strip that is in the case of the printer
class MainLightingHardware(abstractHardware.AbstractHardware):
    __LEDS = ["red", "green", "blue", "white"]
    def __init__(self):
        super().__init__()

        # HACK ideally, we will ask udev or the device tree what hardware
        # is available, but we cannot do this yet. So instead, we check
        # if the new path is available, and if not fall back to the old
        # file node. But even this new path should come from the machine itself
        # rather then hardcoding it here.
        self.__leds = dict()
        if os.path.isdir("/sys/class/leds/pca963x:red:cabin_light"):
            for led in MainLightingHardware.__LEDS:
                self.__leds[led] = LinuxLEDClass("pca963x", led, "cabin_light")
            self.__setBrightness = self._linuxLEDsetBrightness
        else:
            self.__leds["rgbw"] = MarlinLEDDebug()
            self.__setBrightness = self._marlinLEDsetBrightness

        # Color balance data should come from the hardware definition as it is dependent of chosen LED.
        self.__color_balance = {"red": 1.0, "blue": 0.6, "green": 0.6, "white": 1.0}

    ## Helper function to set the brightness via the linux led class
    #  @param red: float - Red output in range 0.0 to 1.0
    #  @param green: float - Green output in range 0.0 to 1.0
    #  @param blue: float - Blue output in range 0.0 to 1.0
    #  @param white: float - White output in range 0.0 to 1.0
    def _linuxLEDsetBrightness(self, red, green, blue, white):
        self.__leds["red"].setBrightness(red)
        self.__leds["green"].setBrightness(green)
        self.__leds["blue"].setBrightness(blue)
        self.__leds["white"].setBrightness(white)

    ## Helper function to set the brightness via marlin
    #  @param red: float - Red output in range 0.0 to 1.0
    #  @param green: float - Green output in range 0.0 to 1.0
    #  @param blue: float - Blue output in range 0.0 to 1.0
    #  @param white: float - White output in range 0.0 to 1.0
    def _marlinLEDsetBrightness(self, red, green, blue, white):
        self.__leds["rgbw"].setBrightness(red, green, blue, white)

    ## Set the main cabin lights to a color.
    #  @param red: float - Red output in range 0.0 to 1.0
    #  @param green: float - Green output in range 0.0 to 1.0
    #  @param blue: float - Blue output in range 0.0 to 1.0
    def setRGB(self, red, green, blue):
        white = min(red, green, blue)
        red -= white
        green -= white
        blue -= white

        red *= self.__color_balance["red"]
        green *= self.__color_balance["green"]
        blue *= self.__color_balance["blue"]
        white *= self.__color_balance["white"]

        self.__setBrightness(red, green, blue, white)
