import logging
import re
import time
from collections import namedtuple
from datetime import datetime

from griffin import signal
from griffin import thread
from griffin.math.crc16 import Crc16
from griffin.math.movingWindowStatistics import MovingWindowStatistics
from griffin.preferences.registry import Registry
from .marlinLowLevelCommunication import MarlinLowLevelCommunication
from .networkPacket import NetworkPacket

log = logging.getLogger(__name__.split(".")[-1])


## MarlinProtocolHandler
#  This object handles all the nasty complex details of dealing with the Marlin protocol.
#  It sends data to Marlin, and reads back the results. It keeps track of which reply belongs to which results
#  and it accounts for resend requests.
#  Further details can be found at docs/Serial_Protocol.odt
class MarlinProtocolHandler:
    # Number of moves that marlin can put in the planner before a motion move will block the communications.
    __MARLIN_PLANNER_BUFFER_SIZE = 15
    # Number of lines marlin can receive in its buffer without causing buffer overflows.
    __MARLIN_RECEIVE_BUFFER_SIZE = 5    # See comments in Marlin: we can expect that a buffer of 5 packets will be handled by Marlin just fine
    # Time duration after which to restart communication when a timeout was discovered.
    __CONNECTION_TIMEOUT_RESTART = 1.0
    # Receive timeout to determine if Marlin is still sending in seconds
    __RECEIVE_TIMEOUT = 10.0

    # To calculate P-value statistics: what should the window size be?
    __P_VALUE_STATS_WINDOW = 50

    # Message tuple index definitions for humans to read and understand
    __MESSAGE_TUPLE = namedtuple('MessageTuple', 'sequence_number, message, callback_function')

    # Marlin Protocol Control Indicators
    __MARLIN_OK = "ok"
    __MARLIN_REJECT = "rej"

    # Statistics counters
    __READ_ERRORS = "read_errors"
    __MARLIN_ERRORS = "marlin_errors"
    __MARLIN_WARNINGS = "marlin_warnings"
    __NO_ACK_WARNINGS = "no_ack_warning"
    __RECONNECTS = "reconnects"
    __RESENDS = "resends"
    __TIMEOUTS = "timeouts"
    __COMMUNICATION_STALLS = "communication_stalls"
    __SEQUENCE_NUMBER_ADJUSTMENTS = "sequence_number_adjustments"
    # This counter is increased if the planner buffer is reported empty, while the queue has data waiting.
    # Note: It will increase during the start of a print. But with perfect buffering it should not increase during a single extrusion print. (Might increase when switching nozzles)
    __PLANNER_BUFFER_EMPTY = "planner_buffer_empty"
    __READ_SUCCESS = "read_success"
    __WRITE_SUCCESS = "write_success"

    def __init__(self, controller, port_name, queue):
        self.__controller = controller
        self.__queue = queue
        self.__idle_handler = None
        self.__comm = MarlinLowLevelCommunication(port_name)
        self.__communication_thread = thread.Thread("Serial communication thread", self.__communicationThreadFunction)

        self.__acks = 0
        self.__statistics = MovingWindowStatistics(self.__P_VALUE_STATS_WINDOW)  # Rolling window of ... samples

        self.__planner_buffer_space = MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE
        self.__previous_planner_buffer_space = MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE
        self.__connected = False
        self.__received_data = []
        self.__partially_received_line = ""
        self.__last_received_timestamp = time.monotonic() - MarlinProtocolHandler.__CONNECTION_TIMEOUT_RESTART      # Subtract the offset for faster connection on startup.
        self.__resend_number = None
        self.__resend_data = []
        ## Current sequence number for the packet
        self.__current_sequence_number = 0
        # A list of all messages that have been sent, but still need a reply. Each entry contains a MessageTuple with (sequence_number, message, callback_function)
        self.__send_messages_that_still_need_ack = []

        # Hack to prevent problems with the communications when probing. When probing the bed the communication is stalled for a long time. So we need to disable the timeout.
        self.__is_probing = False

        self.onConnect = signal.Signal(type=signal.Signal.ASync)
        self.onDisconnect = signal.Signal(type=signal.Signal.ASync)
        self.onError = signal.Signal(type=signal.Signal.ASync)
        self.onWarning = signal.Signal(type=signal.Signal.ASync)

        # Error counter(s)
        self.__errors = Registry()
        self.__errors.setAsInt(self.__READ_ERRORS, 0)
        self.__errors.setAsInt(self.__MARLIN_ERRORS, 0)
        self.__errors.setAsInt(self.__MARLIN_WARNINGS, 0)
        self.__errors.setAsInt(self.__NO_ACK_WARNINGS, 0)
        self.__errors.setAsInt(self.__RECONNECTS, 0)
        self.__errors.setAsInt(self.__RESENDS, 0)
        self.__errors.setAsInt(self.__TIMEOUTS, 0)
        self.__errors.setAsInt(self.__COMMUNICATION_STALLS, 0)
        self.__errors.setAsInt(self.__SEQUENCE_NUMBER_ADJUSTMENTS, 0)
        self.__errors.setAsInt(self.__PLANNER_BUFFER_EMPTY, 0)

        self.__success = Registry()
        self.__success.setAsInt(self.__READ_SUCCESS, 0)
        self.__success.setAsInt(self.__WRITE_SUCCESS, 0)

        self.__debug_logfile_name = "/tmp/debug_%s_%s.log" % (controller.getSystemService().getMachineName(), time.strftime("%Y-%m-%d_%H.%M.%S"))

    # @return a dictionary reporting how often various errors occurred and calculates a total number of errors+warnings
    def getErrorCounts(self):
        counts = self.__errors.getAll()
        counts["total"] = sum(counts.values())
        return counts

    # @return a dictionary reporting how often various success occurred
    def getSuccessCounts(self):
        return self.__success.getAll()

    # Set the "idle command handler", which is a callback handler that is called whenever the communication needs
    # a command to send to Marlin to keep the communication active. This handler is called when the planner buffer is full and there is no instant gcode to send.
    # This handler function should return a tuple containing the GCode line and a reply handler for this line of gcode.
    # @param: handler, a function which will be called when there is no command to send to Marlin.
    def setIdleCommandHandler(self, handler):
        self.__idle_handler = handler

    # Start the communication with Marlin. This is not started in the __init__ so the owner of this object has time to register callbacks.
    def start(self):
        self.__communication_thread.start()

    # @return: bool, True if we are properly connected with Marlin and communication is possible. False if we are still trying to connect.
    def isConnected(self):
        return self.__connected

    # @return: The number of movements currently still planned to be executed by Marlin. When this is 0 the printer is standing still.
    def getPlannedMovementCount(self):
        return MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE - self.__planner_buffer_space

    def __communicationThreadFunction(self):
        while True:
            while not self.__comm.isOpen():
                self.__comm.open()
            self.__communicationTryReceive()
            self.__communicationTrySend()

    def __communicationTryReceive(self):
        try:
            # Serial.readline can return 3 different things:
            # A zero length string: No data received within the timeout.
            # A string with a newline at the end: A full line was received.
            # A string without a newline at the end: Timeout happened during receiving of a line.
            # The timeout is set to 100ms. If a timeout happens during a partial line, we need to stitch
            # this together later on, as the rest of the code assumes that it gets full lines of text.
            line = self.__comm.readline()
            self.__success.setAsInt(self.__READ_SUCCESS, self.__success.getAsInt(self.__READ_SUCCESS) + 1)# Increase read success count
        except:
            # If we get an exception during the serial read, we better close the serial port.
            self.__errors.setAsInt(self.__READ_ERRORS, self.__errors.getAsInt(self.__READ_ERRORS) + 1)
            log.exception("Exception %i during serial read, closing serial port.", self.__errors.getAsInt(self.__READ_ERRORS))
            self.__comm.close()
            self.__connected = False
            self.onDisconnect.emit()
            line = ""

        if line[-1:] == '\n':
            line = self.__partially_received_line + line
            self.__partially_received_line = ""
            if "||" in line:
                remainder, debug_string = line.split("||")
                # Remainder of the line, before the '||' marker is for future processing.
                self.__partially_received_line += remainder

                try:
                    with open(self.__debug_logfile_name, "a") as f:
                        now = datetime.utcnow()
                        f.write("%s %s" % (now.strftime("%Y-%m-%d %H:%M:%S.%f"), debug_string))
                except OSError as ex:
                    log.warning("Error writing debug log file '%s'. Details: ", self.__debug_logfile_name, ex)

            else:
                self.__processIncomingData(line.strip())
        else:
            if len(line) > 0:  # Timeout during a line, store the data for later use.
                self.__partially_received_line += line
            self.__processReceiveTimeout(time.monotonic() - self.__last_received_timestamp)

    def __processIncomingData(self, line):
        if line.startswith("LOG:"):
            log.info(line[4:])
            return
        log.debug("Recv: %s", line)

        if "Error:" in line or "ERROR" in line:
            self.__errors.setAsInt(self.__MARLIN_ERRORS, self.__errors.getAsInt(self.__MARLIN_ERRORS) + 1)
            log.warning("Got error line %i from printer: %s", self.__errors.getAsInt(self.__MARLIN_ERRORS), line)
            self.onError.emit(line.split(":")[1].strip(), line)

        if "Warning:" in line or "WARNING:" in line:
            self.__errors.setAsInt(self.__MARLIN_WARNINGS, self.__errors.getAsInt(self.__MARLIN_WARNINGS) + 1)
            log.warning("Got warning line %i from printer: %s", self.__errors.getAsInt(self.__MARLIN_WARNINGS), line)
            self.onWarning.emit(line.split(":")[1].strip(), line)

        # The main thread loop does not account for being connected or not, only for an open communications port,
        # so it's possible to loose the connection and reconnect because indirectly an M105 might be send (__processReceiveTimeout)
        # trying to re-establish a connection (without messing with the queue)
        if not self.__connected:
            if MarlinProtocolHandler.__MARLIN_OK in line or MarlinProtocolHandler.__MARLIN_REJECT in line:
                self.__errors.setAsInt(self.__RECONNECTS, self.__errors.getAsInt(self.__RECONNECTS) + 1)

                # Reset any errors that might have happened already. If the error is still there, this will trigger a new error message which we can capture properly.
                # Else the error message might have been sent before we were listening.
                # If the command is acknowledged, then emit the signal for connection re-established
                self.__queue.putInstant("M999", self.__emitConnect, allow_block=False)
                self.__connected = True # This allows for sending (existing packets in queue's)

                # Only update the receive_data_time when we received decent data. Otherwise log messages could hang the system.
                self.__last_received_timestamp = time.monotonic()

        else:
            self.__received_data.append(line)
            if MarlinProtocolHandler.__MARLIN_OK in line or MarlinProtocolHandler.__MARLIN_REJECT in line:
                self.__processIncomingAck()

                # Only update the receive_data_time when we received decent data. Otherwise log messages could hang the system.
                self.__last_received_timestamp = time.monotonic()

    ## @brief Emit the connect signal, this indicates we are ready to send data now.
    #  @param data The data coming from the M999 command (can be ignored)
    def __emitConnect(self, data):
        log.info("M999 handled, communication with Marlin established")
        self.onConnect.emit()

    def __processIncomingAck(self):
        ## Marlin returns a N <number> for the sequence number (for either OK or REJ)
        #  and P <number> to indicate the room left in the planbuffer (only for OK)
        self.__acks += 1

        receive_index = None
        reject_sequence_number = None
        log.debug("ProcessACK: %s", str(self.__received_data))
        for line in self.__received_data:

            if line.startswith(MarlinProtocolHandler.__MARLIN_REJECT) and self.__isVerifiedCrc(line):
                self.__errors.setAsInt(self.__RESENDS, self.__errors.getAsInt(self.__RESENDS) + 1)
                reject_sequence_number = self.__getValue("N", line)

            if line.startswith(MarlinProtocolHandler.__MARLIN_OK) and self.__isVerifiedCrc(line):
                n_value = self.__getValue("N", line)
                receive_index = self.__findMessageWaitingForAckIndexBySequenceNumber(n_value)

                p_value = self.__getValue("P", line)
                if p_value is not None and 0 <= p_value <= MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE:
                    if self.__previous_planner_buffer_space != MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE and p_value == MarlinProtocolHandler.__MARLIN_PLANNER_BUFFER_SIZE and self.__queue.getNormalQueueSize() > 0:
                        self.__errors.setAsInt(self.__PLANNER_BUFFER_EMPTY, self.__errors.getAsInt(self.__PLANNER_BUFFER_EMPTY) + 1)
                    self.__previous_planner_buffer_space = self.__planner_buffer_space
                    self.__planner_buffer_space = p_value

                # Some logging on p_value
                p_statistics = self.__statistics.process(p_value)
                if self.__acks % self.__P_VALUE_STATS_WINDOW == 0:
                    log.debug("p_value statistics (window={n}): min: {stat[1]}, max: {stat[2]}, avg: {stat[0]}".format(stat=p_statistics, n=self.__P_VALUE_STATS_WINDOW))

        # Both ack and nack will imply that earlier sent packets have been handled ok, but that in that case the ack response has gone haywire
        if reject_sequence_number is None and receive_index is not None:
            self.__acknowledgePacketsReceived(receive_index)
            # Call the callback function for the current ack.
            if self.__send_messages_that_still_need_ack[0].callback_function is not None:
                self.__send_messages_that_still_need_ack[0].callback_function("\n".join(self.__received_data))

            # Remove the entry from the list of messages that still need an ack.
            self.__send_messages_that_still_need_ack = self.__send_messages_that_still_need_ack[1:]
        elif reject_sequence_number is not None:
            nack_index = self.__findMessageWaitingForAckIndexBySequenceNumber(reject_sequence_number)
            if nack_index is not None and nack_index >= 1:
                # If the nack index is not the first one, this implies ack of all the messages before
                self.__acknowledgePacketsReceived(nack_index)
            self.__handleResend(reject_sequence_number)
        self.__received_data = []

    ## @brief Gets the index for the given sequence number of the message waiting for an ack
    #  @param sequence_number The sequence number to find in the ack-message list
    #  @return Returns the index or None if the sequence number could not be found
    def __findMessageWaitingForAckIndexBySequenceNumber(self, sequence_number):
        for index, message in enumerate(self.__send_messages_that_still_need_ack):
            if message.sequence_number == sequence_number:
                return index
        return None

    ## @brief Acknowledges all waiting messages up to the one specified and removes them from the waiting list
    #  @param The index of the packet in the queue of the messages waiting for ack that is to become the first.
    def __acknowledgePacketsReceived(self, index):
        log.debug("Processing acknowledged packets up to index %d, messages waiting for ack: %r", index, repr(self.__send_messages_that_still_need_ack))
        # If the receive_index is not position 0, then we missed an ack for earlier messages. Be sure to call the callback handlers for these messages.
        for n in range(index):
            message_tuple = self.__send_messages_that_still_need_ack[n]
            self.__errors.setAsInt(self.__NO_ACK_WARNINGS, self.__errors.getAsInt(self.__NO_ACK_WARNINGS) + 1)
            log.warning("Did not see an ack (%i) for line: %d: %s", self.__errors.getAsInt(self.__NO_ACK_WARNINGS), message_tuple.sequence_number, message_tuple.message)
            if message_tuple.callback_function is not None:
                message_tuple.callback_function("")
        # Remove the entries from the list of messages that still need an ack.
        self.__send_messages_that_still_need_ack = self.__send_messages_that_still_need_ack[index:]

    def __handleResend(self, resend_nr):
        # Remember the first resend request number (most likely there will arrive more requests).
        if self.__resend_number is None:
            log.debug("Marlin request to resend: %d %s", resend_nr, repr(self.__send_messages_that_still_need_ack))
            self.__resend_number = resend_nr

    def __processReceiveTimeout(self, timeout):
        if not self.__connected:
            if timeout >= MarlinProtocolHandler.__CONNECTION_TIMEOUT_RESTART:
                self.__errors.setAsInt(self.__TIMEOUTS, self.__errors.getAsInt(self.__TIMEOUTS) + 1)
                self.__last_received_timestamp = time.monotonic() # Reset the last_received_time stamp here to prevent a flood of M105 requests

                # On each attempt, reconfigure the serial port.
                # This works around the kernel bug where serial settings are not properly applied.
                self.__errors.setAsInt(self.__RECONNECTS, self.__errors.getAsInt(self.__RECONNECTS) + 1)
                self.__comm.refreshPortConfiguration()

                # This one can not use the normal buffering system as this is in the connection phase
                # Also used during possible re-connections
                log.debug("Sending M105 during (re-)connection phase.")
                self.__write("M105") # Send directly and wait for Marlin response (likely to be ack or nack with the expected sequence number)
        elif self.__resend_number is not None:
            # Data transmission was stopped at receiving the first resend request. Wait for all lower level communications to settle down.
            # Not an optimal implementation, but with the low number of errors we see we don't care for now. 
            if timeout > 0.1:
                index = self.__findMessageWaitingForAckIndexBySequenceNumber(self.__resend_number)
                number_of_messages_waiting_for_ack = len(self.__send_messages_that_still_need_ack)

                if index is None and number_of_messages_waiting_for_ack > 0:
                    # First check: if the resend_number equals the last message's sequence + 1, it's ok; meaning:
                    # all messages would have gotten an ack, but all the ack responses got lost or garbled
                    next_sequence_number = NetworkPacket.calculateNextSequenceNumber(self.__send_messages_that_still_need_ack[-1].sequence_number)
                    if self.__resend_number == next_sequence_number:
                        log.debug("All messages have been acknowledged!")
                        self.__send_messages_that_still_need_ack = []
                        number_of_messages_waiting_for_ack = 0
                    else:
                        ## This case should not be happening, hence the warning messages
                        self.__errors.setAsInt(self.__SEQUENCE_NUMBER_ADJUSTMENTS, self.__errors.getAsInt(self.__SEQUENCE_NUMBER_ADJUSTMENTS) + 1)
                        log.warning("Resend data request for sequence %d, but not found, before adjusting %s ", self.__resend_number, repr(self.__send_messages_that_still_need_ack))
                        # Iterate over transmit queue and adjust the resend number for each message
                        for index in range(number_of_messages_waiting_for_ack):
                            line_to_ack = self.__send_messages_that_still_need_ack[index]
                            self.__send_messages_that_still_need_ack[index] = self.__MESSAGE_TUPLE(NetworkPacket.calculateNextSequenceNumber(self.__resend_number, index), line_to_ack.message, line_to_ack.callback_function)
                        log.warning("Resend data request for sequence %d, but not found, after adjusting %s ", self.__resend_number, repr(self.__send_messages_that_still_need_ack))
                        # Correct the current sequence number base on resend, in case there is a problem with wrong (n)ack value
                        self.__current_sequence_number = NetworkPacket.calculateNextSequenceNumber(self.__resend_number, number_of_messages_waiting_for_ack)
                    index = 0  # Restart sending the first

                log.debug("Planning resend data starting with sequence # %d; # of packets: %d; next new packet with sequence %d", index, number_of_messages_waiting_for_ack, self.__current_sequence_number)
                self.__resend_data += self.__send_messages_that_still_need_ack[index:]
                self.__send_messages_that_still_need_ack = self.__send_messages_that_still_need_ack[:index]
                self.__resend_number = None
        else:
            if timeout > MarlinProtocolHandler.__RECEIVE_TIMEOUT and not self.__is_probing:
                self.__last_received_timestamp = time.monotonic()
                self.__errors.setAsInt(self.__COMMUNICATION_STALLS, self.__errors.getAsInt(self.__COMMUNICATION_STALLS) + 1)
                if self.__send_messages_that_still_need_ack:
                    message_type = self.__send_messages_that_still_need_ack[0]
                    log.debug("Sending %d - %s because communication looks stalled." % (message_type.sequence_number, message_type.message))
                    # Try the message that we still have no ack for in an attempt to recover the communication.
                    self.__writeSequence(message_type.sequence_number, message_type.message)
                    self.__success.setAsInt(self.__WRITE_SUCCESS, self.__success.getAsInt(self.__WRITE_SUCCESS) + 1)
                else:
                    log.debug("Sending M105 because communication looks stalled.")
                    # Just write a random sequence number, in the hope to recover our sequence number from the resend.
                    self.__writeSequence(0, "M105")

    def __communicationTrySend(self):
        # Only when connected and there is nothing to resend already, we try to send the next batch of data, or the resend data if there is any
        if self.__connected and self.__resend_number is None:
            if self.__resend_data:
                sequence_number, send_line, send_ack_handle_function = self.__resend_data.pop(0)
                self.__writeSequence(sequence_number, send_line)
                self.__success.setAsInt(self.__WRITE_SUCCESS, self.__success.getAsInt(self.__WRITE_SUCCESS) + 1)
                self.__send_messages_that_still_need_ack.append(self.__MESSAGE_TUPLE(sequence_number, send_line, send_ack_handle_function))
            elif len(self.__send_messages_that_still_need_ack) < MarlinProtocolHandler.__MARLIN_RECEIVE_BUFFER_SIZE:
                # If we have room in the planner, get something from the queue to send.
                send_line, send_ack_handle_function = self.__queue.get(self.__planner_buffer_space > 0)
                if send_line is not None and ("G0" in send_line or "G1" in send_line):
                    # If we got something from the queue, we have less room in the planner buffer when we send this. So already account for this.
                    self.__planner_buffer_space -= 1

                if send_line is None and self.__idle_handler is not None:
                    send_line, send_ack_handle_function = self.__idle_handler(self.__current_sequence_number)

                if send_line is not None:
                    if "G30" in send_line:   # Dirty hack, disable the timeout handling while probing with the capacitive sensor
                        self.__is_probing = True
                        send_ack_handle_function = self.__generateProbingFinishedFunction(send_ack_handle_function)

                    self.__send_messages_that_still_need_ack.append(self.__MESSAGE_TUPLE(self.__current_sequence_number, send_line, send_ack_handle_function))
                    self.__write(send_line)
                    self.__success.setAsInt(self.__WRITE_SUCCESS, self.__success.getAsInt(self.__WRITE_SUCCESS) + 1)

    def __generateProbingFinishedFunction(self, ack_handle_function):
        return lambda reply: self.__probingFinished(reply, ack_handle_function)

    def __probingFinished(self, reply, ack_function):
        self.__is_probing = False
        if ack_function is not None:
            ack_function(reply)

    # @return: A dictionary containing debugging information about the queue. Should only be used for debugging!
    def debugGetInfo(self):
        result = {
            "planner_buffer_space": self.__planner_buffer_space,
            "__last_received_timestamp": self.__last_received_timestamp - time.monotonic(),
            "connected": self.__connected,
            "current_sequence_number": self.__current_sequence_number,
            "is_probing": self.__is_probing,
        }
        if len(self.__send_messages_that_still_need_ack) > 0:
            result["send_messages_that_still_need_ack"] = {}
            for sequence_number, message, callback in self.__send_messages_that_still_need_ack:
                result["send_messages_that_still_need_ack"]["%d" % sequence_number] = message
        return result

    # Write a line of GCode to marlin. With the proper checksum protocol. Increase sequence number for next line.
    # @param: line, unicode string containing the data to send, without a newline.
    def __write(self, line):
        self.__writeSequence(self.__current_sequence_number, line)
        self.__current_sequence_number = NetworkPacket.calculateNextSequenceNumber(self.__current_sequence_number)

    ## @brief Write a line of GCode to marlin. With the proper checksum protocol. Can be used to resend packets using older sequence numbers
    # and will not increase the internal sequence number
    # @param: sequence_number, integer to indicate the sequence number to use
    # @param: line, unicode string containing the data to send, without a newline.
    def __writeSequence(self, sequence_number, line):
        packet = NetworkPacket(sequence_number, line)
        self.__comm.write(packet)

    ## @brief Verifies if the line has a correct CRC16
    #  @param line The line to verify (ok or nack)
    #  @return Returns true if the line has been verified to be without error, otherwise false.
    def __isVerifiedCrc(self, line):
        received_crc = self.__getCrcValue(line)
        if received_crc is None:
            log.warning("CRC is missing from line '%s'", line)
            return False

        to_calculate = bytes(line[:line.index("*") - 1].encode())
        calculated_crc = Crc16(to_calculate).hexdigest()
        #log.debug("Line = '%s', bytes = '%r', Received crc = %r, calculated crc = %r", line, to_calculate.decode(), received_crc, calculated_crc)
        return  int(received_crc, 16) == int(calculated_crc, 16)

    ## @brief Function used to extract a value behind a "key" in GCode lines.
    #  @param key The key to find
    #  @param line The line in which to find the key
    #  @param default The default value, if the key could not be found (or cast into the return_type)
    #  @param return_type The type of the value to be returned, defaults to int
    def __getValue(self, key, line, default=None, return_type=int):
        res = re.search("%s([0-9\.]+)" % key, line)
        if res is None:
            return default
        try:
            return return_type(res.group(1))
        except ValueError:
            return default

    ## @brief Parses the line for the special character *. After this character 4 bytes of Hexadecimal digits should be found
    #  @param line The line to search for the CRC value
    #  @param default The default value, if the key could not be found
    def __getCrcValue(self, line, default=None):
        res = re.search("\*([0-9A-F]{4}$)", line)
        if res is None:
            return default
        try:
            return res.group(1)
        except ValueError:
            return default
