from griffin.math.crc16 import Crc16
import logging
from copy import deepcopy

log = logging.getLogger(__name__.split(".")[-1])

## See https://confluence.ultimaker.com:8443/x/NQdI for more details

## @brief Simple network packet for improved communications with Marlin driver
# The packet consists of the following values:
# Position 0; Length 2 bytes: Start of Packet (header, value 0xFFFF)
# Position 2; Length 1 byte: Sequence Number (0 .. MAX_SEQUENCE_NUMBER-1)
# Position 3; Length 1 byte: Length of Payload (1 .. MAX_CMD_SIZE)
# Position 4; Length 1 .. MAX_CMD_SIZE bytes: Start of Payload data
# Position 4 + Length of Payload; Length 2 bytes: CRC16
class NetworkPacket:
    ## @brief This needs to be the same value as defined in Marlin!
    __MAX_CMD_SIZE = 96

    ## @brief Maximum sequence number allowed (0..MAX_SEQUENCE_NUMBER-1)
    __MAX_SEQUENCE_NUMBER = 255

    ## @brief Header byte
    __HEADER_BYTE = 0xFF

    ## @brief Calculate the fixed protocol size bytes (see class description), simple exclude the length of the payload
    __FIXED_PROTOCOL_SIZE = 2 + 1 + 1 + 2

    ## @brief Indexes (fixed) for the data structure
    __HEADER_INDEX = 0
    __SEQUENCE_NUMBER_INDEX = 2
    __PAYLOAD_SIZE_INDEX = 3
    __PAYLOAD_INDEX = 4
    
    def __init__(self, sequence_number, payload):
        payload_length = len(payload)
        assert payload_length <= NetworkPacket.__MAX_CMD_SIZE, "Maximum length of payload overflow %d > %d, '%r'" % (payload_length, NetworkPacket.__MAX_CMD_SIZE, payload)
        assert payload_length > 0, "Payload cannot be empty"
        ## Because of the cyclic numbering, the sequence number wraps around on __MAX_SEQUENCE_NUMBER
        #  therefore only allowing number from 0 to __MAX_SEQUENCE_NUMBER - 1
        assert sequence_number >= 0 and sequence_number < NetworkPacket.__MAX_SEQUENCE_NUMBER
        assert isinstance(payload, str) or isinstance(payload, bytes)

        NetworkPacket.__assert_is_ascii(payload)

        self.__data = bytearray(payload_length + NetworkPacket.__FIXED_PROTOCOL_SIZE)
        self.__data[NetworkPacket.__HEADER_INDEX] = NetworkPacket.__HEADER_BYTE
        self.__data[NetworkPacket.__HEADER_INDEX + 1] = NetworkPacket.__HEADER_BYTE
        self.__data[NetworkPacket.__SEQUENCE_NUMBER_INDEX] = sequence_number
        self.__data[NetworkPacket.__PAYLOAD_SIZE_INDEX] = payload_length
        crc_start_index = NetworkPacket.__PAYLOAD_INDEX + payload_length
        crc_end_index = crc_start_index + 1

        payload_bytes = payload
        if isinstance(payload, str):
            payload_bytes = payload.encode()
        self.__data[NetworkPacket.__PAYLOAD_INDEX:crc_start_index] = payload_bytes
        self.__data[crc_start_index:crc_end_index] = Crc16(bytes(self.__data[NetworkPacket.__SEQUENCE_NUMBER_INDEX:crc_start_index])).digest()

    @classmethod
    def __assert_is_ascii(cls, payload):
        try:
            if isinstance(payload, str):
                payload.encode('ascii')
            else:
                payload.decode('ascii')
        except UnicodeDecodeError:
            assert False, "Bytes with non-ASCII characters"
        except UnicodeEncodeError:
            assert False, "String with non-ASCII characters"

    ## @brief Returns a copy of the packet contents
    #  @return Returns a byte array of the contents
    def getBytes(self):
        return deepcopy(self.__data)

    ## @brief Gets the payload back as message
    #  @return The payload as string
    def getPayload(self):
        payload_length = self.__data[NetworkPacket.__PAYLOAD_SIZE_INDEX]
        return deepcopy(self.__data[NetworkPacket.__PAYLOAD_INDEX:NetworkPacket.__PAYLOAD_INDEX + payload_length].decode())

    ## @brief Gets the CRC back as value
    #  @return The CRC as unsigned integer
    def getCrc(self):
        payload_length = self.__data[NetworkPacket.__PAYLOAD_SIZE_INDEX]
        crc_start = NetworkPacket.__PAYLOAD_INDEX + payload_length
        crc_end = crc_start + 2
        hex_string = ''.join('{:02x}'.format(b) for b in self.__data[crc_start:crc_end])
        return int(hex_string, 16)

    ## @brief Calculates the next sequence number, given a sequence number
    #  @param sequence_number The sequence number as base value
    #  @param increment The optional increment
    #  @return Returns the new sequence number in line (circular increase)
    @classmethod
    def calculateNextSequenceNumber(cls, sequence_number, increment=1):
        return (sequence_number + increment) % NetworkPacket.__MAX_SEQUENCE_NUMBER

    def __repr__(self):
        return ''.join('{:02x}'.format(byte) for byte in self.__data)

    def __str__(self):
        return "Sequence: {seq_nr}, payload length {len}, payload '{payload}', crc {crc}".format(
            seq_nr=self.__data[2],
            len=self.__data[3],
            payload=self.getPayload(),
            crc=hex(self.getCrc())
        )
