import re
import logging
from griffin import signal
from griffin.printer.controller import Controller
from griffin.printer.procedures.procedureStep import ProcedureStep

log = logging.getLogger(__name__.split(".")[-1])


## @brief Report the distance an axis traveled before it reached the homing position
class CompareHomingDistances(ProcedureStep):

    ## @brief initialize a new step
    # @param key str the name of the step
    # @param expected_value float the expected homing distance. The difference between this and the measured value is signalled via stepDataReady
    # @param axis str the name of the axis to perform this comparison with
    def __init__(self, key, expected_value, axis):
        super().__init__(key)

        self.expected_value = expected_value
        self.axis = axis

        self.stepDataReady = signal.Signal()

        self.__metadata = {}


    ## @brief Performs a homing motion and measures the distance traveled for homing.
    # This distance is compared to an expected value and the absolute different is signalled via stepDataReady.
    def run(self):
        self.__metadata = {}
        c = Controller.getInstance()

        c.waitForQueueToBeEmpty(wait_for_motion_controller=True)
        data = c.sendRawGCode("G28 %s0" % self.axis, wait=True)  # This returns the distance traveled for homing
        log.debug("Marlin reply: %s", data)

        m = re.match(r"home_distance = (-?\d+\.\d+)", data)
        homing_dist = float(m.group(1))

        diff = abs(self.expected_value - homing_dist)

        self.__metadata["diff"] = diff
        self.stepDataReady.emit(diff)

    def getMetaData(self):
        return self.__metadata