from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.temperature.waitForHotendTemperatureStep import WaitForHotendTemperatureStep
from griffin.printer.procedures.temperature.setHotendTemperatureStep import SetHotendTemperatureStep
from griffin.printer.procedures.steps.sleepStep import SleepStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep
from griffin.printer.procedures.wizard.checkCartridgeLoaded import CheckCartridgeLoaded

import logging

log = logging.getLogger(__name__.split(".")[-1])


## @brief This procedure implements the heater test process, according to https://docs.google.com/document/d/1_BHDZxUcYv4od-7QAX-P7bbRzeO_n-smnPM71eoGmM0
#  1. Wait for the hotends to have a temperature of T degrees C
#  2. Turn of object cooling fans
#  3. Wait for stable temperature
#  4. Save start temperature of the hotends
#  5. Turn heaters on for Xs
#  6. Wait Y seconds
#  7. Save end temperature of the hotends
class Pt100HeaterTestProcedure(Procedure):

    ## @brief Default starting temperature (in degrees C)
    __MINIMAL_TEMPERATURE_THRESHOLD = 100

    ## @brief Default heatup time (in seconds)
    __HEATING_TIME = 10

    ## @brief Default extra time to wait after turning off heating (in seconds)
    __DELAY_BEFORE_MEASURING = 4

    ## @brief Timeout to wait for a stable temperature (in seconds)
    __STABILIZE_TEMPERATURE_TIMEOUT = 60

    ## @brief Metadata key for status
    __STATUS_KEY = "STATUS"

    ## @brief Stages of status
    __STAGE_PREPARING = "PREPARING"
    __STAGE_STABILIZING = "STABILIZING"
    __STAGE_HEATING = "HEATING"
    __STAGE_HEATING_DONE = "HEATING_DONE"
    __STAGE_FINISHED = "FINISHED"

    ## @brief Initializes the procedure
    #  @param key The procedure key
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)
        self.__meta_data = { }
        self.__hotends = []
        self.__hotend_count  = 0

    # Setup the steps for the procedure.
    def setup(self, args):
        controller = Controller.getInstance()

        start_temperature = args.get("target", self.__MINIMAL_TEMPERATURE_THRESHOLD)
        heating_time = int(args.get("heating_time", self.__HEATING_TIME))
        settling_time = int(args.get("settling_time", self.__DELAY_BEFORE_MEASURING))

        self.__meta_data = { self.__STATUS_KEY: self.__STAGE_PREPARING }
        self.__hotend_count  = controller.getPropertyValue("hotend_count")
        self.__hotends = [None] * self.__hotend_count
        for hotend_index in range(self.__hotend_count):
            self.__hotends[hotend_index] = controller.getPrintHeadController().getHotendSlot(hotend_index)

        steps = []

        # 0) Check for both hotends being loaded
        for hotend_index in range(self.__hotend_count):
            steps.append(CheckCartridgeLoaded("CHECK_CARTRIDGE_LOADED_%d" % hotend_index, hotend_index))

        # 1) Waiting to reach minimum temperature of T degrees
        for hotend_index in range(self.__hotend_count):
            steps.append(SetHotendTemperatureStep("HEATUP_HOTEND_%d" % hotend_index, {"hotend_nr": hotend_index, "target": start_temperature}))
        for hotend_index in range(self.__hotend_count):
            steps.append(WaitForHotendTemperatureStep("HEATUP_HOTEND_WAIT_%d" % hotend_index, {"hotend_nr": hotend_index}))

        # 2) Turn off the fans
        steps.append(SetPropertyStep("STOP_FANS", "printer.head.0", "cooling_fan_speed", 0.0))
        steps[-1].stepFinished.connect(self.__WaitForStableTemperature)

        # 3) Wait for stable temperature
        steps.append(SleepStep("WAIT_FOR_STABLE_TEMPERATURE", self.__STABILIZE_TEMPERATURE_TIMEOUT))
        # 4) Get start temperature(s)
        steps[-1].stepFinished.connect(self.__getStartTemperatures)

        # 5) Start heating for X seconds
        for hotend_index in range(self.__hotend_count):
            # Note: this assumes the given start_temp is < standby temperature (to prevent filament from burning)
            steps.append(SetHotendTemperatureStep("HEATUP_HOTEND_%d" % hotend_index, {"hotend_nr": hotend_index, "target": "printing_temperature"}))
        steps.append(SleepStep("HEATING_TIME_DELAY", heating_time))
        steps[-1].stepFinished.connect(self.__waitForHeaterCooldown)

        # 6) Wait bit, so temperature can settle down
        steps.append(SleepStep("DELAY_TIME_BEFORE_MEASURING", settling_time))
        # 7) After finishing the step, get the end temperature(s)
        steps[-1].stepFinished.connect(self.__getFinalTemperatures)

        # Start cooling down already
        for hotend_index in range(self.__hotend_count):
            steps.append(SetHotendTemperatureStep("FINAL_COOLDOWN_%d" % hotend_index, {"hotend_nr": hotend_index, "target": 0}))

        steps.append(WaitMessageStep("WAIT_FOR_USER_TO_FINISH", "FINISH"))

        self._addStepSequence(steps)

        # Additional abort steps
        abort_steps = []
        for hotend_index in range(self.__hotend_count):
            abort_steps.append(SetHotendTemperatureStep("FINAL_COOLDOWN_%d" % hotend_index, {"hotend_nr": hotend_index, "target": 0}))
        self._changeToStepSequence(abort_steps)
        self.setAbortstepForMultiple(steps, abort_steps[0])

        return True

    ## @brief Before measuring really starts, a delay is necessary to stabilize the temperature
    def __WaitForStableTemperature(self):
        self.__meta_data[self.__STATUS_KEY] = self.__STAGE_STABILIZING

    ## @brief Gets the start temperatures for the hotends and put them in the meta data
    def __getStartTemperatures(self):
        self.__getHotendTemperatures("start_temperature_hotend_")
        self.__meta_data[self.__STATUS_KEY] = self.__STAGE_HEATING
        log.debug("Start temperatures stored: %r", self.__meta_data)

    ## @brief Heating is stopped, waiting for temperature to settle down
    def __waitForHeaterCooldown(self):
        self.__meta_data[self.__STATUS_KEY] = self.__STAGE_HEATING_DONE

    ## @brief Gets the finishing temperatures for the hotends and put them in the meta data
    def __getFinalTemperatures(self):
        self.__getHotendTemperatures("end_temperature_hotend_")
        self.__meta_data[self.__STATUS_KEY] = self.__STAGE_FINISHED
        log.debug("End temperatures stored: %r", self.__meta_data)

    ## @brief Gets the temperatures for the hotends and put them in the meta data, identified by the prefix
    #  @param prefix The prefix to create a unique key (combined with the hotend index)
    def __getHotendTemperatures(self, prefix):
        for hotend_index in range(self.__hotend_count ):
            self.__meta_data[prefix + str(hotend_index)] = self.__hotends[hotend_index].getPropertyValue("current_temperature")

    def getMetaData(self):
        if self.getActiveStep() is not None:
            step_meta_data = self.getActiveStep().getMetaData()
            if step_meta_data is not None:
                self.__meta_data.update(step_meta_data)
        return self.__meta_data
