import importlib
import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.initialization.initiateHardwareInitialization import InitiateHardwareInitialization
from griffin.printer.procedures.initialization.waitForHardwareInitializationFinished import WaitForHardwareInitializationFinished
from griffin.printer.procedures.procedure import Procedure

log = logging.getLogger(__name__.split(".")[-1])


## This procedure starts any required hardware initalization and waits for it to finish.
#  This is done during startup, and the printer is in the state BOOTING while this procedure is running.
#
#  Optionally, it can report that it is updating internal firmware when this is required.
#  Note that the reporting of firmware update progress is a leftover from the slow update of the motion controller,
#  and can potentially be removed in the future
class HardwareInitializationProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_BOOT)

    ## Setup the hardware initialization procedure. Read the machine configuration to figure out what to initialize.
    #  And construct the steps for actual initialization execution.
    def setup(self, args):
        if Controller.getInstance().getDriver().isConnected():
            # We cannot initiate the hardware if the driver is already connected. This would screw up the system.
            # Normally this would never happen. But as we have a dbus API to start procedures, someone could start this by mistake.
            log.error("Skipping hardware setup since already connected. Cannot initialize twice.")
            return False

        self.__hardware_initializers = []
        hardware = Controller.getInstance().getPropertyValue("hardware")
        for name, hardware_data in hardware.items():
            if "initialization" in hardware_data:
                initialization_data = hardware_data["initialization"]
                module, class_name = initialization_data["class"].rsplit(".", 1)
                initializer_class = getattr(importlib.import_module(module), class_name)

                self.__hardware_initializers.append({
                    "name": name,
                    "class": initializer_class,
                    "parameters": initialization_data["parameters"],
                    "firmware": initialization_data["firmware"],
                    "object": None,
                    "thread": None
                })

        check_step = InitiateHardwareInitialization("FIRMWARE_UPDATE_CHECKING", self.__hardware_initializers)
        write_step = WaitForHardwareInitializationFinished("FIRMWARE_UPDATE_WRITING", self.__hardware_initializers)
        self.setFirstStep(check_step)
        # Add the condition that when the check firmware step detects that we are updating, that we switch to the updating step.
        check_step.setNextConditionalStep("UPDATING", write_step)
        return True

    ## Get the progress information for the hardware initialization.
    def getMetaData(self):
        progress = 1.0
        for initializer in self.__hardware_initializers:
            if initializer["object"] is not None:
                progress = min(initializer["object"].getProgress(), progress)
        return {"progress": progress}
