import logging
import time

from griffin.printer.faultHandler import FaultHandler, FaultCode
from griffin.printer.procedures.procedureStep import OutcomeBase
from griffin.printer.procedures.procedureStep import ProcedureStep
from griffin.thread import Thread

log = logging.getLogger(__name__.split(".")[-1])


## Procedure step to initiate the hardware initialization.
#  This will look at the given hardware initializers and start those in different threads.
#  It will wait till these are done, or till one of them starts with a firmware update.
#  It will exit with a specific outcome when it detects a firmware update, to indicate
class InitiateHardwareInitialization(ProcedureStep):
    def __init__(self, key, hardware_initializers):
        super().__init__(key, outcomes=["UPDATING"])
        self.__hardware_initializers = hardware_initializers

    ## Run this step, start threads for all the hardware initializers and then monitor those till they are finished or an update is detected.
    def run(self):
        for initializer in self.__hardware_initializers:
            name = initializer["name"]
            initializer["object"] = initializer["class"]()
            initializer["thread"] = Thread("Hardware initializer: %s" % (name), target=self.__runHardwareInitializer, args=(name, initializer["object"], initializer["parameters"], initializer["firmware"]))
            initializer["thread"].start()

        while True:
            for initializer in self.__hardware_initializers:
                if initializer["object"].getProgress() > 0.0:
                    return "UPDATING"

            any_initializer_busy = any([initializer["thread"].is_alive() for initializer in self.__hardware_initializers])
            if not any_initializer_busy:
                return OutcomeBase.Success
            time.sleep(0.1)

    ## Run a specific hardware initializer in a thread.
    #  @param: name, name of this initializer, used for debugging/logging
    #  @param: initializer, initializer object.
    #  @param: parameters, Dictionary of parameters passed to the connection function of the initializer.
    #  @param: firmware, filename for this hardware.
    def __runHardwareInitializer(self, name, initializer, parameters, firmware):
        try:
            if not initializer.connect(**parameters):
                FaultHandler.error("Failed to connect to hardware [%s] for initialization" % name, FaultCode.MOTION_CONTROLLER_UPDATE_FAILED)
                initializer.close()
                return
            if not initializer.program(firmware):
                FaultHandler.error("Failed to program firmware [%s]" % name, FaultCode.MOTION_CONTROLLER_UPDATE_FAILED)
                initializer.close()
                return
            log.debug("Finished init for hardware [%s]" % name)
        except:
            log.exception("Exception while initializing hardware.")
            FaultHandler.error("Failed to program firmware [%s] for initialization" % name, FaultCode.MOTION_CONTROLLER_UPDATE_FAILED)
        initializer.close()
