import logging
import time

from griffin.math.vector3 import Vector3
from griffin.printer.buildPlateLocations import BuildPlateLocations
from griffin.printer.controller import Controller
from griffin.printer.faultHandler import FaultHandler, FaultCode
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.pre_and_post_print.auto_bed_level_adjust.calcAndSetZDifferenceStep import CalcAndSetZDifferenceStep
from griffin.printer.procedures.pre_and_post_print.auto_bed_level_adjust.calculatePlaneStep import CalculatePlaneStep
from griffin.printer.procedures.pre_and_post_print.auto_bed_level_adjust.capacitiveOffsetStep import CapacitiveOffsetStep
from griffin.printer.procedures.pre_and_post_print.auto_bed_level_adjust.gotoPositionAndSetZ0Step import GotoPositionAndSetZ0Step
from griffin.printer.procedures.pre_and_post_print.auto_bed_level_adjust.probeZStep import ProbeZStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.gotoLocationStep import GotoLocationStep
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.temperature.setBedTemperatureStep import SetBedTemperatureStep
from griffin.printer.procedures.temperature.setHotendTemperatureStep import SetHotendTemperatureStep
from griffin.printer.procedures.temperature.waitForBedTemperatureStep import WaitForBedTemperatureStep
from griffin.printer.procedures.temperature.waitForHotendTemperatureStep import WaitForHotendTemperatureStep

log = logging.getLogger(__name__.split(".")[-1])


## Procedure to measure bed position and skew, as well as the height difference between hotends.
## Also it controls nozzle temperature
class AlignZAxisProcedure(Procedure):
    AUTOMATIC_BED_LEVEL_MODE_REDO_AFTER_X_DAYS = 2
    AUTOMATIC_BED_LEVEL_MODE_BED_TEMP = 35

    ALL_BED_LEVEL_CORRECTION_MODES = ["manual", "always", "automatic", "once a day", "once a week", "at boot"]
    DEFAULT_BED_LEVEL_CORRECTION_MODE = "automatic"

    __SECONDS_PER_DAY = 24 * 60 * 60
    __SECONDS_PER_WEEK = __SECONDS_PER_DAY * 7

    ## Height above the plane to start the probe, this is used when we do not have a previous probe to use,
    #  But we do have data from the total plane measured with the other nozzle.
    #  We need to account for variations in the bed height, but also that the nozzle could not have lifted correctly and thus we
    #  have invalid measurement compared to what we are expecting. We do not want to start with the nozzle inside the buildplate.
    __PROBE_START_HEIGHT_WITH_PLANE_DATA = 2.5
    ## If we have probe
    __MINIMAL_PROBE_START_HEIGHT = 0.5

    # How much bed skew do we account for. If the bed is tilted, and we haven't measured a 3 point plane yet.
    # We do not now how much skew the bed has. But we can still use previous probe data to enhance the starting position for the next probe position.
    __BED_SKEW_PER_MM = 2.0 / 100.0

    # How much "uncertainty" do we have per mm of travel in the XY distance on the Z height.
    # This indicates how flat the bed is. The furter we move away from a previous probe position the less certain we are about the
    # actual height of the bed due to glass not being flat.
    __SURFACE_HEIGHT_VARIATION_UNCERTAINTY_PER_MM = 1.0 / 100.0

    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)

        self._at_boot = True

        self.__nr_of_probes = 1
        self.__progress = 0
        self.__verbose = Controller.getInstance().getSystemService().isDeveloperModeActive()
        self.__plane_storage = {}

        self.onFinished.connect(self.__onProcedureFinished)

    def setup(self, args):
        self.__progress = 0
        nr_of_probes = 0

        try:
            require_active_leveling = args.get("force_probing", False) or self.__requireActiveLeveling()
        except ValueError:
            return False

        # Get the list of steps that we want for this form of the procedure.
        if require_active_leveling:
            self._at_boot = False
            steps = self.__generateStepsForActiveLeveling(args.get("end_early", False))
        else:
            steps = self.__generateApplySavedSettingsSteps()

        # Check for Probe steps, we want to report progress on probe steps.
        for step in steps:
            if isinstance(step, ProbeZStep):
                nr_of_probes += 1
                step.stepFinished.connect(self.__reportProbeDone)

        self.__nr_of_probes = max(1, nr_of_probes)

        # Generate the step chain.
        self._addStepSequence(steps)
        return True

    ## Generate the steps for a full active leveling cycle,
    #  This will probe the bed at 6 points and calculate the resulting skew/height and store it.
    #  After this it will apply these parameters and setup the printer ready for printing (unless early_end is set)
    #  @param early_end: bool, if is set, then end after saving the data, do not setup the data to be used.
    #  @return a list of ProcedureSteps to go trough this procedure.
    def __generateStepsForActiveLeveling(self, early_end):
        controller = Controller.getInstance()
        hotend_count = controller.getPropertyValue("hotend_count")

        # Generate a probe data storage, this is an array with elements for each hotend, and in each hotend the probe locations.
        probe_hotend_storage = [[] for _ in range(0, hotend_count)]

        # Storage for the plane that indicates the tilt of the bed. This is calculated after the first 3 probes with the right hotend.
        # The "plane" key will be filled in this dictionary.
        self.__plane_storage = {}

        ######################################################################
        # Setup the first flow, doing the full alignment.
        # Precondition: buildplate should be >150mm from nozzle, i.e. homed.
        ######################################################################
        steps = [
            # Switch on the buildplate temperature. Non-blocking
            # This saves time as automatic bed leveling is often executed
            # just before a print is started and then this step has to be performed anyway.
            SetBedTemperatureStep("START_HEATUP_BED", {"target": "initial_temperature"}),
        ]

        # Heat up nozzles to melt any filament stuck to the nozzle and push it aside on touching the buildplate.
        for hotend_index in range(hotend_count):
            steps.append(SetHotendTemperatureStep("HEATUP_HOTEND_%d" % (hotend_index), {"hotend_nr": hotend_index, "target": "standby_temperature"}))
        for hotend_index in range(hotend_count):
            steps.append(WaitForHotendTemperatureStep("HEATUP_HOTEND_WAIT_%d" % (hotend_index), {"hotend_nr": hotend_index}))

        steps += [
            OtherProcedureStep("SWITCH_TO_HOTEND_1", "SWITCH_ACTIVE_HOTEND", args={"target": 1}, duration_estimate_sec=5),
            CapacitiveOffsetStep("SETTING_CAPACITVE_SENSOR_OFFSET", self.__verbose),     # Reset the capacitive offset. Should be done far away from the build plate, i.e. 150 mm. or more.
        ]
        steps += self.__generateTreePointLevelingSteps(probe_hotend_storage[1], 1)
        steps += [
            # Safe a few seconds by starting to heat the hotend 0 a.s.a.p, but do not block waiting.
            SetHotendTemperatureStep("START_HEATUP_HOTEND_0", {"hotend_nr": 0, "target": "standby_temperature"}),

            # Calculate the plane of the probes of the right nozzle and store this in the plane_storage["plane"]
            CalculatePlaneStep("CALCULATE_PLANE", probe_hotend_storage=probe_hotend_storage[1], plane_storage=self.__plane_storage, plane_storage_key="plane", hotend_index=1),

            OtherProcedureStep("SWITCH_TO_HOTEND_0", "SWITCH_ACTIVE_HOTEND", args={"target": 0}, duration_estimate_sec=5),

            # During the probing Marlin temporarily switched off the nozzle heaters to minimize PWM noise.
            # Nozzle temperature dropped about 10 degrees during probing.
            # Wait for the left nozzle to be on target temperature again (we don't care for the right nozzle as that one is already finished).
            #
            # Safety note !!!
            # In other situations where the bed is hot we move the bed all the way down to give the user more space
            # to move his hands inside the closure. Here we decided not to do so since moving the bed down takes about
            # the same time as heating the nozzle up again. All those extra bed movements are more dangerous and more
            # confusing than leaving the bed at the high position.
            OtherProcedureStep("HEATUP_HOTEND_WAIT_0", "SET_HOTEND_TEMPERATURE_WAIT", {"hotend_nr": 0, "target": "standby_temperature"}),
        ]
        steps += self.__generateTreePointLevelingSteps(probe_hotend_storage[0], 0)
        steps += [
            CalcAndSetZDifferenceStep("CALCULATING_ALIGN_Z_AXIS", probe_hotend_storage),
        ]
        if early_end:
            return steps

        steps += [
            GotoLocationStep("MOVE_TO_PRIME_POSITION", "CLEAN_NOZZLES_MAINTENANCE"),

            # Do heat-up here so that we only move the bed up once we start printing, this is to satisfy DEKRA SAFETY requirements!
            # Also do it in this procedure because otherwise the bed moves up, down and up again later etc.
            WaitForBedTemperatureStep("WAIT_BED_TEMPERATURE", {"target": "initial_temperature", "heatup_only": True}),
        ]

        # We know that we are in a print at this point in the code, as else the early_end flag was set.
        metadata = controller.getProcedure("PRINT").getGcodeMetaData()
        for i in range(hotend_count):
            steps.append(SetHotendTemperatureStep("HEATUP_HOTEND_%d" % (i), {"hotend_nr": i, "target": "initial_temperature" if metadata.getInitialTemperature(i) is not None else 0 }))
        for i in range(hotend_count):
            if metadata.getInitialTemperature(i) is None:
                continue
            steps.append(WaitForHotendTemperatureStep("HEATUP_HOTEND_WAIT_%d" % (i), {"hotend_nr": i, "target": "initial_temperature", "heatup_only": True}))

        steps += [
            ## NOTE: after this step you are not allowed to home because it resets the skew (because homing resets your reference frame)
            GotoPositionAndSetZ0Step("MOVE_TO_POS_AND_SET_0", BuildPlateLocations().getPosition("ACTIVE_LEVEL_3").x, BuildPlateLocations().getPosition("ACTIVE_LEVEL_3").y)
        ]

        return steps

    def __generateTreePointLevelingSteps(self, probe_hotend_storage, hotend_index):
        steps = []
        for n in range(1, 4):
            position = BuildPlateLocations.getInstance().getPosition("ACTIVE_LEVEL_%d" % (n), hotend_index=hotend_index)
            steps += [
                GotoPositionStep("MOVE_TO_POS_%d_%d" % (hotend_index, n), calculate_position_function=self.__generateBedProbePositionFunctionFor(probe_hotend_storage, position["x"], position["y"], hotend_index=hotend_index)),
                ProbeZStep("PROBE_Z_AXIS_%d_%d" % (hotend_index, n), probe_hotend_storage, self.__verbose),
            ]
        steps += [
            GotoPositionStep("LOWER_BED_AFTER_PROBING_%d" % (hotend_index), z=BuildPlateLocations.getSafeTravelHeight()),
            HomeHeadStep("RE-HOMING_AFTER_PROBING_%d" % (hotend_index)), # Home after probing. The probing turns off the X/Y motors, and we can (and have) lose position because of this.
        ]
        return steps

    ## Generate a function that will calculate the Z position for active leveling.
    #  This will use the supplied plane in the plane_storage if available
    #  @param probe_hotend_storage: storage list that contains all previous probes for this hotend.
    #           If this is filled the last probe will be used to correct the probe start position closer towards the bed.
    #  @param x: x position
    #  @param y: y position
    #  @param hotend_index: integer, the index of the hotend that we are generating the probe position for.
    #  @return a closure that returns a {"x","y","z"} dictionary for the position to start leveling at.
    def __generateBedProbePositionFunctionFor(self, probe_hotend_storage, x, y, *, hotend_index):
        def calculatePositionFunction():
            # We enter this function with 4 possible cases:
            # 1) No data. Just start at the start height.
            # 2) Plane data is available. Thanks to the plane we can start closer to the bed at any position. But we do need to account for the fact that the bed is uneven.
            # 3) Previous probe data is available. If we account for skew we can start closer to the bed possibly.
            # 4) Both plane and probe data is available. This is the most difficult but most powerful one. From the plane data we can correct for the skew on top of the probe data. So we can start potentially closer then for case 3.
            # As each increasing case should provide a better result, we can just take the lowest value of these 4 results (when available)

            z = BuildPlateLocations.getActiveLevelStartHeight(hotend_index)

            previous_probe = None if len(probe_hotend_storage) < 1 else probe_hotend_storage[-1]

            plane_z = None
            if self.__plane_storage.get("plane") is not None: # Case 2)
                # If we have a plane, we want to start __PROBE_START_HEIGHT_WITH_PLANE_DATA above this plane.
                # So we can probe faster.
                plane_z = self.__plane_storage.get("plane").intersectRay(Vector3(x, y, 0), Vector3(0, 0, 1))
                z = plane_z + self.__PROBE_START_HEIGHT_WITH_PLANE_DATA
            if previous_probe is not None: # Case 3)
                # If we have a previous probe data, for this specific hotend, so we potentially can start closer to the bed.
                # Calculate how far we are from our previous probe point, and then apply an safety margin depending on how far we are from the previous probe point.
                xy_distance_between_current_and_previous_probe = (Vector3(x, y, previous_probe.z) - previous_probe).length()

                # Account for a skewed bed in finding the next start height.
                z_uncertainty = xy_distance_between_current_and_previous_probe * self.__BED_SKEW_PER_MM

                z = min(z, previous_probe.z + self.__MINIMAL_PROBE_START_HEIGHT + z_uncertainty)

                if plane_z is not None: # Case 4)
                    # calculate the height where the plane would say the previous probe should have hit.
                    previous_expected_plane_z = self.__plane_storage.get("plane").intersectRay(Vector3(previous_probe.x, previous_probe.y, 0), Vector3(0, 0, 1))
                    # Use this z position to adjust our plane calculated z position
                    previous_z_adjustment_to_plane = previous_probe.z - previous_expected_plane_z

                    # Account for the uncertainty that we added because the previous probe position has a distance from our new probe position
                    z_uncertainty = xy_distance_between_current_and_previous_probe * self.__SURFACE_HEIGHT_VARIATION_UNCERTAINTY_PER_MM

                    z = min(z, plane_z + previous_z_adjustment_to_plane + self.__MINIMAL_PROBE_START_HEIGHT + z_uncertainty)
            return {"x": x, "y": y, "z": z}
        return calculatePositionFunction

    ## Generate the steps needed to use previously measured/setup data.
    #  This is used when a print is started and the decision was made not to do an active leveling probe but use old data.
    #  @return list of ProcedureSteps
    def __generateApplySavedSettingsSteps(self):
        controller = Controller.getInstance()
        hotend_count = controller.getPropertyValue("hotend_count")

        steps = [
            OtherProcedureStep("SWITCH_TO_HOTEND_0", "SWITCH_ACTIVE_HOTEND", args={"target": 0})
        ]
        # Switch off all the nozzles. Required because we want to switch on the buildplate and the
        # power supply has not enough power to heat up both nozzles and build plate from a cold situation.
        for hotend_index in range(hotend_count):
            steps.append(SetHotendTemperatureStep("COOLDOWN_HOTEND_%d" % (hotend_index), {"hotend_nr": hotend_index, "target": 0}))
        steps += [
            GotoLocationStep("MOVE_TO_PRIME_POSITION", "CLEAN_NOZZLES_MAINTENANCE"),

            # Do heat-up here so that we only move the bed up once we start printing, this is to satisfy DEKRA SAFETY requirements!
            # Also do it in this procedure because otherwise the bed moves up, down and up again later etc.
            WaitForBedTemperatureStep("WAIT_BED_TEMPERATURE", {"target": "initial_temperature", "heatup_only": True}),
        ]

        metadata = controller.getProcedure("PRINT").getGcodeMetaData()
        for i in range(hotend_count):
            steps.append(SetHotendTemperatureStep("HEATUP_HOTEND_%d" % (i), {"hotend_nr": i, "target": "initial_temperature" if metadata.getInitialTemperature(i) is not None else 0}))
        for i in range(hotend_count):
            if metadata.getInitialTemperature(i) is None:
                continue
            steps.append(WaitForHotendTemperatureStep("HEATUP_HOTEND_WAIT_%d" % (i), {"hotend_nr": i}))

        steps += [
            ## NOTE: after this step you are not allowed to home because it resets the skew(because homing resets your reference framework)
            GotoPositionAndSetZ0Step("MOVE_TO_POS_AND_SET_0", BuildPlateLocations().getPosition("ACTIVE_LEVEL_3").x, BuildPlateLocations().getPosition("ACTIVE_LEVEL_3").y)
        ]

        return steps

    ## Check if we need to do a active leveling or not
    # @return bool: True if we want to active level, False if not.
    def __requireActiveLeveling(self):
        controller = Controller.getInstance()
        print_head_controller = controller.getPrintHeadController()

        # Figure out the last time the current combination of hotends was updated with leveling data.
        # For this we use the lowest possible update time.
        last_change = float("inf")
        for hotend_index in range(controller.getPropertyValue("hotend_count")):
            z_height_property = print_head_controller.getHotendSlot(hotend_index).getPropertyAsObject("z_height")
            last_change = min(last_change, z_height_property.getLastChanged())

        # Get our current active leveling mode, and do a basic sanity check.
        leveling_mode = controller.getPreference("BED_LEVEL_CORRECTION_MODE")
        if leveling_mode is None:
            leveling_mode = self.DEFAULT_BED_LEVEL_CORRECTION_MODE
            controller.setPreference("BED_LEVEL_CORRECTION_MODE", leveling_mode)
        elif leveling_mode == "manual" and not controller.hasBuildPlateLevelData():
            # no bed level known stop here.
            FaultHandler.warning("no bed level known, please level manually first", FaultCode.BUILD_PLATE_LEVEL_UNKNOWN)
            raise ValueError("No bed level known")
        elif leveling_mode not in self.ALL_BED_LEVEL_CORRECTION_MODES:
            leveling_mode = self.DEFAULT_BED_LEVEL_CORRECTION_MODE

        # Now that we have data about the active leveling mode, figure out if we want to active level or not.
        log.info("using leveling mode %s (configured mode %s)", leveling_mode, str(controller.getPreference("BED_LEVEL_CORRECTION_MODE")))
        if leveling_mode == "at boot":
            if self._at_boot:
                log.info("probe because boot flag is set!")
                return True
        elif leveling_mode == "automatic":
            if not controller.hasBuildPlateLevelData():
                log.info("probe because BED or Z is not set!")
                return True
            elif last_change + self.AUTOMATIC_BED_LEVEL_MODE_REDO_AFTER_X_DAYS * self.__SECONDS_PER_DAY < time.time():
                log.info("probe because LAST_PROBE is not recent enough(%d days)!", self.AUTOMATIC_BED_LEVEL_MODE_REDO_AFTER_X_DAYS)
                return True
            elif controller.getBed().getPropertyValue("current_temperature") < self.AUTOMATIC_BED_LEVEL_MODE_BED_TEMP:
                log.info("probe because bed temp(%dC) is less than %dC", int(controller.getBed().getPropertyValue("current_temperature")), self.AUTOMATIC_BED_LEVEL_MODE_BED_TEMP)
                return True
            # probe data retrieved based on nozzle pair so re-leveling on a changed pair is implied.
        elif leveling_mode == "once a week":
            if not controller.hasBuildPlateLevelData():
                log.info("probe because LAST_PROBE is not set!")
                return True
            elif last_change + self.__SECONDS_PER_WEEK < time.time():
                log.info("probe because LAST_PROBE is not recent enough(7 days)!")
                return True
        elif leveling_mode == "once a day":
            if not controller.hasBuildPlateLevelData():
                log.info("probe because LAST_PROBE is not set!")
                return True
            elif last_change + self.__SECONDS_PER_DAY < time.time():
                log.info("probe because LAST_PROBE is not recent enough(1 day)!")
                return True
        elif leveling_mode == "manual":
            return False    # no probing, manual already reverted to automatic if BED is not set in offset
        elif leveling_mode == "always":
            log.info("probe because leveling_mode == \"always\"!")
            return True
        return False

    ## callback function for the probe step being finished.
    def __reportProbeDone(self):
        self.__progress += 1

    ## Procedure is finished, reset the progress.
    def __onProcedureFinished(self, procedure):
        self.__progress = 0

    def getMetaData(self):
        c = Controller.getInstance()
        mode = c.getPreference("BED_LEVEL_CORRECTION_MODE")

        if mode not in self.ALL_BED_LEVEL_CORRECTION_MODES:
            mode = self.DEFAULT_BED_LEVEL_CORRECTION_MODE
        
        # use super class getMetaData to return stepinfo progress data
        metaData = super().getMetaData()
        metaData.update({"current": mode, "target": mode})

        return metaData
