from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.queueRawGCodeWaitStep import QueueRawGCodeWaitStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.temperature.waitForMultipleHotendsTemperatureStep import WaitForMultipleHotendsTemperatureStep
from griffin.math import mapping
import logging

log = logging.getLogger(__name__.split(".")[-1])


## This procedure deprimes the nozzles so that the next print can start with approximately known start conditions,
##  the material is not stuck to the hotend and no drop forming at the end of the filament.
class DeprimeProcedure(Procedure):
    FUDGE_FACTOR_PRINTING_TEMPERATURE = 35.0 # degC
    FUDGE_FACTOR_RETRACTION_LENGTH = 1 # mm
    QUICK_RETRACT_DIST = 4.5

    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_UNKNOWN)
        self.__cooldown = True
        self.__steps = None

    def setup(self, args):
        c = Controller.getInstance()
        hotend_count = c.getPropertyValue("hotend_count")
        hotend_data_store = args.get("hotend_data_store")
        if callable(hotend_data_store):# extract data from hotend cleaning wizard
            hotends = hotend_data_store()
        else:
            hotends = range(hotend_count)

        self.__cooldown = args.get("cooldown", True)
        # Determine whether or not to deprime all hotends or just 1
        if "hotend_nr" in args:
            try:
                hotend_nr = int(args.get("hotend_nr"))
                # Make sure the range is from nr to nr +1 , otherwise this gives a []
                if hotend_nr < hotend_count:
                    hotends = range(hotend_nr, hotend_nr + 1)
            except ValueError:
                # Do nothing, assume(!) all hotends are to be deprimed
                pass

        deprime_amount = c.getPropertyValue("deprime_retraction_amount")
        deprime_speed = c.getPropertyValue("deprime_retraction_speed")
        idle_amount = c.getPropertyValue("idle_retraction_amount")

        initial_steps = []
        set_temp_steps = []
        wait_steps = []
        after_temp_steps = []
        second_wait_steps =[]
        after_second_wait = []
        cooldown_steps = []

        active_hotend = c.getGCodeActiveHotend()

        printing_temperature = c.getMaterialManager().getPrintingTemperature(active_hotend)
        temp = c.getPrintHeadController().getHotendSlot(active_hotend).getPropertyValue("current_temperature")
        retracted = c.getHotendRetractedLength(active_hotend)
        quick_retracted = False

        if temp > printing_temperature - self.FUDGE_FACTOR_PRINTING_TEMPERATURE and retracted < self.QUICK_RETRACT_DIST:
            log.info("quick retracting hotend(%d) %.1f mm to prevent oozing", active_hotend, self.QUICK_RETRACT_DIST)
            initial_steps.append(GotoPositionStep("QUICK_RETRACT", e=-self.QUICK_RETRACT_DIST, speed=deprime_speed, relative_positions=True))
            quick_retracted = True
        initial_heatup_hotends = []
        cooldown_hotends = []

        for hotend_nr in hotends:
            hotend = c.getPrintHeadController().getHotendSlot(hotend_nr)
            present = hotend.getPropertyValue("is_present")
            if not present:
                log.info("hotend({hotend_nr}) is not present".format(hotend_nr=hotend_nr))
                continue

            standby_temperature = c.getMaterialManager().getStandbyTemperature(hotend_nr)
            printing_temperature = c.getMaterialManager().getPrintingTemperature(hotend_nr)
            retracted = c.getHotendRetractedLength(hotend_nr)

            log.info("hotend({hotend_nr}) is already {retracted} mm retracted.".format(hotend_nr=hotend_nr, retracted=retracted))
            # Retract even more when the temperature has dropped below the standby temperature (+ 5C).
            # This extra retraction is used to break off the string that is attached to the end of the material.
            # Often this is used in a two-stage process: first deprime to stop oozing and then in a second
            # call, with lower temperature, retract the idle_amount to break off the string.

            moved = 0
            if hotend_nr == active_hotend and quick_retracted:
                moved -= self.QUICK_RETRACT_DIST
            should_wait_on_hotend = False

            if retracted <= deprime_amount - self.FUDGE_FACTOR_RETRACTION_LENGTH:
                if hotend.getPropertyValue("current_temperature") < printing_temperature - self.FUDGE_FACTOR_PRINTING_TEMPERATURE:
                    log.info("going to heat up hotend(%d) because it is %d degC", hotend_nr, hotend.getPropertyValue("current_temperature"))
                    set_temp_steps.append(SetPropertyStep("START_SETTLE_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr),
                        "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "pre_tune_target_temperature", printing_temperature))
                    initial_heatup_hotends.append(hotend_nr)
                after_temp_steps.append(QueueRawGCodeWaitStep("SELECT_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr), "T{hotend_nr}".format(hotend_nr=hotend_nr)))
                moved2 = retracted + -moved - deprime_amount
                after_temp_steps.append(GotoPositionStep("RETRACT_{hotend_nr}".format(hotend_nr=hotend_nr), e=moved2, speed=deprime_speed, relative_positions=True))
                log.info("retracting hotend(%d) %d mm", hotend_nr, moved2)
                moved += moved2
                should_wait_on_hotend = True

            # moved is negative but added up to the retracted amount so inverted.
            if retracted + -moved <= deprime_amount + idle_amount - self.FUDGE_FACTOR_RETRACTION_LENGTH:
                after_temp_steps.append(SetPropertyStep("START_COOLDOWN_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr),
                    "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "pre_tune_target_temperature", standby_temperature))
                after_second_wait.append(QueueRawGCodeWaitStep("SELECT_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr), "T{hotend_nr}".format(hotend_nr=hotend_nr)))
                distance = retracted + -moved - deprime_amount - idle_amount
                after_second_wait.append(GotoPositionStep("DEPRIME_{hotend_nr}".format(hotend_nr=hotend_nr), e=distance, speed=deprime_speed, relative_positions=True))
                log.info("depriming hotend(%d) %d mm", hotend_nr, distance)
                should_wait_on_hotend = True

            cooldown_steps.append(SetPropertyStep("START_COOLDOWN_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr), "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "pre_tune_target_temperature", 0.0))
            if should_wait_on_hotend:
                cooldown_hotends.append(hotend_nr)

        if len(initial_heatup_hotends) > 0:
            wait_steps.append(WaitForMultipleHotendsTemperatureStep("INITIAL_HEATUP_WAIT", initial_heatup_hotends))

        if self.__cooldown and len(after_second_wait) > 0: # Only wait here if we have added temperature set commands in the after_temp_steps list. We check this by abusing the after_second_wait list which only has items when there are temperature commands within the after_temp_steps list.
            second_wait_steps.append(WaitForMultipleHotendsTemperatureStep("COOLDOWN_HOTENDS_WAIT", cooldown_hotends))
        steps = initial_steps + set_temp_steps + wait_steps + after_temp_steps + second_wait_steps + after_second_wait

        # Reset both values tracking filament position
        for hotend_nr in range(c.getPropertyValue("hotend_count")):
            if c.getPrintHeadController().getHotendSlot(hotend_nr).getPropertyValue("is_present") and self.__cooldown:
                steps.append(SetPropertyStep("SET_DEPRIMED_{hotend_nr}".format(hotend_nr=hotend_nr), "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "is_primed", False))
        # the previous 3 lines must be happen before the cooldown_steps as that is the point where the abort hooks in!

        # Set the abort step as the first cooldown step. This aborts the depriming, but leaves the system in a proper state.
        self.setAbortstepForMultiple(steps, cooldown_steps[0])

        if self.__cooldown:
            steps += cooldown_steps
        steps.append(QueueRawGCodeWaitStep("RESET_HOTEND", "T{hotend_nr}".format(hotend_nr=active_hotend)))


        self._addStepSequence(steps)
        self.__steps = steps

        return True

    def getMetaData(self):
        metadata = {"progress": 0.0}

        active_step = self.getActiveStep()
        if active_step is not None and self.__steps is not None and active_step in self.__steps:
            total_long_steps = 0
            current_long_step = 0
            for step in self.__steps:
                if step == active_step:
                    current_long_step = total_long_steps
                if isinstance(step, WaitForMultipleHotendsTemperatureStep):
                    total_long_steps += 1
            if total_long_steps > 0:
                metadata["progress"] = mapping.linearMap(current_long_step + active_step.getMetaData().get("progress", 0.0), 0, total_long_steps, 0.0, 1.0)
        return metadata
