import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.procedureStep import ProcedureStep


log = logging.getLogger(__name__.split(".")[-1])


## This step saves the position, speed and temperature data during the pausing of the print.
class PauseStep(ProcedureStep):
    ## Used by ResumeStep as well
    QUICK_RETRACT_LENGTH_ON_PAUSE = 6.5 #< quick retract distance [mm]
    RETRACT_LENGTH_ON_PAUSE = 26.5 #< total retract distance [mm]

    ## Internal only
    __Z_HOP_HEIGHT = 10 #< move the bed down [mm]
    __HORIZONTAL_WIPE_DISTANCE = 10   #< Horizontal wipe distance [mm]

    def __init__(self, key):
        super().__init__(key)

    def run(self):
        c = Controller.getInstance()

        move_speed = c.getPropertyValue("default_travel_speed") * (1 / (c.getPropertyValue("movement_speed_modifier") / 100)) * 60
        deprime_speed = c.getPropertyValue("deprime_retraction_speed") * 60

        # Pause the controller so no more gcode will be added to motion queue.
        c.setPaused(True)
        # With the controller now in a paused state, motion queues might still be moving but the tracked position
        # is where Marlin will stop.
        # This removes the need to wait for the queue on Marlin to finish! So we can do the retract as quick as possible.
        saved_state = c.getPosition()   # Also saves last used speed!

        # Wipe left or right to prevent running into the end limits
        wipe_x_pos = saved_state["X"] - self.__HORIZONTAL_WIPE_DISTANCE if saved_state["X"] > self.__HORIZONTAL_WIPE_DISTANCE else saved_state["X"] + self.__HORIZONTAL_WIPE_DISTANCE

        c.sendRawGCode("G92 E0")
        c.sendRawGCode("G1 E%d F%f" % (-self.QUICK_RETRACT_LENGTH_ON_PAUSE, deprime_speed * 2))         # High speed retract to prevent oozing
        c.sendRawGCode("G0 X%f Y%f F%d" % (wipe_x_pos, saved_state["Y"], move_speed))                   # wipe nozzle
        c.sendRawGCode("G0 Z%f F%d" % (saved_state["Z"] + self.__Z_HOP_HEIGHT, move_speed))               # hop
        c.sendRawGCode("G1 E%d F%f" % (-self.RETRACT_LENGTH_ON_PAUSE, deprime_speed))                   # Long retract
        c.waitForQueueToBeEmpty()

        cooling_fan_speed = c.getPrintHeadController().getPropertyContainer().get("cooling_fan_speed")
        saved_state["cooling_fan_speed"] = cooling_fan_speed.get()
        saved_state["active_hotend"] = c.getGCodeActiveHotend()

        cooling_fan_speed.set(0)

        log_msg = "saved printer state: {X:f}, {Y:f}, {Z:f}, {F:f}, cooling_fan_speed {cooling_fan_speed:.3g}"
        for hotend_nr in range(c.getPropertyValue("hotend_count")):
            saved_state["T{}".format(hotend_nr)] = c.getPrintHeadController().getHotendSlot(hotend_nr).getPropertyValue("pre_tune_target_temperature")
            log_msg += " Temperature[{0}] {{T{0}:.3g}} E{0}:{{E{0}}}".format(hotend_nr)
        log.info(log_msg.format(**saved_state))

        c.sendRawGCode("G0 F%d" % move_speed)       # Ensure correct speed at step finish

        c.getProcedure("RESUME_PRINT").setSavedState(saved_state)
