from griffin import signal
from griffin import thread
from threading import Lock

from .procedureStep import OutcomeBase
from .procedureStep import ProcedureStep
from griffin.thread import Thread

import time
import logging
import dbus

from collections import namedtuple

# Dedicated tuple is needed to ensure atomic updates of stepinfo metadata
ProcedureStepInfo = namedtuple('ProcedureStepInfo', ['completed', 'remaining', 'active'])

log = logging.getLogger(__name__.split(".")[-1])

#  Base class for all procedures.
#  Procedures are created during startup of the printer service by the controller.
#  Each procedure has a unique key which identifies this procedure.
#  Every procedure has one or more steps that get executed when this procedure is ran.
#  The procedure has a setup step with parameters, during this setup step, the procedure can report that it cannot run.
class Procedure:
    # Not every procedure currently has a clearly defined type. Procedures that not clearly need to set the printer general state have this type.
    TYPE_UNKNOWN = 0
    # When the machine is printing, a procedure with this type is active.
    TYPE_PRINT = 1
    # We have a lot of maintenance related procedures, like change material. When these are active the machine should be marked as in maintenance and not ready for printing.
    TYPE_MAINTENANCE = 2
    # We have some procedures that run during booting. When these procedures are active, the machine is not ready yet and thus should report that.
    TYPE_BOOT = 3

    # Procedure types. Every procedure has a type, which is used to generate the general printer service state. (see printerService.py)
    TYPES = [TYPE_UNKNOWN, TYPE_PRINT, TYPE_MAINTENANCE, TYPE_BOOT]

    def __init__(self, key, procedure_type):
        super().__init__()

        self.onStart = signal.Signal()
        self.onNextStep = signal.Signal()
        self.onFinished = signal.Signal()

        assert procedure_type in Procedure.TYPES

        # A halted procedure must stop a.s.a.p. and _not_ execute any new step(s) or path.
        self.__is_halted = False
        self.__key = key
        self.__type = procedure_type

        # The current state of this procedure.
        self.__procedure_outcome = None

        # Changing the active_step and _abort is critical.
        self.__active_step_lock = Lock()

        # The current step being executed. This can also be one of the abort
        # steps.
        self.__active_step = None

        # Initial step of the normal path to be executed.
        self.__initial_step = None

        # Disables the abort for this procedure.
        self.__abort_disabled = False

        # May be set by abort() or by receiving the ABORT message. During
        # normal path, don't start the next step, but the first step of the
        # "abort path" instead. The "abort path" isn't affected.
        self.__abort_flag = thread.Event()
        # Check if we forced to take an abort branch. This flag is used to only take an abort branch
        # once per abort request.
        self.__abort_branch_forced = False

        self.__ready_event = thread.Event()
        self.__ready_event.set()

        self.__procedure_step_info = None

        self.__prev_active_step_progress = 0

    def setup(self, args):
        return True

    # The default start()
    # Note it's not allowed to start the same instance in different threads,
    # but it's common that the previous thread continues because steps are
    # run in a separate thread.
    # @return boolean: True if this procedure was properly started.
    def start(self):
        return self._protectedStart(skip_driver_check=self.isBootProcedure())

    # Protected start, called from start
    # @param skip_driver_check: boolean to indicate that we do not need to wait for the driver to be connected before starting this procedure.
    # @return boolean: True if this procedure was properly started.
    def _protectedStart(self, *, skip_driver_check):
        # Just don't run with error logged if trying to run twice. The
        # caller should first check canBeExecuted().
        if not self.__ready_event.is_set():
            log.error("Start procedure failed, because previous not ready")
            return
        assert(self.__initial_step is not None)
        assert(self.__active_step is None)
        if not skip_driver_check:
            # Everything that is not a boot procedure, should wait till the driver is connected before it can run,
            # As the boot procedure(s) initialize the hardware and prepare the system for usage.
            from griffin.printer.controller import Controller
            driver = Controller.getInstance().getDriver()
            if not driver.isConnected():
                log.info("Wait for Motion Controller connection.")
                while not driver.isConnected():
                    time.sleep(0.1)
                log.info("Motion Controller connected.")

        return self._startStepLoop()

    # Calculate a progress fraction based on completed, active and remaining steps
    # @param completed: array containing information on completed steps. Elements are dictionaries containing "step" and "elapsed" fields
    # @param active: Dictionary containing information on active step. Keys: "step", "elapsed" and "estimate"
    # @param remaining: array containing completed steps. Elements are dictionaries containing "step" and "estimate" fields
    # @return progress fraction
    def _calculateProgress(self, completed, active, remaining):
        elapsed = sum([i["elapsed"] for i in completed])
        elapsed += active["elapsed"]
        estimate = active["estimate"]
        estimate += sum([i["estimate"] for i in remaining])

        return elapsed / (elapsed + estimate)

    def getMetaData(self):
        info = self.__procedure_step_info
        if info is None or not isinstance(info.active, ProcedureStep):
             return {"EMPTY_ORIGIN": "Procedure %s" % self}

        meta_data = info.active.getMetaData()

        # Expose generic procedure abort flag. This is not named just "aborted" because it might interfer with existing aborted
        # flags that may have a (slightly) different meaning.
        meta_data["procedure_aborted"] = self.isAborted()

        # by default, without info, assume there is no progress yet
        step_progress = 0
        # check for stepinfo first, because progress may be a sub-sub-sub-step progress value and not representative for overall step progress
        if "stepinfo_completed" in meta_data:       
            step_progress = self._calculateProgress(meta_data["stepinfo_completed"], meta_data["stepinfo_active"], meta_data["stepinfo_remaining"])
        elif "progress" in meta_data:
            step_progress = meta_data["progress"]
            if step_progress < 0 or step_progress > 1:
                log.warning("Invalid step progress detected on step {k}: {sp} (resetting to 0)".format(k=info.active.getKey(), sp=step_progress))
                step_progress = 0

        if step_progress < self.__prev_active_step_progress:
            # these warnings should never happen, because it negatively influences user experience
            log.warning("Step progress decreased on step {k}: {prev} --> {curr}".format(k=info.active.getKey(), curr=step_progress, prev=self.__prev_active_step_progress))

        self.__prev_active_step_progress = step_progress

        meta_data["stepinfo_completed"] = dbus.Array(info.completed, signature="a{sv}")
        meta_data["stepinfo_remaining"] = dbus.Array(info.remaining, signature="a{sv}")
        
        step_duration_estimate = info.active.getDurationEstimate()
        meta_data["stepinfo_active"] = dbus.Dictionary({
            "step": info.active.getKey(), 
            "elapsed": step_duration_estimate * step_progress,      # scale progress to given estimate to prevent progress jumps
            "estimate": step_duration_estimate * (1-step_progress) }, signature="sv")

        return meta_data

    ## @brief Add the given sequence of steps as first step to this procedure.
    # @param steps list of steps to add to this procedure
    def _addStepSequence(self, steps):
        self._changeToStepSequence(steps)
        self.setFirstStep(steps[0])

    ## @brief Change the given sequence of steps into a step sequence.
    # @param steps list of steps to be made into a sequence.
    @staticmethod
    def _changeToStepSequence(steps):
        for n in range(0, len(steps) - 1):
            steps[n].setNextDefaultStep(steps[n + 1])

    ## @brief set a condition to many steps at once.
    # @param from_steps a collection of steps to set the given condition on
    # @param condition to condition (or transition) to apply
    # @param to_step the step that the given from_ steps should branch to if the condition/transition is given by the step in from_
    def _setApplyConditionToSteps(self, from_steps, condition, to_step):
        for step in from_steps:
            step.setNextConditionalStep(condition, to_step)

    def isActive(self):
        return not self.__ready_event.is_set()

    def canBeExecuted(self):
        if self.__ready_event.is_set():
            return True
        return False

    def getKey(self):
        return self.__key

    def isPrintProcedure(self):
        return self.__type == Procedure.TYPE_PRINT

    def isMaintenanceProcedure(self):
        return self.__type == Procedure.TYPE_MAINTENANCE

    def isBootProcedure(self):
        return self.__type == Procedure.TYPE_BOOT

    def setFirstStep(self, step):
        assert(isinstance(step, ProcedureStep))
        self.__initial_step = step
        return step

    ## @brief set the abort step of a lot of steps at once
    # @param from_ the steps to set the Aborted-transitions of
    # @param abortStep the step to transition to in case any of the steps in from_ is aborted
    def setAbortstepForMultiple(self, from_steps, to_abort_step):
        return self._setApplyConditionToSteps(from_steps, OutcomeBase.Aborted, to_abort_step)

    def getActiveStep(self):
        """ Returns the reference to the active step.

        Note this reference SHOULD only be used for read only access. The caller
        will never be sure it this is still the active step.

        """

        return self.__active_step

    def setAbortDisabled(self, abort_disabled):
        """ Set the disable abort for the procedure.

        Note this should be handled with care, normally every step and procedure
        is expected to be able to abort again, within a reasonable amount of
        time. However, for example for "shutdown/postprint" actions, the
        usage might be legit.
        """
        self.__abort_disabled = abort_disabled

    def isAbortDisabled(self):
        return self.__abort_disabled

    def abort(self):
        """ Abort the procedure if abort isn't disabled or procedure is halted. """

        if self.__is_halted:
            # Aborting will trigger execution of the abort steps, in case of a halt, this is not desired.
            # We also like to prevent any log messages appearing after a halt, hence this check is kept out of
            # the following sequence.
            return

        # We want to be sure that we inform the real active step, not "half way
        # in between". Otherwise the new step may become active without being
        # notified.
        log.info("{procedure} is going to be aborted".format(procedure=self))
        with self.__active_step_lock:
            # Abort only possible during Running state.
            if not self.__abort_disabled:
                self.__abort_flag.set()
                if self.__active_step:
                    self.__active_step.abort()

    def receiveMessage(self, message):
        """ Handle the default procedure message and pass on messages to step.

        ABORT message is handled by the procedure, but still passed on for
        printFileStep compatibility. Rest aborts on the _abort flag.
        """

        if self.__is_halted:
            # When halted, we do not want any processing of any message at all.
            # We also like to prevent any log messages appearing after a halt, hence this check is kept out of
            # the following sequence.
            return

        # We want to be sure that we inform the real active step, not "half way
        # in between". Otherwise the new step may become active without being
        # notified.

        log.info("Message: '%s' for procedure '%s'", message, self)
        with self.__active_step_lock:

            if message == "ABORT":
                # Abort only possible during Running state.
                if not self.__abort_disabled:
                    log.info("Abort message received for procedure %s", self)
                    self.__abort_branch_forced = False
                    # We do pass the message, but also make sure we will stop the
                    # next steps. Also cancel the active step.
                    self.__abort_flag.set()
                    if self.__active_step:
                        log.info("Pass abort to %s", self.__active_step.getKey())
                        self.__active_step.abort()
                else:
                    log.info("Abort message ignored for procedure %s", self)
            else:
                # Pass on. Note the default step handler handles ABORT. But if it's
                # superseeded it might be blocked anyway.
                if self.__active_step is not None:
                    self.__active_step.receiveMessage(message)

    def isAborted(self):
        return self.__abort_flag.is_set()

    def join(self):
        self.__ready_event.wait()

    def getProcedureOutcome(self):
        return self.__procedure_outcome

    def _startStepLoop(self, threaded=True):
        self.__active_step = self.__initial_step
        self.__procedure_outcome = None
        self.__ready_event.clear()
        self.__abort_flag.clear()
        self.__abort_branch_forced = False

        self.onStart.emit(self, self.__active_step)

        self.__ready_event.clear()
        if threaded:
            thread = Thread("Procedure {}".format(self.__key), self.__stepLoop, daemon=True)
            thread.start()
        else:
            self.__stepLoop()

    # Get happy flow steps that are expected to follow the given step
    # @param from_step: ProcedureStep object representing the step from which to determine next steps
    # @return array of dictionaries with keys 'step' and 'estimate' that tell which step takes what time (estimated, in seconds)
    def __getRemainingSteps(self, from_step):
        result = []

        if not from_step:
            return []

        current_step = from_step.getNextStep(OutcomeBase.Success)
        
        while current_step is not None:
            duration = current_step.getDurationEstimate()
            result.append({'step': current_step.getKey(), 'estimate': duration})
            current_step = current_step.getNextStep(OutcomeBase.Success)

        return result

    # @brief Halts the currently active step.
    # Sets the procedure to HALTED as well as ABORTED, but does NOT trigger any abort sequence.
    def halt(self):
        self.__is_halted = True
        self._abort = True

        with self.__active_step_lock:
            if self.__active_step:
                self.__active_step.halt()

    def __stepLoop(self):
        # Set the initial outcome to error, in case we run no steps we directly fall trough to the end.
        outcome = OutcomeBase.Success
        
        self.__procedure_step_info = ProcedureStepInfo(
            completed = [], 
            active = self.__active_step, 
            remaining = self.__getRemainingSteps(self.__active_step))

        while self.__active_step is not None and not self.__is_halted:
            self.onNextStep.emit(self, self.__active_step)

            try:
                log.debug("Executing step {active}".format(active=self.__active_step))
                outcome = self.__active_step._run()

                log.debug("Outcome {out} from {active}".format(active=self.__active_step, out=outcome))
                if not outcome:
                    outcome = OutcomeBase.Success

                if self.__is_halted:
                    outcome = OutcomeBase.Error
                elif self.isAborted() and not self.__abort_branch_forced:
                    log.warning("{proc} is aborted and overruling outcome {out} from {step}".format(proc=self, out=outcome, step=self.__active_step))
                    outcome = OutcomeBase.Aborted
                    self.__procedure_outcome = OutcomeBase.Aborted
                    self.__abort_branch_forced = True
            except Exception:
                log.exception("Step {step} yielded an exception".format(step=self.__active_step))
                outcome = OutcomeBase.Error
                self.__procedure_outcome = OutcomeBase.Error

            with self.__active_step_lock:
                # Only move to the next step if the procedure is not halted.
                if not self.__is_halted:
                    # If the outcome is Aborted or Error, the self._procedure_state is set accordingly and when that is the last step,
                    next_step = self.__active_step.getNextStep(outcome)

                    log.info("{proc} transitioning from {act} --> {nxt}".format(act=self.__active_step, nxt=next_step, proc=self))
                    completed = self.__procedure_step_info.completed + [{'step': self.__active_step.getKey(), 'elapsed': self.__active_step.getDurationEstimate()}]

                    self.__active_step = next_step
                    self.__prev_active_step_progress = 0

                    self.__procedure_step_info = ProcedureStepInfo(
                        completed = completed,
                        active = self.__active_step, 
                        remaining = self.__getRemainingSteps(self.__active_step))

        # Procedure is finished, if no outcome was set, we will copy the outcome of the latest step.
        # This means if a procedure has no explicit error/abort handling, the error/abort will bubble upwards.
        # Else it will not bubble upwards, and bubbling has to be made explicit.
        if self.__procedure_outcome is None:
            self.__procedure_outcome = outcome

        # Do not change state, trigger events or finish the procedure if it's halted.
        if self.__is_halted:
            self.__procedure_outcome = OutcomeBase.Error
        else:
            self.__ready_event.set()
            self.onFinished.emit(self)

    def __repr__(self):
        return "{cls}(key='{key}')".format(cls=self.__class__.__name__, key=self.getKey())
