from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.programming_station.waitForHotendProfileStep import WaitForHotendProfileStep
from griffin.timer import Timer

import logging
import time

log = logging.getLogger(__name__.split(".")[-1])

## @brief This procedure runs forever to allow the printcore programming station to write the EEPROMs of PrintCores
## Feedback of the status is done by setting the printcore leds:
#  Off: no print core inserted
#  Purple pulsing: Cannot program the PrintCore due to lack of selection of a PrintCore Profile (select profile, reinsert PrintCore)
#  Orange: detected PrintCore inserted
#  Cyan: the PrintCore EEPROM is being written
#  Green: the PrintCore has been programmed successfully
#  Red: there was a problem writing the EEPROM!
#  These colors are defined in the @ref HeadSlotController
class ProgramPrintCoresProcedure(Procedure):
    ## @brief Initializes the procedure
    #  @param key The procedure key
    def __init__(self, key):
        log.info("Constructing ProgramPintCoresProcedure")
        super().__init__(key, Procedure.TYPE_MAINTENANCE)
        self.__controller = None
        self.__waitForHotendProfileStep = WaitForHotendProfileStep("WAIT_FOR_HOTEND_PROFILE")
        self.__current_profile = ""

    # Setup the steps for the procedure.
    def setup(self, args):
        log.info("Setting up ProgramPintCoresProcedure")
        self.__setupCallbackHandlers()
        self.__waitForHotendProfileStep.stepDataReady.connect(self.__onNewHotendProfileSelected)
        steps = [
            self.__waitForHotendProfileStep
        ]
        self._addStepSequence(steps)

        return True

    ## @brief Sets up the controller and the change handler for the print cores
    def __setupCallbackHandlers(self):
        self.__controller = Controller.getInstance()
        self.__pch = self.__controller.getPrintHeadController()
        self.__hcm = self.__pch.getHotendCartridgeManager()

        # Using the hotend data ready signal from the HCM is not the right one, as it will signal as well every 15s
        hotend_count = self.__controller.getPropertyValue("hotend_count")
        self.__hotends = [None] * hotend_count
        for hotend_index in range(0, hotend_count):
            self.__hotends[hotend_index] = self.__pch.getHotendSlot(hotend_index)
            hotend_is_present_property = self.__hotends[hotend_index].getPropertyAsObject("is_present")
            hotend_is_present_property.onChange.connect(self.__generateHotendChangeCallback(hotend_index))

    ## @brief: Generate a closure that can be used as callback from property changes to update the material settings for a certain hotend nr.
    #          This is done to capture the hotend_nr variable in the closure.
    #  @param The hotend_nr to pass to the callback function
    #  @return Callback function
    def __generateHotendChangeCallback(self, hotend_nr):
        return lambda property, value: self.__onHotendChanged(hotend_nr, value)

    ## @brief The callback function which will be called when a hotend insertion or removal is detected and act accordingly:
    #         When a hotend is inserted, it will be added to the event queue - if there has been a profile selected!
    #         Depending on the situation, the hotend leds will start indicating the status
    #  @param hotend_index The index of the hotend slot (0-1)
    #  @param is_present True if the hotend is present (inserted) or False in case of removal
    def __onHotendChanged(self, hotend_index, is_present):
        hotend = self.__hotends[hotend_index]
        if is_present:
            if self.__current_profile == "":
                log.info("Cannot program hotend without a hotend profile")
                hotend.setPropertyValue("programming_state", "no_profile_selected")
            else:
                hotend.setPropertyValue("programming_state", "idle")
                execute_programming = Timer("PROGRAM_HOTEND", 0.5, self.__programHotend, hotend_index)
                execute_programming.setSingleShot(True)
                execute_programming.setThreaded(True)
                execute_programming.start()
        else:
            hotend.setPropertyValue("programming_state", "removed")

    ## @brief The actual programming of the hotend
    #  @param hotend_index The hotend which should be programmed
    def __programHotend(self, hotend_index):
        assert self.__current_profile != "", "The current profile should be known!"

        hotend = self.__hotends[hotend_index]
        log.info("HotEnd %d inserted, start programming", hotend_index)
        hotend.setPropertyValue("programming_state", "writing")

        # Retry 10 times, with 0.1second delay. The hotend object isn't always created yet when we arrive here,
        # so we want to try a few times till it is.
        retry_count = 10
        for n in range(retry_count):
            if self.__hcm.programHotend(hotend_index, self.__current_profile):
                log.info("Hotend %d programmed", hotend_index)
                hotend.setPropertyValue("programming_state", "programmed")
                return
            else:
                time.sleep(0.1)
        log.info("Failed to program hotend %d", hotend_index)
        hotend.setPropertyValue("programming_state", "error")

    ## @brief This function will be called when the step has data available
    #  @param step_data Contains the file name of the profile
    def __onNewHotendProfileSelected(self, step_data):
        log.info("New selected profile: %s", step_data)
        self.__current_profile = step_data
