from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.temperature.setBedTemperatureStep import SetBedTemperatureStep
from griffin.printer.procedures.steps.sleepStep import SleepStep

import logging

log = logging.getLogger(__name__.split(".")[-1])


# This procedure sets the Bed temperature.
class SetBedPreheatTemperatureProcedure(Procedure):
    BED_HOTEND_NR = -1

    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_UNKNOWN)
        self.__state_property = None
        self.__controller = None
        self.onFinished.connect(self.__disconnectStateProperty)

    # Setup the steps for the procedure.
    # the target temperature may be influenced by the temperature tweak, the user adjustable value.
    def setup(self, args):
        self.__setupStateProperty()

        # Do not start pre-heat when printer is not idle!
        if self.__state_property.get() != "idle":
            return False

        target = self.__controller.getBed().deriveTemperature(**args)
        if target < 0:
            return False

        steps = [
            SetBedTemperatureStep("START_HEATUP_BED", {"target": target})
        ]

        # If the target == 0, the timeout is not necessary
        if (target > 0):
            timeout = int(args["timeout"])
            steps += [
                SleepStep("WAIT_FOR_TIMEOUT", timeout),
                SetBedTemperatureStep("COOL_DOWN_BED", {"target": 0})
            ]

        self._addStepSequence(steps)
        self.setAbortstepForMultiple(steps, None)
        return True

    ## @brief Collect the metadata
    #  @return Returns the metadata as a dictionary
    def getMetaData(self):
        data = {"ACTIVE": False}
        if self.getActiveStep() is not None and isinstance(self.getActiveStep(), SleepStep):
            data.update(self.getActiveStep().getMetaData())
            data["ACTIVE"] = True
        return data

    ## @brief Check for the printer state changes. If this isn't idle, then abort this procedure
    #  @param property_object The property_object, in this case the state
    #  @param value The value it was set to
    def __onPrinterStatePropertyChanged(self, property_object, value):
        if value != "idle":
            if self.isActive():
                self.abort()
                # We disable the bed temperature here.
                # We are currently running before the procedure code that is changing the system state from idle to something else,
                # and this procedure will be actually running after we return. So we can set the bed temperature to 0 here without having a race condition.
                self.__controller.getBed().setPropertyValue("pre_tune_target_temperature", 0.0)

    ## @brief Sets up the controller and the change handler for the state property
    def __setupStateProperty(self):
        self.__controller = Controller.getInstance()
        self.__state_property = self.__controller.getPropertyContainer().get("state")
        self.__state_property.onChange.connect(self.__onPrinterStatePropertyChanged)

    ## @brief Disconnects the change handler for the state property
    #  @param procedure The procedure
    def __disconnectStateProperty(self, procedure):
        if self.__controller is not None and self.__state_property is not None:
            self.__state_property.onChange.disconnect(self.__onPrinterStatePropertyChanged)
        self.__state_property = None
        self.__controller = None