import time

from griffin.printer.procedures.procedureStep import ProcedureStep
from griffin.printer.heatableObject.heatableObject import HeatableObject
import logging

log = logging.getLogger(__name__.split(".")[-1])


## Base class to set a temperature on a heatable object.
#  Never use this class directly, but use one of its subclasses.
class SetHeatableObjectTemperatureStep(ProcedureStep):
    ## Create a new SetHeatableObjectTemperatureStep.
    # @param key: string, name of this step.
    # @param target_temperature_configuration: None or dictionary, containing the key/values that are used to configure this step.
    #                               see HeatableObject.deriveTemperature functions for details.
    def __init__(self, key, target_temperature_configuration, heatable_object):
        super().__init__(key)

        assert isinstance(heatable_object, HeatableObject), "SetHeatableObjectTemperatureStep requires a heatable object."

        self.__target_temperature_configuration = target_temperature_configuration
        self.__target_temperature = None
        self.__target_heatable_object = heatable_object
        self.__progress = -1.0
        self.__skip = False

    ## Update the target temperature with a dictionary of parameters.
    #  see HeatableObject.deriveTemperature functions for details.
    # @param **args: Key value arguments passed to deriveTemperature.
    # @return Returns True if target temperature is set, otherwise False
    def updateTargetTemperature(self, **args):
        target = self.__target_heatable_object.deriveTemperature(**args)
        if target is None or target < 0:
            return False
        self.__target_temperature = target
        return True

    ## Private function to update the heatable object that we are pointing at.
    #  this is used by the hotend subclass to update to a specific hotend.
    #  @param new_heatable_object: HeatableObject subclass, new heatable object to set the temperature on and wait for with this instance.
    def _updateHeatableObject(self, new_heatable_object):
        assert isinstance(new_heatable_object, HeatableObject), "WaitForHeatableObjectStep requires a heatable object."

        self.__target_heatable_object = new_heatable_object

    ## Overruled run function from base class.
    def run(self):
        if self.__target_temperature_configuration is not None:
            self.updateTargetTemperature(**self.__target_temperature_configuration)

        if self.__target_temperature is not None:
            self.__target_heatable_object.setPropertyValue("pre_tune_target_temperature", self.__target_temperature)
            log.info("Setting target temperature of %s to %.0fC", self.__target_heatable_object.getName(), self.__target_temperature)

    def getDurationEstimate(self):
        return 0.1