from griffin.printer.controller import Controller
from griffin.printer.faultHandler import FaultHandler, FaultCode
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.temperature.setHotendTemperatureStep import SetHotendTemperatureStep
from griffin.printer.procedures.dummyStep import DummyStep

import logging

log = logging.getLogger(__name__.split(".")[-1])


# This procedure sets the hotend temperature
class SetHotendTemperatureProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_UNKNOWN)
        self.__step = SetHotendTemperatureStep("SET_HOTEND_TEMPERATURE")

    # Setup the steps for the procedure.
    # The target temperature may be influenced by the temperature tweak, the user adjustable value.
    # The hotend_nr parameter MUST be set in the args dictionary except when this command is used from gcode (M104).
    def setup(self, args):
        ## Determine hotend index first, before getting target temperature!
        self.__step.updateTargetHotend(**args)
        target = Controller.getInstance().getPrintHeadController().getHotendSlot(self.__step.getHotendIndex()).deriveTemperature(**args)
        if target is None:
            self.setFirstStep(DummyStep("DUMMYING"))
            return True

        if target < 0:
            FaultHandler.warning("Incorrect printing temperature", FaultCode.MATERIAL_TEMPERATURE)   # For displaying
            return False

        self.setFirstStep(self.__step)
        self.__step.updateTargetTemperature(target=target)
        log.info("Heating up hotend %d to %.0fC", self.__step.getHotendIndex(), target)
        return True

    # Returns the metadata of the procedure.
    def getMetaData(self):
        controller = Controller.getInstance()
        hotend = controller.getPrintHeadController().getHotendSlot(self.__step.getHotendIndex())
        return {
            "original_target": hotend.getPropertyValue("pre_tune_target_temperature"),
            "user_tweak": hotend.getPropertyValue("tune_offset_temperature"),
            "current": hotend.getPropertyValue("current_temperature"),
            "target": hotend.getPropertyValue("target_temperature"),
            "hotend_nr": self.__step.getHotendIndex(),
            "display_hotend_nr": self.__step.getHotendIndex() + 1,
            "max_temperature": hotend.getPropertyValue("max_target_temperature")
        }
