import time

from griffin.printer.procedures.procedureStep import ProcedureStep
from griffin.printer.heatableObject.heatableObject import HeatableObject
import logging

log = logging.getLogger(__name__.split(".")[-1])


## Base class to set a temperature on a heatable object and wait for it to be reached.
#  Never use this class directly, but use one of its subclasses.
class WaitForHeatableObjectsStep(ProcedureStep):
    __WAIT_LOGGING_TIME = 5.0

    ## Create a new WaitForHeatableObjectStep.
    # @param key: string, name of this step.
    # @param target_temperature_configuration: None or dictionary, containing the key/values that are used to configure this step.
    #                               see HeatableObject.deriveTemperature functions for details.
    def __init__(self, key, target_temperature_configuration, step_timeout=None):
        super().__init__(key, step_timeout=step_timeout)

        self.__target_temperature_configuration = target_temperature_configuration
        self.__target_temperature = None
        self.__target_heatable_objects = []

        # Adds tolerance to the check.
        self.__wait_tolerance = 0

        # Boolean flags to see if we only want to wait for heatup or cooldown.
        self.__wait_for_heatup = True
        self.__wait_for_cooldown = True

        self.__progress = -1.0
        self.__skip = False

    ## Update the target temperature with a dictionary of parameters.
    #  see HeatableObject.deriveTemperature functions for details.
    # @param **args: Key value arguments passed to deriveTemperature.
    #            Parameter target is mandatory and sets the destination temperature. Other parameters are ignored when this is omitted !!
    #            wait_tolerance: specifies the +/- temperature deviation that will be accepted. 
    #            cooldown_only: when true specifies that we only wait when cooling down.
    #            heatup_only:  when true specifies that we only wait when heating up.
    # @return Returns True if target temperature is set, otherwise False
    def updateTargetTemperature(self, **args):
        #DIRTY!
        target = self.__target_heatable_objects[0].deriveTemperature(**args)
        if target is None or target < 0:
            return False
        self.__target_temperature = target

        self.__wait_tolerance = args.get("wait_tolerance", 0)
        if args.get("cooldown_only", False):
            self.__wait_for_heatup = False
            self.__wait_for_cooldown = True
        elif args.get("heatup_only", False):
            self.__wait_for_heatup = True
            self.__wait_for_cooldown = False
        else:
            self.__wait_for_heatup = True
            self.__wait_for_cooldown = True

        return True

    ## Private function to update the heatable object that we are pointing at.
    #  this is used by the hotend subclass to update to a specific hotend.
    #  @param new_heatable_objects: A list of HeatableObject subclass, new heatable object to set the temperature on and wait for with this instance.
    def _setHeatableObjects(self, new_heatable_objects):
        assert isinstance(new_heatable_objects, list), "WaitForHeatableObjectStep requires a list of heatable object."
        for new_heatable_object in new_heatable_objects:
            assert isinstance(new_heatable_object, HeatableObject), "WaitForHeatableObjectStep requires a list of heatable object."

        self.__target_heatable_objects = new_heatable_objects

    ## Overruled run function from base class.
    def run(self):
        self.__skip = False

        if self.__target_temperature_configuration is not None:
            self.updateTargetTemperature(**self.__target_temperature_configuration)

        if self.__target_temperature is not None:
            for heatable_object in self.__target_heatable_objects:
                heatable_object.setPropertyValue("pre_tune_target_temperature", self.__target_temperature)
                log.info("Setting target temperature of %s to %.0fC", heatable_object.getName(), self.__target_temperature)

        start_temperatures = [heatable_object.getPropertyValue("current_temperature") for heatable_object in self.__target_heatable_objects]

        last_log_time = time.monotonic()
        while not self._abort and not self.__skip:
            if all(self.__isTargetTemperatureReached(heatable_object) for heatable_object in self.__target_heatable_objects):
                break

            total_progress = 1.0
            for heatable_object, start_temperature in zip(self.__target_heatable_objects, start_temperatures):
                current = heatable_object.getPropertyValue("current_temperature")
                target = heatable_object.getPropertyValue("target_temperature")

                start_delta = abs(start_temperature - target)
                current_delta = abs(current - target)

                if heatable_object.isTargetTemperatureReached(): # Strange hack because the above delta calculation is incorrect.
                    current_delta = 0.0

                if start_delta != 0.0:
                    progress = 1.0 - (current_delta / start_delta)
                else:
                    progress = 0.0

                total_progress = min(progress, total_progress)

            self.__progress = max(min(total_progress, 1.0), 0.0)

            # Sleep a bit to prevent busy looping.
            time.sleep(0.5)

            if time.monotonic() > last_log_time + self.__WAIT_LOGGING_TIME:
                last_log_time = time.monotonic()
                for heatable_object, start_temperature in zip(self.__target_heatable_objects, start_temperatures):
                    current = heatable_object.getPropertyValue("current_temperature")
                    target = heatable_object.getPropertyValue("target_temperature")
                    if current < target:
                        log.info("Waiting for %s heatup: %.1f/%.1f (wait_tolerance: %.1f)", heatable_object.getName(), current, target, self.__wait_tolerance)
                    else:
                        log.info("Waiting for %s cooling: %.1f/%.1f (wait_tolerance: %.1f)", heatable_object.getName(), current, target, self.__wait_tolerance)

        if self._abort:
            for heatable_object in self.__target_heatable_objects:
                heatable_object.setPropertyValue("pre_tune_target_temperature", 0.0)

    def __isTargetTemperatureReached(self, heatable_object):
        if heatable_object.isTargetTemperatureReached():
            return True

        current = heatable_object.getPropertyValue("current_temperature")
        target = heatable_object.getPropertyValue("target_temperature")
        # Check if we need to continue because we only wait for heating up or cooling down.
        if not self.__wait_for_heatup and current <= target:
            log.debug("Not waiting for heatup")
            return True
        if not self.__wait_for_cooldown and current >= target:
            log.debug("Not waiting for cooldown")
            return True

        # If we have a configured tolerance, see if we are within that tolerance and then continue.
        if self.__wait_tolerance:
            if abs(current - target) <= self.__wait_tolerance:
                log.debug("Within wait_tolerance of {wt}".format(wt=self.__wait_tolerance))
                return True
        return False

    ## Overruled reveiveMessage class.
    #  Called when a different service messages this procedure.
    #  @param message: string, containing the message that was send to the procedure in which this step was created.
    def receiveMessage(self, message):
        # This allows waiting for the target to be reached to be skipped, this does not turn off the target, so it will be reached eventually.
        if message == "SKIP":
            self.__skip = True

    ## Overrules metadata function.
    #  Overruled to give progress information as well as other data about this temperature step.
    # @return dictionary
    def getMetaData(self):
        return {
            "progress": float(self.__progress),
            "time": int(-1),
            "time_total": int(-1)
        }

    def getDurationEstimate(self):
        return 30