import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.print.pausePrintProcedure import PauseStep
from griffin.printer.procedures.print.switchActiveHotendStep import SwitchActiveHotendStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.queueRawGCodeWaitStep import QueueRawGCodeWaitStep
from griffin.printer.procedures.steps.gotoLocationStep import GotoLocationStep
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.steps.homeBedStep import HomeBedStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.steps.storeValueStep import StoreValueStep
from griffin.printer.procedures.temperature.setHotendTemperatureStep import SetHotendTemperatureStep
from griffin.printer.procedures.waitForQueueEmptyStep import WaitForQueueToBeEmptyStep
from griffin.printer.procedures.wizard.moveMaterialWaitMessageStep import MoveMaterialWaitMessageStep
from griffin.printer.procedures.wizard.unloadMaterialStep import UnloadMaterialStep
from griffin.printer.procedures.wizard.waitForUserToSelectMaterialStep import WaitForUserToSelectMaterialStep
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep
from griffin.printer.properties.printer.printJobStateProperty import PrintJobStateProperty


log = logging.getLogger(__name__.split(".")[-1])


# The wizard procedure which handles the change material option for the end user.
class ChangeMaterial(Procedure):
    EXTRA_RETRACT_LENGTH = 40
    INSERT_SAFETY_LENGTH = 40
    SKIP_START = "skip_start"
    SKIP_END = "skip_end"
    EXPECTED_MATERIAL = "expected_material"
    __HEATUP_DEGREES_PER_SECOND = 2.0  #  [°C/s]  A typical hotend heats up with 2 °C/s
    __FORWARDING_DURATION = 15.0  #  [s] How long does forwarding the material take
    __TEMPERATURE_TOLERANCE = (__HEATUP_DEGREES_PER_SECOND * __FORWARDING_DURATION)  #  degC/s * s = degC

    __USER_SELECT_MATERIAL_TIMEOUT = 600  # [s] Give the user some time to select a material
    __USER_CONFIRM_EXTRUSION_TIMEOUT = 180  # [s] Give the user some time to confirm material is being extruded

    def __init__(self, key, wizard_type="change"):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)
        self._hotend_nr = 0
        self._wizard_type = wizard_type
        self.__controller = Controller.getInstance()
        self._bowden_tube_length = self.__controller.getPropertyValue("bowden_tube_length")
        self._skip_start = False
        self._skip_end = False
        self._expected_material = None
        self._in_print = False
        self._job_state = None
        self.__temperature_storage = {"unload_temp": 0, "load_temp": 0}

    def setup(self, args):
        self._hotend_nr = int(args.get("hotend_nr", 0))
        self._skip_start = bool(args.get(self.SKIP_START, False))
        self._skip_end = bool(args.get(self.SKIP_END, False))
        if self._wizard_type == "select":
            self._skip_start = True
            self._skip_end = True
        self._expected_material = args.get(self.EXPECTED_MATERIAL, None)
        if self._expected_material == "None":
            self._expected_material = None

        self._in_print = self.__controller.getProcedure("PRINT").isActive()   #  bool(args.get(self.IN_PRINT, False))

        self._job_state = self.__controller.getPropertyValue("job_state")
        if not self._job_state in [PrintJobStateProperty.PRINTING, PrintJobStateProperty.PAUSED, PrintJobStateProperty.NO_JOB]:
            return False

        self._old_active_hotend = self.__controller.getGCodeActiveHotend()

        ## the expected material is used for display in the welcome setup but since it has to be changed dynamically, the expected material can be set to the wizard running this change material,
        # aka, the welcome setup wizard. The welcome setup stores the type of nozzle that the user has selected to use and from that an expected material can be derived,
        if isinstance(self._expected_material, Procedure):
            meta = self._expected_material.getMetaData()
            if meta["next_expected_hotend_id"] == "BB 0.4":
                self._expected_material = "PVA"
            else:
                self._expected_material = None

        if self._expected_material is not None:
            log.info("expected material: %s", self._expected_material)

        # If the hotend is not there, then materials should not be loaded/unloaded (changed) nor moved since all of that requires heating up the hotend
        if not self.__controller.getPrintHeadController().getHotendCartridgeManager().isPresent(self._hotend_nr):
            return False

        # Generate all the steps for this wizard procedure.
        steps = []
        if not self._skip_start:
            steps += self._generatePreparationSteps()
        else:
            steps.append(QueueRawGCodeWaitStep("SELECT_HOTEND", "T{hotend_nr}".format(hotend_nr=self._hotend_nr)))

        # If the new material has a much lower printing_temperature than the old material, the old material will not melt and will stick in the hotend.
        # Be sure to reset this, when the procedure is set-up so we don't keep the temperatures of a previously unloaded material here by accident somehow.
        # Because we take the max and at least load_temp is always set, the actual used temperature is never really 0.
        self.__temperature_storage = {"unload_temp": 0, "load_temp": 0}

        if self._wizard_type == "change":
            steps += self._generateUnloadMaterialSteps(pre_heat=True)
            steps += self._generateLoadMaterialSteps(pre_heat=False)
        elif self._wizard_type == "select":
            steps += self._generateSelectMaterialSteps()
        elif self._wizard_type == "unload":
            steps += self._generateUnloadMaterialSteps()
        elif self._wizard_type == "load":
            steps += self._generateLoadMaterialSteps()
        else:
            log.error("Unknown change material wizard type: %s", self._wizard_type)

        if not self._skip_end:
            steps += self._generateFinalizationSteps()
            self.setAbortstepForMultiple(steps[:-1], steps[-1])

        self._addStepSequence(steps)

        return True

    def _generatePreparationSteps(self):
        steps = []

        if self._in_print:
            if self._job_state == PrintJobStateProperty.PRINTING:
                steps.append(OtherProcedureStep("PAUSE_BEFORE_CHANGE_MATERIAL", "PAUSE_PRINT"))
                gcode_processor = self.__controller.getGCodeProcessor()
                gcode_processor.setRetractedLength(
                    self._old_active_hotend,
                    gcode_processor.getRetractedLengthForHotend(self._old_active_hotend) +
                        PauseStep.RETRACT_LENGTH_ON_PAUSE
                )
        else:
            steps += [
                # Prepare the printer for material change.
                # The printer state is unknown, so home the bed first, then home the head, making sure the head will not collide with clips.
                HomeBedStep("HOME_BED"),
                HomeHeadStep("HOME_HEAD"),
                SwitchActiveHotendStep("RESET_ACTIVE_HOTEND", "reset"),
            ]

        if not self._in_print or self._hotend_nr != self._old_active_hotend:
            steps.append(OtherProcedureStep("SWITCH_TO_HOTEND", "SWITCH_ACTIVE_HOTEND", {"target": self._hotend_nr}))

        steps.append(GotoLocationStep("MOVE_HEAD_TO_MAINTENANCE_POS", "CLEAN_NOZZLES_MAINTENANCE"))
        return steps

    def _unloadForPrimedSteps(self):
        return [
            # Heat up the hotend to printing temperature, so we can extract the material properly.
            OtherProcedureStep("HEATUP_WAIT", "SET_HOTEND_TEMPERATURE_WAIT", {"hotend_nr": self._hotend_nr, "target": "printing_temperature", "heatup_only": True}),

            # Move the material backwards, first to a short pull, later on the long move.
            OtherProcedureStep("REVERSE_MATERIAL_A", "MOVE_MATERIAL", {"hotend_nr": self._hotend_nr, "target": -10, "speed": 10})
        ]

    def _generateUnloadMaterialSteps(self, *, pre_heat=False):
        primed = self.__controller.getPrintHeadController().getHotendSlot(self._hotend_nr).getPropertyValue("is_primed")
        steps = []

        if primed:
            steps += self._unloadForPrimedSteps()
        elif pre_heat:
            # Hotend is heated for primed material removal already, else start pre-heating to speed-up the loading.
            steps.append(
                # A safe value of 100C is chosen that won't degrade the materials.
                OtherProcedureStep("PRE_HEAT_NOZZLE", "SET_HOTEND_TEMPERATURE", {"hotend_nr": self._hotend_nr, "target": 100})
            )

        steps += [
            OtherProcedureStep("REVERSE_MATERIAL_B", "MOVE_MATERIAL", {"hotend_nr": self._hotend_nr, "target": -self._bowden_tube_length - self.EXTRA_RETRACT_LENGTH}),
            WaitForQueueToBeEmptyStep("WAIT_REVERSE_MATERIAL_FINISHED"),
            WaitMessageStep("WAIT_FOR_USER_TO_REMOVE_MATERIAL", "DONE"),
            UnloadMaterialStep("UNLOADING_MATERIAL", self._hotend_nr)
        ]

        if not pre_heat:
            # Ofcourse we don't want a cooldown after a pre-heat.
            steps.append(SetHotendTemperatureStep("START_COOLDOWN_HOTEND", {"hotend_nr": self._hotend_nr, "target": 0}))

        return steps

    def _generateSelectMaterialSteps(self):
        return [
            # TODO: remove after EM-403
            # This step was introduced as a result of EM-129 but should be removed after EM-403 implementation.
            # This step is a clumsy implementation for changing the current selected material type. In fact,
            # what we want is a simple menu routine instead of calling this wizard. Problem is that the material
            # property is now saved by the Wizard. A cleaner implementation can be created when we have material
            # properties and then the Wizard can wait for the next step by being triggered on a material change.

            # explain which nozzle material is supposed to be inserted into in selection step because we're already using rfid detection
            # Have the user select the material that he will insert.
            WaitForUserToSelectMaterialStep("USER_SELECT_MATERIAL", self._hotend_nr, only_select_material=True)
        ]

    def _generateLoadMaterialSteps(self, *, pre_heat=True):
        steps = []

        # If the new material has a much lower printing_temperature than the old material, the old material will not melt and will stick in the hotend.
        # So, we get the printing temp of the hotend's last_material property stored in the EEPROM later take the max of that 'old' temp and the 'new' temp of new material
        steps.append(StoreValueStep("STORE_PREV_MATERIAL_TEMP", self.__temperature_storage, "unload_temp",
                                    self.__getPrintTempOfLastMaterial))
        # So, instead of taking the printing_temperature of the new material, take the unload_temp, which is calculated to match both the new and old material.

        if pre_heat:
            # Start pre-heating the nozzle to speed-up the loading.
            # A safe value of 100C is chosen that won't degrade the materials.
            steps.append(OtherProcedureStep("PRE_HEAT_NOZZLE", "SET_HOTEND_TEMPERATURE",
                                            {"hotend_nr": self._hotend_nr, "target": 100}))

        # Special actions when loading Material 1
        if self._hotend_nr == 0:
            # Wait for the user to install the material with installed material guide
            steps.append(WaitMessageStep("WAIT_FOR_USER_TO_PLACE_MATERIAL_WITH_GUIDE", "DONE"))
        else:
            # Wait for the user to install the material
            steps.append(WaitMessageStep("WAIT_FOR_USER_TO_PLACE_MATERIAL", "DONE"))


        # Have the user select the material that he will insert. this step will also wait for the insertion! and have the material nr listed so people will know which material they are changing
        # The procedure is aborted after the timeout; then we don't know the material for sure
        steps.append(WaitForUserToSelectMaterialStep("USER_SELECT_MATERIAL", self._hotend_nr))


        # Get the printing_temperature from the material loaded when this step runs (so the newly loaded material)
        steps.append(StoreValueStep("STORE_LOAD_TEMP", self.__temperature_storage, "load_temp",
                                    lambda: float(self.__controller.getMaterialManager().getPrintingTemperature(self._hotend_nr))))

        # Here, only SET the temperature
        steps.append(OtherProcedureStep("STARTUP_HEATUP_TO_NEW_MATERIAL", "SET_HOTEND_TEMPERATURE", {"hotend_nr": self._hotend_nr, "target": self.__calcTemperatureForBothMaterials}))
        # Then wait until the temperature is printing_temperature - (time needed for forwarding / degrees per second of hotend)
        # Start forwarding when that temp is reached
        log.info("Loading will start early, when hotend is at {diff}degC from printing_temperature.".format(diff=self.__TEMPERATURE_TOLERANCE))

        # Wait until temperature is within some tolerance from the print_temperature
        steps.append(OtherProcedureStep("HEATUP_TO_NEW_MATERIAL_WAIT", "SET_HOTEND_TEMPERATURE_WAIT", {"hotend_nr": self._hotend_nr,
                                                                                                       "target": self.__calcTemperatureForBothMaterials,
                                                                                                       "wait_tolerance": self.__TEMPERATURE_TOLERANCE,
                                                                                                       "heatup_only": True}))
        # then forward material with a fast move.
        steps.append( OtherProcedureStep("FORWARD_MATERIAL",  "MOVE_MATERIAL", {"hotend_nr": self._hotend_nr, "target": self._bowden_tube_length - self.INSERT_SAFETY_LENGTH}))

        steps.append(WaitForQueueToBeEmptyStep("WAIT_FORWARD_MATERIAL_FINISHED"))
        # Wait till the user reports that the material is coming out of the hotend. At the same time, keep feeding forwards.
        # Keep the best temperature for both materials until (new) material comes out.
        # Wait for the user to confirm material coming out of the nozzle. After that time, simply assume it happened.
        steps.append(MoveMaterialWaitMessageStep("WAIT_FOR_MATERIAL_TO_COME_OUT_OF_HOTEND", self._hotend_nr, "forward", step_timeout=self.__USER_CONFIRM_EXTRUSION_TIMEOUT))

        if not self._skip_end and not self._in_print:
            # Simple retract to stop oozing.
            deprime_amount = self.__controller.getPropertyValue("deprime_retraction_amount")
            deprime_speed = self.__controller.getPropertyValue("deprime_retraction_speed")
            steps.append(GotoPositionStep("DEPRIME", e=-deprime_amount, speed=deprime_speed, relative_positions=True))
        else:
            # When we are in a print, we want to return the filament to where it was at the time when this change was initiated.
            filament_position_at_change = self.__controller.getHotendRetractedLength(self._hotend_nr)
            steps.append(GotoPositionStep("RETRACT", e=-filament_position_at_change, speed=15, relative_positions=True))  # return the hotend to have its previously retracted length.
            steps.append(OtherProcedureStep("START_COOLDOWN", "SET_HOTEND_TEMPERATURE", {"hotend_nr": self._hotend_nr, "target": "standby_temperature"}))

        return steps

    def _generateFinalizationSteps(self):
        steps = []
        # Finalization step. Also used as abort step.
        if not self._in_print:
            steps.append(OtherProcedureStep("FINISHING", "GENERIC_WIZARD_FINISHING_PROCEDURE"))
        elif self._hotend_nr != self._old_active_hotend:
            steps.append(OtherProcedureStep("SWITCH_TO_HOTEND", "SWITCH_ACTIVE_HOTEND", {"target": self._old_active_hotend}))
        if self._in_print and self._job_state == PrintJobStateProperty.PRINTING:
            steps.append(OtherProcedureStep("RESUME_AFTER_CHANGE_MATERIAL", "RESUME_PRINT", {"allow_resume_even_if_maintenance_procedures_active": True}))

        return steps

    # Override the getMetaData function to add our selected hotend nr.
    # This so consumers of the dbus-API know which hotend is currently being accessed with this procedure.
    def getMetaData(self):
        result = super().getMetaData()
        if result is None:
            result = {}
        result["hotend_nr"] = self._hotend_nr
        result["display_hotend_nr"] = self._hotend_nr + 1
        if isinstance(self._expected_material, str) or self._expected_material is None:
            result["expected_material"] = str(self._expected_material)

        return result

    ## @brief Calculate a temperature that is good for both the unloaded material and the new loaded material
    # This is done by taking the max of both, so that both materials will melt.
    def __calcTemperatureForBothMaterials(self):
        best_temperature = max(self.__temperature_storage.values())
        log.info("Calculating best temperature from {d}: {best}".format(d=self.__temperature_storage, best=best_temperature))
        return best_temperature

    # If the new material has a much lower printing_temperature than the old material, the old material will not melt and will stick in the hotend.
    # So, we store the temp with which to unload the old material and later take the max of that 'old' temp and the 'new' temp of new material
    #
    # Get the printing_temperature from the material loaded when this step runs (so the unloaded material)
    def __getPrintTempOfLastMaterial(self):
        try:
            log.debug('get_print_temp_of_last_material')
            last_material_guid = self.__controller.getPrintHeadController().getHotendSlot(
                self._hotend_nr).getPropertyValue("last_material_guid")
            last_material_fields = self.__controller.getMaterialManager().getMaterialNameFieldsByGuid(
                last_material_guid)
            log.debug("last_material = {last}: {fields}".format(last=last_material_guid, fields=last_material_fields))
            last_print_temp = self.__controller.getMaterialManager().getPrintingTemperatureForGuid(
                self._hotend_nr, last_material_guid)
            log.info("last_print_temp = {temp} ({last}: {fields})".format(temp=last_print_temp, last=last_material_guid,
                                                                          fields=last_material_fields))
            return float(last_print_temp)
        except:
            return 0
