import logging
from griffin.math.vector3 import Vector3

from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.homeBedStep import HomeBedStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.print.switchActiveHotendStep import SwitchActiveHotendStep
from griffin.printer.buildPlateLocations import BuildPlateLocations
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.waitForQueueEmptyStep import WaitForQueueToBeEmptyStep
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep
from griffin.printer.procedures.wizard.findNozzleOffsetStep import FindNozzleOffsetStep
from griffin.printer.procedures.wizard.checkCartridgeLoaded import CheckCartridgeLoaded

log = logging.getLogger(__name__.split(".")[-1])


## Manual Z offset procedure.
#  This procedure allows to configure the Z offset for the difference between hotend 0 and hotend 1.
class ManualNozzleOffsetProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)

    # Setup the Manual Z offset procedure.
    # @param args: dictionary, containing:
    #     chained_after_leveling: True if this procedure is combined with the manual bed leveling. If this is the case, we do not need to setup the printer by homing.
    def setup(self, args):
        steps = []
        if not args.get("chained_after_leveling", False):
            steps += [
                CheckCartridgeLoaded("CHECK_CARTRIDGE_LOADED", 0),
                CheckCartridgeLoaded("CHECK_CARTRIDGE_LOADED", 1),
                HomeBedStep("HOME_BED"),
                HomeHeadStep("HOME_HEAD"),
                SwitchActiveHotendStep("SWITCH_TO_HOTEND", target="reset"),
            ]
        steps += [
            SwitchActiveHotendStep("SWITCH_TO_HOTEND", target=1),

            GotoPositionStep("MOVE_HEAD_TO_BACK_POS", calculate_position_function=self.__getStartingPosition),
            WaitForQueueToBeEmptyStep("INIT_COMPLETE"),
            WaitMessageStep("WAIT", "CONTINUE"),
            WaitForQueueToBeEmptyStep("WAITING"),
            FindNozzleOffsetStep("FIND_OFFSET"),

            HomeBedStep("HOME_BED"),
            OtherProcedureStep("FINISHING", "GENERIC_WIZARD_FINISHING_PROCEDURE"),
        ]

        self._addStepSequence(steps)
        self.setAbortstepForMultiple(steps[:-1], steps[-1])

        return True

    def __getStartingPosition(self):
        position = BuildPlateLocations.getInstance().getPosition("CENTER_REAR_UNDER_2ND_NOZZLE", hotend_index=1)
        position += Vector3(0, 0, Controller.getInstance().getPrintHeadController().getHotendSlot(0).getPropertyValue("z_height"))
        return {"x": position.x, "y": position.y, "z": position.z}
