import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.steps.powerDownSteppersStep import PowerDownSteppersStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.gotoLocationStep import GotoLocationStep
from griffin.printer.procedures.steps.homeBedStep import HomeBedStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.steps.sleepStep import SleepStep
from griffin.printer.procedures.temperature.waitForMultipleHotendsTemperatureStep import WaitForMultipleHotendsTemperatureStep
from griffin.printer.procedures.waitForQueueEmptyStep import WaitForQueueToBeEmptyStep
from griffin.printer.procedures.temperature.setHotendsTemperatureStep import SetHotendsTemperatureStep
from griffin.printer.procedures.wizard.selectCleaningMaterialStep import SelectCleaningMaterialStep
from griffin.printer.procedures.wizard.selectPrintCoresToCleanStep import SelectPrintCoresToCleanStep
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep

log = logging.getLogger(__name__.split(".")[-1])


# The wizard procedure which guides the user to cleaning the hotend's nozzle(s)
# For the cold pull, the hotend is steered through a temperature profile:
# +
# | * __STAGE_HOT_PULL ('melt' the filament so the dirt and gunk mixes with the pulling filament
# |  \
# |   \    Screen: Maintain pressure on filament
# |    \
# |     \
# |      * __STAGE_STOP_PUSHING (when it's OK for the user to stop pushing in filament)
# |       \
# |        \
# |         \  Screen: You can release pressure on the filament now
# |          \
# |           * __STAGE_COLD_PULL (When the dirt fused with the pulling filament so we can pull it out)
# |            +------------ Screen: Stabilizing temperature
# |                                Screen: Pull the filament ...
# |
# +----------------------------+

class NozzleCleaning(Procedure):
    __STAGE_HOT_PULL        = 0 # temperature stages to be used as an index for __CLEANING_TEMPERATURES_* parameters
    __STAGE_STOP_PUSHING    = 1 # temperature stages to be used as an index for __CLEANING_TEMPERATURES_* parameters
    __STAGE_COLD_PULL       = 2 # temperature stages to be used as an index for __CLEANING_TEMPERATURES_* parameters
    __CLEANING_TEMPERATURES_PC = [270, 210, 135] # cleaning with polycarbonate temperature stages
    __CLEANING_TEMPERATURES_PLA = [230, 150, 75] # cleaning with PLA temperature stages
    __STABILISE_TEMPERATURE_TIME = 30 ## seconds to wait after reaching the cold pull temperature
    DATA_STORAGE_PRINTCORE_PREFIX = "printcore_"

    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)

    def setup(self, args):
        hotend_data_store = []

        # function used as callable for the SetHotendsTemperatureStep and the WaitForHotendsTemperatureStep to return the list of hotends to do this for (defined at runtime instead of at creation time)
        def calculate_hotends():
            hotends = [ int(data[len(self.DATA_STORAGE_PRINTCORE_PREFIX):]) for data in hotend_data_store if str(data).startswith(self.DATA_STORAGE_PRINTCORE_PREFIX) ]
            log.info("hotends to be cleaned: %s", str(hotends))
            return hotends

        # function used to generate a callable for the SetHotendsTemperatureStep and the WaitForHotendsTemperatureStep to return the temperature function to do this for (defined at runtime instead of at creation time)
        def calculate_temperature(stage_nr):
            return lambda: self.__CLEANING_TEMPERATURES_PLA[stage_nr] if "MATERIAL_PLA" in hotend_data_store else self.__CLEANING_TEMPERATURES_PC[stage_nr]

        # function used as callable for the SetPropertyStep to set the list of property owners at runtime.
        def calculate_hotend_properties():
            return ["printer.head.0.hotend.%d" % int(data[len(self.DATA_STORAGE_PRINTCORE_PREFIX):]) for data in hotend_data_store if str(data).startswith(self.DATA_STORAGE_PRINTCORE_PREFIX) ]

        # Generate all the steps for this wizard procedure.
        initial_steps = [
            HomeBedStep("HOME_BED"),
            HomeHeadStep("HOME_HEAD"),
            GotoLocationStep("GOTO_CORNER", "COLD_PULL"),
            WaitForQueueToBeEmptyStep("WAIT_FOR MOVE"),
            WaitMessageStep("SHOW_LINK_TO_WEBSITE", "CONTINUE", user_interaction=True),
            SelectPrintCoresToCleanStep("SELECT_PRINTCORES_TO_CLEAN", hotend_data_store),
            SelectCleaningMaterialStep("SELECT_CLEANING_MATERIAL", hotend_data_store),
            OtherProcedureStep("DEPRIME", "DEPRIME", args={"hotend_data_store": calculate_hotends, "cooldown": False}),
            SetHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_SET", {
                "hotend_nr": calculate_hotends,
                "target": calculate_temperature(self.__STAGE_HOT_PULL)} ),
            WaitForMultipleHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_WAIT", "all"),
            SetPropertyStep("INTERACTION_REQUIRED_ON", calculate_hotend_properties, "interaction_required", True),
            WaitMessageStep("REMOVE_BOWDEN_TUBE", "CONTINUE", user_interaction=True),
            SetPropertyStep("INTERACTION_REQUIRED_OFF", calculate_hotend_properties, "interaction_required", False),
        ]
        heatup_before_hot_pull_steps = [
            SetHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_SET", {
                "hotend_nr": calculate_hotends,
                "target": calculate_temperature(self.__STAGE_HOT_PULL)} ),
            WaitForMultipleHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_WAIT", "all"),
        ]
        hot_pull_steps = [
            SetPropertyStep("INTERACTION_REQUIRED_ON", calculate_hotend_properties, "interaction_required", True),
            WaitMessageStep("PERFORM_HOT_PULL", "CONTINUE", user_interaction=True),
            SetPropertyStep("INTERACTION_REQUIRED_OFF", calculate_hotend_properties, "interaction_required", False),
        ]
        heatup_before_cold_pull_steps = [
            SetHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_SET", {
                "hotend_nr": calculate_hotends,
                "target": calculate_temperature(self.__STAGE_HOT_PULL)} ),
            WaitForMultipleHotendsTemperatureStep("HEATUP_TO_CLEANING_TEMP_WAIT", "all"),
        ]
        cold_pull_steps = [
            SetPropertyStep("INTERACTION_REQUIRED_ON", calculate_hotend_properties, "interaction_required", True),
            WaitMessageStep("START_COLD_PULL", "CONTINUE", user_interaction=True),

            SetHotendsTemperatureStep("COOLDOWN_TO_STOP_PUSHING_TEMP", {
                "hotend_nr": calculate_hotends,
                "target": calculate_temperature(self.__STAGE_STOP_PUSHING)}),
            WaitForMultipleHotendsTemperatureStep("WAIT_FOR_STOP_PUSHING_TEMP", "all"),
            SetPropertyStep("INTERACTION_REQUIRED_OFF", calculate_hotend_properties, "interaction_required", False),

            SetHotendsTemperatureStep("COOLDOWN_TO_COLD_PULL_TEMP", {
                "hotend_nr": calculate_hotends,
                "target": calculate_temperature(self.__STAGE_COLD_PULL)}),
            WaitForMultipleHotendsTemperatureStep("WAIT_FOR_COLD_PULL_TEMP", "all"),

            SleepStep("NORMALIZE_TEMPERATURE_AFTER_PRINT", self.__STABILISE_TEMPERATURE_TIME),
            SetPropertyStep("INTERACTION_REQUIRED_ON", calculate_hotend_properties, "interaction_required", True),
            WaitMessageStep("PERFORM_COLD_PULL", "CONTINUE", user_interaction=True),
            SetPropertyStep("INTERACTION_REQUIRED_OFF", calculate_hotend_properties, "interaction_required", False),
            WaitMessageStep("REPEAT", "CONTINUE", user_interaction=True, alternate_outcome_messages=["REPEAT"])
        ]
        self._changeToStepSequence(heatup_before_hot_pull_steps + [hot_pull_steps[0]])
        self._changeToStepSequence(heatup_before_cold_pull_steps + [cold_pull_steps[0]])

        hot_or_cold_pull = WaitMessageStep("GO_TO_HOT_OR_COLD_PULL", None, user_interaction=True, alternate_outcome_messages=["COLD_PULL", "HOT_PULL"])
        hot_or_cold_pull.setNextConditionalStep("COLD_PULL", heatup_before_cold_pull_steps[0], confirm_loop_is_ok=True)
        hot_or_cold_pull.setNextConditionalStep("HOT_PULL", heatup_before_hot_pull_steps[0], confirm_loop_is_ok=True)

        cold_pull_steps[-1].setNextConditionalStep("REPEAT", hot_or_cold_pull, confirm_loop_is_ok=True)

        finish_steps = [
            WaitMessageStep("REATTACH_BOWDEN_TUBE", "CONTINUE", user_interaction=True),
            SetHotendsTemperatureStep("TURN_OFF_HEATERS", {"hotend_nr": list(range(Controller.getInstance().getPropertyValue("hotend_count"))), "target": 0} ),
            HomeHeadStep("HOME_HEAD"),
            PowerDownSteppersStep("POWER_DOWN_STEPPERS")
        ]

        self._addStepSequence(initial_steps + hot_pull_steps + cold_pull_steps + finish_steps)
        self.setAbortstepForMultiple(initial_steps, finish_steps[1]) # skip reattach
        self.setAbortstepForMultiple(hot_pull_steps + cold_pull_steps + [hot_or_cold_pull], finish_steps[0])

        return True

    # Override the getMetaData function to add our selected hotend nr.
    # This so consumers of the dbus-API know which hotend is currently being accessed with this procedure.
    def getMetaData(self):
        result = super().getMetaData()

        return result
