from timeit import default_timer as timer

import logging
log = logging.getLogger(__name__.split(".")[-1])


## @brief Measures time as a contextmanager
# Usage:
# >>> with Stopwatch("dummy", minimum_for_print=0.5):
# >>>     do_work()
# Then, if do_work() takes longer than 0.5 seconds, the actual time will be reported
class Stopwatch:

    ## @brief Stopwatch constructor
    # @param name str the name of Stopwatch
    # @param threshold float If the measured duration is longer than this threshold, a message will be logged when the contextmanager exits.
    def __init__(self, name, threshold=0.1):
        self.name = name
        self.threshold = threshold

    ## Part of the Contextmanager protocol in Python
    def __enter__(self):
        self.start = timer()
        return self

    ## Part of the Contextmanager protocol in Python
    def __exit__(self, *args):
        t = timer() - self.start
        if t >= self.threshold:
            log.warning("{name} : {dur:0.3g} seconds".format(name=self.name, dur=t))
        self.time = t