import logging
import os
import shutil
import socket
import subprocess
import threading
import time
import urllib.request
import urllib.error

from typing import Optional, Dict

from griffin import signal

from griffin.preferences.registryFile import RegistryFile
from griffin.thread import Thread
from griffin.timer import Timer

log = logging.getLogger(__name__.split(".")[-1])


# SystemUpdate object
# This class handles updating the system with new images.
# Updating the system is a relative complex process, the following features are supported:
# * Download a remote file
# * Checking a file signature
# * Initiate an update with the file
class SystemUpdate:
    STATE_IDLE = "IDLE"
    STATE_INIT_DOWNLOAD = "INIT_DOWNLOAD"
    STATE_DOWNLOADING = "DOWNLOADING"
    STATE_COPYING = "COPYING"
    STATE_VERIFYING = "VERIFYING"
    STATE_INSTALLING = "INSTALLING"
    STATE_FAILED_DOWNLOAD = "FAILED_DOWNLOAD"
    STATE_FAILED_USB = "FAILED_USB"
    STATE_FAILED_SIGNATURE = "FAILED_SIGNATURE"

    # Location of the version identification file
    __VERSION_FILE = "/etc/jedi_version"
    # Special development version
    __DEVELOPMENT_VERSION = "DEVELOPMENT_VERSION"

    __UPDATE_TEMP_FILENAME = "/tmp/um-update.swu"

    __UPDATE_NOTIFY_FILE = "system_update_notification.json"
    __UPDATE_KEY_PREV = "previous_version"
    __UPDATE_IGNORE_VERSION = "IGNORE_VERSION"

    __PRIVATE_UPDATE_SERVER_HOSTNAME = "monolith"  # Name of the update server
    __UPDATE_SERVER_IP_ADDRESS = "10.183.0.4"  # Internal IP address of the update server
    __PUBLIC_UPDATE_SERVER_HOSTNAME = "software.ultimaker.com"  # Host name of the public update server
    __FIRMWARE_DOWLOAD_ROOT = "/releases/firmware/"  # The root path on the server where the firmwares are stored
    __FIRMWARE_DOWNLOAD_FILE_NAME = "um-update.swu"  # The root filesystem file name to look for on the update server
    __VERSION_DOWNLOAD_FILE_NAME = "um-update.swu.version"  # The version file name to look for on the update server

    __START_DELAY_CHECK_INTERNAL_RELEASE = 30 ## Delay time in seconds before starting the thread that performs the checking
    __DELAY_CHECK_INTERNAL_RELEASE = 5 * 60 ## Delay time in seconds between checking the accessibility to the UPDATE_SERVER_HOSTNAME

    SYSTEM_RELEASE_TYPE_STABLE = "stable"
    SYSTEM_RELEASE_TYPE_TESTING = "testing"
    SYSTEM_RELEASE_TYPE_INTERNAL = "internal" ## Naming of the internal (developer) version

    __HAS_VERSION_STRINGS_UPDATE_TIMEOUT = 60 * 60
    __NO_VERSION_STRINGS_UPDATE_TIMEOUT = 2 * 60
    __SOCKET_TIME_OUT = 2 * 60 ## Socket time out in seconds if no data is transmitted (due to lost connection) during downloading of the firmware, so it will not wait forever

    #  @param bom_number The machine article number
    def __init__(self, bom_number: int) -> None:
        super().__init__()

        self.onUpdateStart = signal.Signal()
        self.onPreUpdateSuccessful = signal.Signal()
        self.onFirmwareVersionLoaded = signal.Signal()
        self.onFirmwareVersionLoadFailed = signal.Signal()
        self.onNewFirmwareVersionDetected = signal.Signal()

        self.__bom_number = bom_number  # type: int

        self._update_thread = None

        self._download_progress = 0.0
        self._state = self.STATE_IDLE

        self.__update_versions = {}
        self.__system_update_urls = {
            SystemUpdate.SYSTEM_RELEASE_TYPE_STABLE: self.__generateSystemUpdateUrlsFor(
                SystemUpdate.SYSTEM_RELEASE_TYPE_STABLE),
            SystemUpdate.SYSTEM_RELEASE_TYPE_TESTING: self.__generateSystemUpdateUrlsFor(
                SystemUpdate.SYSTEM_RELEASE_TYPE_TESTING),
        }  # type : Dict[str, Optional[Dict[str,str]]]
        for key in self.__system_update_urls.keys():
            self.__update_versions[key] = ""
        self.__update_versions_thread = None
        self.__update_notification_store = RegistryFile(self.__UPDATE_NOTIFY_FILE)

        socket.setdefaulttimeout(SystemUpdate.__SOCKET_TIME_OUT)

        check_for_internal_release = Timer("CHECK_INTERNAL_RELEASE", SystemUpdate.__START_DELAY_CHECK_INTERNAL_RELEASE, self.__checkForInternalRelease)
        check_for_internal_release.setThreaded(True)
        check_for_internal_release.start()

    ## Helper function for creating a dictionary with update file paths, given the update type
    #  @param update_type The update type
    #  @return A dictionary with the paths to the rootfs, the signature and the version file
    def __generateSystemUpdateUrlsFor(self, update_type: str) -> Optional[Dict[str, str]]:
        updateUrl = self.__createUpdateUrlFor(update_type)
        if updateUrl is None:
            return None

        # The current version is appended to the rootfs download url for statistical purpose, it is usefull information to know
        # what version is used before the updrade.
        version = self.getVersion()

        return {
            "update": "{url}{filename}?current_version={current_version}"
                .format(url=updateUrl, filename=SystemUpdate.__FIRMWARE_DOWNLOAD_FILE_NAME, current_version=version),
            "version": "{url}{filename}?current_version={current_version}"
                .format(url=updateUrl, filename=SystemUpdate.__VERSION_DOWNLOAD_FILE_NAME, current_version=version),
        }

    ## Helper function that creates update url strings, given the update type
    #  @param update_type The update type
    #  @return A html url string with the path where to find the latest firmware updates
    def __createUpdateUrlFor(self, update_type: str) -> Optional[str]:
        if update_type == SystemUpdate.SYSTEM_RELEASE_TYPE_STABLE:
            return "http://{server}{path}{bom}/stable/" \
                .format(server=SystemUpdate.__PUBLIC_UPDATE_SERVER_HOSTNAME, path=SystemUpdate.__FIRMWARE_DOWLOAD_ROOT,
                        bom=self.__bom_number)
        elif update_type == SystemUpdate.SYSTEM_RELEASE_TYPE_TESTING:
            return "http://{server}{path}{bom}/testing/" \
                .format(server=SystemUpdate.__PUBLIC_UPDATE_SERVER_HOSTNAME, path=SystemUpdate.__FIRMWARE_DOWLOAD_ROOT,
                        bom=self.__bom_number)
        elif update_type == SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL:
            return "http://{server}{path}{bom}/internal/" \
                .format(server=SystemUpdate.__PRIVATE_UPDATE_SERVER_HOSTNAME, path=SystemUpdate.__FIRMWARE_DOWLOAD_ROOT,
                        bom=self.__bom_number)
        return None

    ## @brief Checks if the internal update server (monolith) can be reached and has the expected address.
    #  This will run on a thread hence the never ending loop
    def __checkForInternalRelease(self):
        while True:
            try:
                if socket.gethostbyname(SystemUpdate.__PRIVATE_UPDATE_SERVER_HOSTNAME) == SystemUpdate.__UPDATE_SERVER_IP_ADDRESS:
                    if SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL not in self.__update_versions:
                        log.info("Adding INTERNAL to update URLs")
                        self.__update_versions[SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL] = ""
                        self.__system_update_urls[SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL] = \
                            self.__generateSystemUpdateUrlsFor(SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL)

            except Exception:
                # In case of any error we just ignore the errors and accept the fact that there was very likely a network issue
                # in which case it's safe to say the internal release is not available
                # Catching the exception also prevents the thread from aborting!
                self.__update_versions.pop(SystemUpdate.SYSTEM_RELEASE_TYPE_INTERNAL, None)   # 'None' parameter avoids throwing an exception when key not found.

            time.sleep(SystemUpdate.__DELAY_CHECK_INTERNAL_RELEASE)

    # Download an update and install it.
    # @param release_type: A key in the __SYSTEM_UPDATE_URLS dict to indicate which release should be downloaded.
    # @return False on error, True on successful start
    def downloadUpdate(self, release_type):
        # Check if there is already an update running.
        if self._update_thread is not None and self._update_thread.is_alive():
            return False

        if release_type == "default":
            release_type = self.SYSTEM_RELEASE_TYPE_STABLE
        # Get our urls from the __system_update_urls dict
        url_info = self.__system_update_urls.get(release_type, None)
        if url_info is None:
            return False

        image_url = url_info["update"]

        self._download_progress = 0.0
        self._state = self.STATE_INIT_DOWNLOAD

        self._update_thread = Thread("SystemUpdate", self._downloadUpdateThread, (image_url,))
        self._update_thread.start()
        return True

    # Load the given update file and install it.
    # @param update_filename: swu filename to install, needs to exist in the system.
    # @return False on error, True on successful start.
    def installLocalUpdate(self, update_filename: str) -> bool:
        # Check if there already is an update running.
        if self._update_thread is not None and self._update_thread.is_alive():
            return False

        self._download_progress = 0.0
        self._state = self.STATE_INIT_DOWNLOAD

        self._update_thread = Thread("SystemUpdate", self._installLocalUpdateThread, args=(update_filename,))
        self._update_thread.start()
        return True

    def getState(self):
        return self._state

    def getDownloadProgress(self):
        return self._download_progress

    ## Get the version strings for firmware updates.
    #  @return a string-string dictionary containing the known release names and version numbers we can update to.
    #          if a version string is not yet known, an empty string is returned.
    def getUpdateTargetVersions(self):
        return self.__update_versions

    ## Retrieve, from the remote server, which firmware updates are available.
    # @returns nothing, but emits the onFirmwareVersionLoaded or onFirmwareVersionLoadFailed signals
    def checkForFirmwareUpdates(self):
        # When the thread to get version numbers is not yet running, create a thread and fetch the version numbers.
        if self.__update_versions_thread is None or not self.__update_versions_thread.is_alive():
            self.__update_versions_thread = Thread("SystemUpdateVersions", self.__retrieveVersionStringsThread, daemon=True)
            self.__update_versions_thread.start()

    ## @brief Read and parses the version file for the current version
    #  @return The current version or the string self.__DEVELOPMENT_VERSION when there is no version file.
    def getVersion(self):
        if os.path.isfile(self.__VERSION_FILE):
            try:
                with open(self.__VERSION_FILE, "r") as f:
                    return ("".join(f.readlines())).strip()
            except Exception:
                log.error("Error while accessing %s", self.__VERSION_FILE)

        return self.__DEVELOPMENT_VERSION

    ## @brief Determines if there is an equal firmware version of the specified release type
    #  @param release_type The release type
    #  @retval False when the versions differ or if the current version is a development version.
    #  @retval True if the version numbers are equal.
    def hasEqualFirmware(self, release_type = SYSTEM_RELEASE_TYPE_STABLE):
        if release_type not in self.__update_versions:
            log.info("Release type '%s' version not yet available", release_type)
            return False

        current_version = self.getVersion()

        # Having the development version to be outdated everytime makes sense so that a development version can always be overwritten by any version
        if (current_version == self.__DEVELOPMENT_VERSION):
            return False

        if "DEV" in current_version.upper():
            # Takes care of the older [major].[minor].[revision]-DEV.YYYYMMDD version
            return False

        if "MOD" in current_version.upper():
            # Takes care of the modified version: MOD-[major].[minor].[revision].YYYYMMDD
            return False

        current_version = current_version.split("-", 1)[-1]

        new_version = self.__update_versions[release_type]
        is_equal = False
        try:
            is_equal = self._createTupleFromVersion(new_version) == self._createTupleFromVersion(current_version)
        except ValueError:
            log.warning("Error comparing firmware '%s' current version '%s' and new version '%s' - for safety assuming the current version is different", release_type, current_version, new_version)
            return False

        log.info("Comparing firmware '%s' old version '%s' with new version '%s' with result: %r", release_type, current_version, new_version, is_equal)
        return is_equal

    ## @brief Converts a version string (major.minor.revision.YYYYMMDD) into seperate values (tuple of 4 values)
    #  @return Returns a tuple (major, minor, revision, YYMMDD)
    def _createTupleFromVersion(self, version):
        return tuple(map(int, (version.split("."))))

    ## @brief Gets the version of the specified firmware release type
    #  @return The version of a certain release type as a string or "" when release type does not exist.
    def getFirmwareVersion(self, release_type):
        if release_type not in self.__update_versions:
            return ""
        return self.__update_versions[release_type]

    def _downloadProgressCallback(self, blocknum, blocksize, totalsize):
        if totalsize > 0:
            self._download_progress = float(blocknum * blocksize) / float(totalsize)

    ## @brief Downloads the firmware image file given by the image url
    #  @param image_url The url for downloading the image file
    def _downloadUpdateThread(self, image_url: str):

        self.onUpdateStart.emit()

        try:
            self._state = self.STATE_DOWNLOADING
            log.info("Starting download of: %s", image_url)
            urllib.request.urlretrieve(image_url, self.__UPDATE_TEMP_FILENAME, self._downloadProgressCallback)

        except Exception:
            log.exception("Exception during download of update")
            self._state = self.STATE_FAILED_DOWNLOAD
            return
        self._installUpdateThread()

    def _installLocalUpdateThread(self, update_filename):
        self.onUpdateStart.emit()

        try:
            self._state = self.STATE_COPYING
            log.info("Copying: %s to tmpfs", update_filename)
            shutil.copy(update_filename, self.__UPDATE_TEMP_FILENAME)
        except Exception:
            log.exception("Exception during copy of update")
            self._state = self.STATE_FAILED_USB
            return
        self._installUpdateThread()

    def _installUpdateThread(self):
        self._state = self.STATE_VERIFYING
        log.info("Files downloaded")
        if not self._verifyImage(self.__UPDATE_TEMP_FILENAME):
            log.info("Signature verification failed.")
            self._state = self.STATE_FAILED_SIGNATURE
            return
        log.info("Signature verification done")

        # Remember the version we're on BEFORE the actual update
        log.info("Storing current version {curr} to file".format(curr=self.getVersion()))
        self.__update_notification_store.setAsString(self.__UPDATE_KEY_PREV, self.getVersion())
        self.__update_notification_store.forceSave()

        self._state = self.STATE_INSTALLING

        try:
            log.info("Performing pre-update procedure.")
            p = subprocess.Popen(["sudo", "/usr/sbin/pre_update.sh", "%s" % (self.__UPDATE_TEMP_FILENAME)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
            result = p.wait() == 0
            if not result:
                log.error("Error trying to pre-update the printer: {error}.".format(error=err))
                self._state = self.STATE_FAILED_DOWNLOAD
                return

            log.info("Pre-update done, rebooting printer to perform the firmware update.")
            self.onPreUpdateSuccessful.emit()
        except Exception as e:
            log.error("Error trying to pre-update the printer: {exception}.".format(exception=e))
            self._state = self.STATE_FAILED_DOWNLOAD

    # Verify if the image we have checks out with any of the public keys we have.
    # @param: Filename of the image file to verify.
    # @return: Returns true when the signature is valid.
    @staticmethod
    def _verifyImage(filename: str) -> bool:
        try:
            p = subprocess.Popen(["/usr/bin/verify_swupdate.sh", filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
            result = p.wait() == 0
            if not result:
                log.error("Error trying to verify signature: {error}.".format(error=err))
            return result
        except Exception as e:
            log.error("Error trying to verify signature: {exception}.".format(exception=e))
            return False

    ## @brief The thread function in which version numbers are retrieved from the server.
    # When this succeeds, the signal 'onFirmwareVersionLoaded' is emitted and another check is scheduled
    # When this fails, the signal 'onFirmwareVersionLoadFailed' is emitted and another check is scheduled
    def __retrieveVersionStringsThread(self):
        for key in self.__update_versions.keys():
            try:
                with urllib.request.urlopen(self.__system_update_urls[key]["version"], timeout=60) as f:
                    version = f.read()
                self.__update_versions[key] = version.decode("utf-8").strip()
                log.debug("Version for release type '%s' found: %s", key, self.__update_versions[key])
                self.onFirmwareVersionLoaded.emit(key)
                if key == self.SYSTEM_RELEASE_TYPE_STABLE and self.getVersion() != self.__update_versions[key] and self.__update_versions[key] != self.__update_notification_store.get(self.__UPDATE_IGNORE_VERSION):
                    self.onNewFirmwareVersionDetected.emit(self.__update_versions[key])
            except Exception:
                # A stacktrace here is overkill - to log failure is enough (most likely a 404 / lost connection)
                log.warning("Error during updating of the update version strings for release type '%s'", key)
                self.onFirmwareVersionLoadFailed.emit(key)

        if all(self.__update_versions.values()):
            # If we have all versions, only update once an hour
            log.debug("Scheduling another checkForFirmwareUpdates in {delay} seconds".format(delay=self.__HAS_VERSION_STRINGS_UPDATE_TIMEOUT))
            threading.Timer(self.__HAS_VERSION_STRINGS_UPDATE_TIMEOUT, self.checkForFirmwareUpdates).start()
        else:
            # If we do not have all version strings yet, update every 2 minutes.
            log.debug("Scheduling another checkForFirmwareUpdates in {delay} seconds".format(delay=self.__NO_VERSION_STRINGS_UPDATE_TIMEOUT))
            threading.Timer(self.__NO_VERSION_STRINGS_UPDATE_TIMEOUT, self.checkForFirmwareUpdates).start()

    ## @brief Returns whether the system was updated to a newer version.
    # @returns bool: True if an update was completed and False otherwise.
    def wasUpdateCompleted(self):
        previous = self.__update_notification_store.getAsString(self.__UPDATE_KEY_PREV, default=self.getVersion())
        current = self.getVersion()
        different = current != previous
        log.info("The current version ({curr}) {eq} previous ({prev})".format(curr=current, prev=previous, eq={True:"!=", False:"=="}[different]))
        return different

    ## @brief Acknowledge that the system was updated so that the user is not notified again
    def acknowledgeUpdateCompleted(self):
        self.__update_notification_store.delete(self.__UPDATE_KEY_PREV)  # Delete the key after acknowledgement, we don't need it anymore.
        self.__update_notification_store.forceSave()

    ## set a specific version as the version to be ignored.
    # @param version The version to be ignored
    def ignoreVersion(self, version):
        self.__update_notification_store.set(self.__UPDATE_IGNORE_VERSION, version)
